#! /usr/bin/env python
from sys_appends import *
import sys
import traceback
import antlr3
import antlr3.tree

import src.globals.constants as constants

# The phases of the optimzier
from src.parser.VictorParserLexer import VictorParserLexer
from src.parser.VictorParserParser import VictorParserParser
import src.vsql_core.plan_generators.plan_generator as planner
import src.normalization.normalization_front as normalize
import src.main_memory.main_memory_normalize as main_mem_normalize


import argparse

parser = argparse.ArgumentParser(description='Victor Front End')
parser.add_argument('specfile', help='the specification filename')
parser.add_argument('--forcemm', dest='forcemm', type=bool, default=False,
                    help='Force Main Memory Execution')

#######################
# The Phases of Victor
#######################
# The parse phase takes in a file to parse.  The output is a set of
# ParserModelInstance or ParserModelCreate statements
def parse_phase(fname):
    char_stream = antlr3.ANTLRFileStream(fname)
    lexer  = VictorParserLexer(char_stream)
    tokens = antlr3.CommonTokenStream(lexer)
    parser = VictorParserParser(tokens)
    
    return parser.top_level().value

# Basic normalization: it creates wrapper functions and ensures that
# the tables necessary to store the models are setup.
# the output is a set of model description
def normalization(tasks):
    return normalize.normalize(tasks)

def concat(x,y):
    return x + y


def planning_phase(tasks):
    if len(tasks) == 0:
        return None
    elif len(tasks) == 1:
        mi_desc = tasks[0]
        return planner.generate_plan(mi_desc)
    else:
        return reduce(concat, [planner.generate_plan(mi_desc) for mi_desc in tasks])

# Sequential execute work

def execute_work(work, nThreads=12):    
    import src.vsql_core.executors.simple_executor as se
    se.execute_tasks(work,process=nThreads)

def should_use_main_memory(forcemm):
    return forcemm

def main(fname, forcemm):
    print "Parsing {0}...".format(fname)
    tasks = parse_phase(fname)

    if should_use_main_memory(forcemm):
        exec_tasks   = filter(main_mem_normalize.is_model_spec,tasks)
        create_tasks = filter(lambda x: not (main_mem_normalize.is_model_spec(x)),tasks)
        normalization(create_tasks)
        
        # We still plan and normalize all the tasks to ensure the relevant SQL structures are setup.
        # Here, we have to go one task at a time so that we can maintain an association between
        # the input main memory tasks, and our tasks.
        for t in exec_tasks:
            work     = planning_phase(normalization([t]))            
            job_spec = main_mem_normalize.main_memory_create_model_instance(t.params)            
            main_mem_normalize.main_mem_executor(job_spec, planned_work=work)

    else:
        print "Normalizing {0} tasks...".format(len(tasks))
        tasks = normalization(tasks)
    
        print "Planning {0} tasks...".format(len(tasks))
        work  = planning_phase(tasks)
    
        if work is not None:
            print "Executing {0} jobs".format(len(work))
            execute_work(work)

args = parser.parse_args()
main(args.specfile, args.forcemm)
