import collections
import examples.CRFs.simple_crf as simple_crf
import examples.CRFs.parse_template as parse_template    
from examples.CRFs.crf_model import SimpleCRFModel
from examples.CRFs.simple_crf import ModelVector

################################################################
#
# The model_type is crf_model_vector x compiled_template
#
################################################################
def internal_initialize_model(template_file_name, stepsize):
    t         = parse_template.parse_template(template_file_name)
    m         = simple_crf.build_model([],[],stepsize)
    return (m, t)

def initialize_model():
    return internal_initialize_model('/scratch.1/VICTOR_SQL/examples/CRFs/template.citeseer', 0.02)

#############################
#
# The Gradient step function
# We are maximizing so the rule is
# w^{k+1} = w^{k} + \nabla L(w^{k}, x)
# where L is the log likelihood.
#
def gradient_step( (m,t), (labeled_doc,)):
    d = parse_template.template_single_document(labeled_doc, t)    
    m = simple_crf.take_gradient_step(m, d)
    return (m,t)


#############################################
# The objective value is the log likelihood
# Expressed below.
def compute_objective_item( (m,t), (labeled_doc,) ):
    d         = parse_template.template_single_document(labeled_doc, t)
    z         = simple_crf.compute_normalization(m, d)
    v         = simple_crf.compute_weight_of_labeled(m, d)
    return z - v


#############################################
# The objective value is the log likelihood
# Expressed below.
def compute_objective_label( (m,t), labeled_doc ):
    d         = parse_template.template_single_document(labeled_doc, t)
    (wp,path) = simple_crf.compute_most_likely_path(m, d)
    tpath     = [ll for (s,u,b,ll) in d]
    return sum([ tpath[i] <> path[i] for i in xrange(len(path)) ])


def compute_label_text( (m,t), labeled_doc ):
    d         = parse_template.template_single_document(labeled_doc, t)
    (wp,path) = simple_crf.compute_most_likely_path(m, d)
    return "{}->{} {} doc={}".format([l.split('\t')[0] for l in labeled_doc], path, m.label_set, [dl[3] for dl in d])

# Stopping condition: TODO MEANINGFUL COMMENTS
def stopping_condition(s, loss):
    if not s.has_key('state'):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 100
