import re,sys
# return 
# name
compiled_match = re.compile('%x\[(-?\d+),(-?\d+)\]')
szDelim        = "\\"


## This takes in a template line in CRF format
## and creates a simple triple data structure:
## (name of the feature, is_the_feature_unigram
def parse_template_line(line,warn=False):
    if ':' in line:
        (name,tags) = line.split(':')
    else:
        name = line
        tags = ""
    isunigram = name[0].lower() == 'u'
    if not isunigram and name[0].lower() <> 'b' and warn:
        print "Warning: Assuming {0} is a bigram feature".format(line)
        
    pairs     = []
    for t in tags.split('/'):
        m = re.search(compiled_match, t)
        if m:
            v= (int(m.group(1)), int(m.group(2)))
            pairs.append(v)
        else:
            if warn:
                print "FAILED TO MATCH {0}".format(t)
    return (name, isunigram, pairs)

# This takes in a template file and returns a list of templates
# These templates are triples as described abvoe.
def parse_template(filename, debug=False):
    fh = open(filename, "r")
    templates = []
    for l in fh:
        l = l.rstrip()
        if len(l) > 0 and l[0] != "#":
            templates.append(parse_template_line(l))
    fh.close()
    # Debug Stats
    if debug:
        nFeatures = len(templates)
        nUnis     = len(filter(lambda x: x[1],templates))
        sys.stderr.write("nFeatures={0} nUnigrams={1}\n".format(nFeatures, nUnis))
        
    return templates

##############################################################
#
# This grounds a particular row column pair in a document.
#
def ground_pair(doc,line_num,r,c):
    absolute_row = r + line_num
    if absolute_row < 0:
        return "BoL{0}".format(absolute_row)
    elif absolute_row >= len(doc):
        return "EoL{0}".format(absolute_row)
    else:
        line = doc[absolute_row]
        if c < 0 or c >= len(line):
            print "line={0} c={1}".format(line,c)
            assert(False)
        if c == len(line) - 1:
            print "WARNING: LOOKING AT THE LABEL IN THE FEATURE!"
        
        return line[c]

##############################
#
# Ground a template and return it as a single string.
# CRF++ assumes this template string is unique.
#
##############################
def ground_template(doc,name,line_num,pairs):
    terms = [ground_pair(doc,line_num,r,c) for (r,c) in pairs]
    return "{0}:{1}".format(name, szDelim.join(terms))

##########################################
# template_document takes as input:
# 1. A document as a list of lines
# 2. A list of templates to apply to the document
# returns the templated document
def template_single_document(doc,templates):
    templated_lines = []
    for line_num in xrange(len(doc)):
        unigram_templates = []
        bigram_templates  = []
        for (name,isu,pairs) in templates:
            template_string = ground_template(doc,name,line_num,pairs)
            if isu :
                unigram_templates.append(template_string)
            else: 
                bigram_templates.append(template_string)
        l = doc[line_num]
        v = l.split('\t')
        label = v[len(v) - 1]
        templated_lines.append((doc[line_num],unigram_templates,bigram_templates,label))
    return templated_lines



##########################################
# template_document takes as input:
# 1. A document as a list of lines
# 2. A list of templates to apply to the document
# ------------------------------------------------------
# It puts output into the following locations
# 3. A list of templated documents (which are themselves list of templated lines). This is the output
def template_document(doc,templates,doc_template):
    templated_lines = []
    for line_num in xrange(len(doc)):
        unigram_templates = []
        bigram_templates  = []
        for (name,isu,pairs) in templates:
            template_string = ground_template(doc,name,line_num,pairs)
            if isu :
                unigram_templates.append(template_string)
            else: 
                bigram_templates.append(template_string)
        l = doc[line_num]
        label = l[len(l) - 1]
        templated_lines.append((doc[line_num],unigram_templates,bigram_templates,label))
    doc_template.append(templated_lines)


#######################################################
#
# This is a frontend function apply templated documents
#
def apply_template(template_file, token_file):
    # Parse the templates.
    templates = parse_template(template_file)

    label_set      = set()
    templated_docs = [] # The main outputs
    doc            = [] # stores the current document

    #########################
    # Open the token file
    token_fh = open(token_file)

    # For each line in the token file.
    for l in token_fh:
        l = l.rstrip()
        # If it's blank a document has ended.
        if len(l) == 0:
            # create the dictionary for the current document.
            if len(doc) > 0:                
                template_document(doc,templates,templated_docs)
            doc = []
        else:     
            raw_features = l.split()
            doc.append(raw_features)
            # Extract the label, since we're here.
            v = l.split()
            l = v[len(v) - 1]
            label_set.add(l)
            
    token_fh.close()
    
    #####################################################
    # check if there is a remaining document in the list
    if(len(doc) > 0):
        line_no = template_document(doc,templates,templated_docs)

    #####################################################
    # Return the document template and the set of labels.
    return (templated_docs,label_set)
