import sys, time, random, argparse
import examples.CRFs.simple_crf

#####################################
#
# Compute Statistics on the dataset #
#
#####################################
def stat_string(m, ds, msg, verbose=False):
    (correct, total, obj) = simple_crf.score_documents(m, ds, verbose=verbose)
    norm = m.model_vector.dot(m.model_vector)
    return "[{0}: correct={1} total={2} percentage={3} train_obj={4} norm_sq={5}]".format(msg, correct, total, float(correct)/total, obj, norm)


def front(template_file, token_file, test_file, nEpochs=20, print_model=False, alpha=5e-1, batch_gradient=False):
    import parse_template        
    sys.stderr.write("template file={0} token_file={1}\n".format(template_file, token_file))
    (ds,label_set) = parse_template.apply_template(template_file, token_file)
        
    # Build the initial model
    build_start = time.time()
    m = simple_crf.build_model(ds,label_set, 1e-1)
    print "Model Constructed in {0} seconds".format(time.time() - build_start)
    
    print stat_string(m, ds, "Initial Numbers")
    counter = 1

    # Starting the training
    train_str = "incremental gradient"
    if batch_gradient:
        train_str = "batch gradient"
    print "**** Starting Training using {0} ****".format(train_str)        
    for epoch in xrange(nEpochs):
        t0 = time.time()
        random.shuffle(ds)
        if batch_gradient:
            stepsize = alpha/(1.0 + counter)
            simple_crf.batch_gradient(m, ds, stepsize)
            counter += len(ds)
        else:
            for d in ds:
                stepsize = alpha/(1.0 + counter)
                counter += 1
                simple_crf.take_gradient_step(m, d, regularize=None, sigma=10.0)            
        tend = time.time()
            
        train_stats  = stat_string(m, ds, "Train Stats")
        test_stats   = ""
        if test_file is not None:
            (test_ds,label_set) = parse_template.apply_template(template_file, test_file)
            test_stats = stat_string(m, test_ds, "Test Stats")

        stat_strings = ' '.join([train_stats, test_stats])
        print "[End Epoch {0} in {1}]\n\t{2}".format(epoch, tend - t0, stat_strings)

    if print_model:
        print "\n**** MODEL ****\n"
        print m.model_vector.print_model()


# Parser Arguments
parser = argparse.ArgumentParser(description='Simple CRF Front End')
parser.add_argument('template_file', help='template file')
parser.add_argument('token_file'   , help='token file')
parser.add_argument('--test_file', help='test file')
parser.add_argument('--epochs', default=10, type=int,help='number of epochs')
parser.add_argument('--batch', type=bool, default=False, help='use a batch')
args = parser.parse_args()

front(args.template_file, args.token_file, args.test_file, args.epochs, batch_gradient=args.batch)
