"""
This module contains functions for linear gradients
"""

import src.vsql_core.utils.victor_utils as victor_utils
import numpy as np
from math import *

#########################################
# NAIVELY ASSUMES CONSTANT STEP SIZE!!!
# THESE ARE HACKS TO MAKE L1 PROX FASTER!!!
##########################################
def lazy_shrink(w, current_time, timestamps, indexes, u):
   """
   Lazy shrink function
   @type w: vector
   @param w: model vector
   @type current_time: double
   @param current_time: current time
   @type timestamps: vector
   @param timestamps: time stamps vector
   @type indexes: vector
   @param indexes: indexes of the example
   @type u: double
   @param u: u
   """
   for i in indexes:
      delta = current_time - timestamps[i]
      ud    = delta* u
      if w[i] > ud:
         w[i] = w[i] - ud
      else:
         if w[i] < -ud:
            w[i] = w[i] + ud
         else:
            w[i] = 0.0


### LOGISTIC!!
def l1_prox_lazy(w, what, timestamps, current_time, steps, stepsize, mu, B, indexes, vecs, y):
   """
   l1 prox lazy function
   @type w: vector
   @param w: model vector
   @type what: vector
   @param what: w hat
   @type timestamps: vector
   @param timestamps: time stamps vector
   @type current_time: double
   @param current_time: current time
   @type steps: double
   @param steps: steps of the incremental stage
   @type stepsize: number
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   @type indexes: vector
   @param indexes: indexes vector of the example's feature vector
   @type vecs: vector
   @param vecs: vector values of the example's feature vector
   @type y: number
   @param y: label of the example
   """
   lazy_shrink(w, current_time, timestamps, indexes, stepsize*mu)
   #lazy_average_update(w, what, timestamps, current_time, indexes)
   # update the timestamps
   for i in indexes:
      timestamps[i]        = current_time
   wx  = victor_utils.dot_dss(w, indexes, vecs)
   sig = victor_utils.sigma(- wx*y)
   victor_utils.scale_and_add_dss(w, indexes, vecs, stepsize*y*sig)   
   incremental_average_sparse(what, w, steps, indexes)


def svm_l1_prox_lazy(w, what, timestamps, current_time, steps, stepsize, mu, B, indexes, vecs, y):
   """
   svm l1 prox lazy function
   @type w: vector
   @param w: model vector
   @type what: vector
   @param what: w hat
   @type timestamps: vector
   @param timestamps: time stamps vector
   @type current_time: double
   @param current_time: current time
   @type steps: double
   @param steps: steps of the incremental stage
   @type stepsize: number
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   @type indexes: vector
   @param indexes: indexes vector of the example's feature vector
   @type vecs: vector
   @param vecs: vector values of the example's feature vector
   @type y: number
   @param y: label of the example
   """
   lazy_shrink(w, current_time, timestamps, indexes, stepsize*mu)
   wx  = victor_utils.dot_dss(w, indexes, vecs)
   if(y * wx < 1):
      victor_utils.scale_and_add_dss(w, indexes, vecs, stepsize*y)
   incremental_average_sparse(what, w, steps, indexes)


def svm_l2_prox_lazy(w, what, wscale, steps, stepsize, mu, B, indexes, vecs, y):
   """
   svm l2 prox lazy function
   @type w: vector
   @param w: model vector
   @type what: vector
   @param what: w hat
   @type wscale: double
   @param wscale: used to scale model vector(w)
   @type steps: double
   @param steps: steps of the incremental stage
   @type stepsize: number
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   @type indexes: vector
   @param indexes: indexes vector of the example's feature vector
   @type vecs: vector
   @param vecs: vector values of the example's feature vector
   @type y: number
   @param y: label of the example
   """
   wscale = 1/(1+stepsize*mu)
   if(wscale < 1e-10):
      victor_utils.scale(w,wscale)
      wscale = 1.0
      
   wx  = victor_utils.dot_dss(w, indexes, vecs)
   if(y * wx < 1):
      victor_utils.scale_and_add_dss(w, indexes, vecs, stepsize*y)
   incremental_average_sparse(what, w, steps, indexes)
