# these are loss functions for linear models
import src.vsql_core.utils.victor_utils as victor_utils
from math import *

def hinge_loss(lm, x, y):
   """
   Calculates and returns hinge loss
   @type w: vector
   @param w: model vector
   @type x: vector
   @param x: feature vector of the example
   @type y: number
   @param y: label of the example

   @rtype: double
   @return: returns the hinge loss
   """
   wx = victor_utils.dot(lm.w, x)
   return max(1 - y*wx,0)

def sparse_hinge_loss(lm,indexes,vecs, y):
   """
   Calculates hinge lss for sparse vectors and returns hinge loss value
   @type w: vector
   @param w: model vector
   @type indexes: vector
   @param indexes: indexes of the feature vector of the example
   @type vecs: vector
   @param vecs: vector values of the feature vector of the example
   @type y: number
   @param y: label of the example

   @rtype: double
   @return: returns the hinge loss
   """
   wx = victor_utils.dot_dss(lm.w, indexes, vecs)
   return max(1 - y*wx,0)

def mse_loss(lm, x, y):
   """
   Calculates and returns mse loss
   @type w: vector
   @param w: model vector
   @type x: vector
   @param x: feature vector of the example
   @type y: number
   @param y: label of the example

   @rtype: double
   @return: returns the mse loss
   """
   wx = victor_utils.dot(lm.w, x)
   v  = wx - y
   return v*v

def logit_loss_ss(model, (index, vecs, y) ):
   """
   Calculates logit loss for sparse vectors and returns the value
   @type w: vector
   @param w: model vector
   @type index: vector
   @param index: indexes of the feature vector of the example
   @type vecs: vector
   @param vecs: vector values of the feature vector of the example
   @type y: number
   @param y: label of the example

   @rtype: double
   @return: returns the logit loss
   """
   lm  = model[0]
   wx  = victor_utils.dot_dss(lm.w, index, vecs)
   err = log( 1+ exp( - y*wx ) )
   return err

def logit_loss_dense(lm, x, y):
   wx  = victor_utils.dot(lm.w, x, y)
   err = log( 1+ exp( - y*wx ) )
   return err
    
