"""
This module contains functions for logitistic regression with an l1 penalty
"""

import src.vsql_core.utils.victor_utils as victor_utils
import examples.LINEAR_MODELS.simple_linear as simple_linear
import math

def logit_l1_grad( model, (indexes, vectors, y) ):
   """
   logit l1 sparse function
   @type w: vector
   @param w: model vector
   @type mu: double
   @param mu: mu
   @type indexes: vector
   @param indexes: indexes vector of the example's feature vector
   @type vectors: vector
   @param vectors: vector values of the example's feature vector
   @type step: double
   @param step: steps of the incremental stage
   @type y: number
   @param y: label of the example
   """
   lm      = model[0]
   wx      = victor_utils.dot_dss(lm.w,indexes,vectors)
   sig     = victor_utils.sigma(-wx*y)
   
   # NB: derivative is - y*x sigma(-wx*y) (so - gradient
   victor_utils.scale_and_add_dss(lm.w,indexes, vectors, lm.stepsize*y*sig)
   victor_utils.l1_shrink_mask(lm.w, lm.mu*lm.stepsize,indexes)
   model[0].take_step()
   return model


def init():
    return (simple_linear.LinearModel(41270,mu=1e-2),)

def stop_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 10
