import numpy, math, array
import src.vsql_core.utils.victor_utils as victor_utils

def l2_project_np(w, b):
    norm_sq = numpy.dot(w,w)
    if norm_sq > b*b:
        n = b/math.sqrt(norm_sq)
        w *= n
    return w
        
class NumpyLinearModel:
    def __init__(self, dim, B, initial_stepsize=0.9):
        self.steps = 0
        self.dim   = dim
        self.w     = numpy.array([1e-2 for i in range(dim)])
        self.initial_stepsize = initial_stepsize
        self.stepsize         = initial_stepsize
        self.B     = B
        self.epoch = 0

    def take_step(self):
        self.steps += 1
        self.stepsize = self.initial_stepsize/(1+self.steps)

def create_numpy_model(dim,B):
    return NumpyLinearModel(dim,B)


### These functions are registered during model specification
### Any model will call them.
def grad( (nlm,mm), (x, y) ):    
    x   = l2_project_np(x,1.0)
    
    wx  = numpy.dot(nlm.w, x)
    err = (wx - y)
    etd = numpy.float(- nlm.stepsize * err)
    
    nlm.w += x*etd
    nlm.w = l2_project_np(nlm.w, nlm.B)
    nlm.take_step()
    return (nlm,mm)

def mse_loss( (nlm,mm) , (x, y) ):    
    x  = l2_project_np(x,1.0)
    
    wx = numpy.dot(nlm.w, x)
    v  = wx - y
    return v*v

############################################################
### This the front end code for a particular model instance
###
def init():
    return (create_numpy_model(54,100.0),numpy.array([0.0 for i in range(10)]))

def stopping_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 10

def dot( (nlm,mm) , x):
    x   = numpy.array(x)
    return numpy.dot(nlm.w, x)

def bulkgrad( (nlm,m), data, perm ):
    xdata = map(lambda x: l2_project_np(numpy.array(x[0]), 1.0), data)
    ydata = array.array('i', map(lambda x: numpy.array(x[1]), data))
    
    w = nlm.w
    nItems = len(data)
    for epoch in xrange(5):
        stepsize = nlm.stepsize
        for i in xrange(nItems):
            x   = xdata[i]
            y   = ydata[i]
            wx  = numpy.dot(nlm.w, x)
            err = (wx - y)
            etd = numpy.float(- stepsize * err)
    
            w += x*etd
            w = l2_project_np(nlm.w, nlm.B)
        nlm.stepsize *= 0.9
