import math, random
import src.vsql_core.utils.victor_utils as victor_utils
import src.vsql_core.utils.low_rank_helper as low_rank_helper

def make_random_vec(dim):
    return [ random.gauss(0,1) * 1e-2 for i in range(dim) ]


def ball_project(x, B, B2):
    norm_square = sum([xi*xi for xi in x])
    if(norm_square > B2):
        norm = math.sqrt(norm_square)
        for i in range(len(x)):
            x[i] = x[i]* B/norm

def make_initial(max_rank, nRows, nCols, B):
    L = []
    R = []
    for i in range(nRows):
        random_vec = make_random_vec(max_rank)
        ball_project(random_vec, B, B*B)
        L.append(random_vec)
        
    for i in range(nCols):
        random_vec = make_random_vec(max_rank)

        ball_project(random_vec, B, B*B)
        R.append(random_vec)
    return (L,R)

class LowRankModel:
    def __init__(self, max_rank, nRows, nCols, B, initial_stepsize=0.9):
        self.steps = 0
        self.max_rank = max_rank
        (L,R)      = make_initial(max_rank, nRows, nCols, B)
        self.L     = L
        self.R     = R
        self.B     = B
        self.B2    = B*B
        self.epoch = 0
        self.stepsize = initial_stepsize

    def take_step(self):
        self.steps += 1
        self.stepsize = 0.9/(1+self.steps)

    def set_L(self, i, li):        
        self.L[i] = li

    def set_R(self, j, rj):
        import os
        print "\t update index {} from pid={}".format(j, os.getpid())
        self.R[j] = rj


def se_loss(m, row, col, rating):
    i   = row - 1 #matlab...
    j   = col - 1
    v      = low_rank_helper.dot(m.L[i], m.R[j]) - rating
    return v*v

def grad(m, row, col, rating):
    L   = m.L
    R   = m.R
    i   = row - 1 # matlab...
    j   = col - 1
    err = low_rank_helper.dot(L[i],R[j])  - rating	
    e   = -(m.stepsize * err)
    tempLi = list(L[i])
    low_rank_helper.scale_and_add(tempLi, R[j],e)   			    
    low_rank_helper.scale_and_add(R[j],   L[i],e)
    L[i] = tempLi
    low_rank_helper.ball_project(L[i],m.B,m.B2)
    low_rank_helper.ball_project(R[j],m.B,m.B2)
    m.set_L(i, L[i])
    m.set_R(j, R[j])
    return m

def initialize_model():
    nRows = 6040
    nCols = 3952 # TODO QUERY FOR THESE
    return LowRankModel(20, nRows, nCols, 1.5)


def stopping_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 5
