import math, random, array
import src.vsql_core.utils.victor_utils as victor_utils
import src.vsql_core.utils.low_rank_helper as low_rank_helper
import src.utils.shared_storage as ss
 
import numpy
cimport numpy

def make_random_vec(dim):
    return numpy.array([ random.gauss(0,1) * 1e-2 for i in range(dim) ])


def numpy_ball_project(x, B, B2):
    norm_square = numpy.dot(x,x)
    if(norm_square > B2):
        norm = math.sqrt(norm_square)
        x *= B/norm


def ball_project(x, B, B2):
    norm_square = sum([xi*xi for xi in x])
    if(norm_square > B2):
        norm = math.sqrt(norm_square)
        for i in range(len(x)):
            x[i] = x[i]* B/norm

def make_initial(max_rank, nRows, nCols, B):
    L = []
    R = []
    for i in range(nRows):
        random_vec = make_random_vec(max_rank)
        ball_project(random_vec, B, B*B)
        L.append(random_vec)
        
    for i in range(nCols):
        random_vec = make_random_vec(max_rank)

        ball_project(random_vec, B, B*B)
        R.append(random_vec)
    return (L,R)

class LowRankModel:
    def __init__(self, max_rank, nRows, nCols, moviemean, B, initial_stepsize=5e-2):
        self.steps = 0
        self.nRows     = nRows
        self.nCols     = nCols
        self.max_rank  = max_rank
        self.B         = B
        self.B2        = B*B
        self.epoch     = ss.get_epoch()
        self.moviemean = moviemean
        self.stepsize  = initial_stepsize

    def take_step(self):
        #self.steps += 1
        #self.stepsize = 0.9/(1+self.steps)
        self.stepsize = 5e-2 * 0.9**(self.epoch)

def se_loss(model, data):
    (m,L,R) = model
    cdef int i = data[0] - 1 # rating
    cdef int j = data[1] - 1
    cdef float rating = data[2]    
    v      = numpy.dot(L[i], R[j]) + m.moviemean - rating
    return v*v

def grad(model, data):
    (m,L,R) = model
    cdef int i        = data[0] - 1
    cdef int j        = data[1] - 1
    cdef float rating = data[2]
    #i   = row - 1 # matlab...
    #j   = col - 1
    err = numpy.dot(L[i],R[j]) + m.moviemean - rating	
    e   = -(m.stepsize * err)
    tempLi = L[i] + e*R[j]
    R[j]   += e*L[i]
    L[i]    = tempLi
    numpy_ball_project(L[i],m.B,m.B2)
    numpy_ball_project(R[j],m.B,m.B2)
    m.take_step()
    return (m,L,R)

def bulkgrad(model, batch_data, perm):
    (m,L,R) = model
    cdef int i, j, k, mini_epoch, nItems
    cdef float stepsize, err, e, B, B2, moviemean, rating
    cdef int nMiniEpochSize = 10
    B         = m.B
    B2        = m.B2

    moviemean = m.moviemean
    row_data = array.array('i', map(lambda x:x[0] -1, batch_data))
    col_data = array.array('i', map(lambda x:x[1] -1, batch_data))
    rat_data = array.array('f', map(lambda x:x[2],    batch_data))
    nItems = len(batch_data)
    m.take_step()
    m.epoch  = ss.get_epoch()*nMiniEpochSize
    stepsize = m.stepsize
    
    for mini_epoch in xrange(nMiniEpochSize):
        for k in xrange(nItems):
            i      = row_data[k]
            j      = col_data[k]
            rating = rat_data[k]
            err = numpy.dot(L[i],R[j]) + moviemean - rating	
            e   = -(stepsize * err)
            tempLi = L[i] + e*R[j]
            R[j]   += e*L[i]
            L[i]    = tempLi
            numpy_ball_project(L[i],B,B2)
            numpy_ball_project(R[j],B,B2)

            

def factor_partition(state, model, seed, nSplits2, t):
    (m,L,R) = model
    import random, math, array
    nRows = m.nRows
    nCols = m.nCols
    if (not state.has_key('seed')) or not (state['seed'] == seed):
        state['seed']     = seed
        random.seed(seed)
        row_perm = range(nRows)
        col_perm = range(nCols)
        random.shuffle(row_perm)
        random.shuffle(col_perm)
        state['row_perm'] = array.array('i', row_perm)
        state['col_perm'] = array.array('i', col_perm)
        print "SHUFFLED for seed={}".format(seed)
        
    row_perm = state['row_perm']
    col_perm = state['col_perm']
    nSplits = int(math.sqrt(nSplits2))

    r0 = t["row"] - 1
    c0 = t["col"] - 1
    if r0 >= nRows or c0 >= nCols or r0 < 0 or c0 < 0 or len(row_perm) < nRows or len(col_perm) < nCols:
        raise ValueError("r={} c={}".format(r, c))

    r   = row_perm[r0]
    c   = col_perm[c0]
    row_bucket = min(int(r/(nRows/nSplits)),nSplits-1)
    col_bucket = min(int(c/(nCols/nSplits)),nSplits-1)
    b   = max(min((row_bucket*nSplits + col_bucket), nSplits2 -1),0)
    return b    

def generic_initialize(dim, nRows, nCols, moviemean, B):   
    m     = LowRankModel(dim, nRows, nCols, moviemean, B)
    (L,R) = make_initial(dim, nRows, nCols, m.B)
    return (m,numpy.array(L),numpy.array(R))



def stopping_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 10



def table_based(table_name, dim=30):
    import src.globals.global_utils as gu
    conn = gu.connect_db()
    cur  = conn.cursor()
    cur.execute("select max(row), max(col), avg(rating) from {table_name}".format(table_name=table_name))
    (rows, cols, moviemean,) = cur.fetchone()
    return generic_initialize(dim, rows, cols, moviemean, 1.5)

### Front end code
def movielens10m_init():
    return table_based("movielens10m")

def movielens1m_init():
    return table_based("movielens1m")

def netflix_init():
    return table_based("lr_data")
