import src.globals.global_utils as global_utils
import src.globals.constants as constants
import sys
import sys_appends

def create_library_setup_function(victor_path):
    fnc_string = """DROP FUNCTION IF EXISTS library_setup() CASCADE;
    CREATE FUNCTION library_setup() RETURNS VOID
    AS $$
    \timport sys
    \tsys.path.append('{victor_path}')
    $$ LANGUAGE plpythonu;
    """.format(victor_path = victor_path)

    return fnc_string

def create_partition_data_fnc():
    fnc_string = """
    DROP FUNCTION IF EXISTS partition_data_fnc(text, text, text, int);
    CREATE FUNCTION partition_data_fnc(part_fnc text, input_view_name text, output_base text, num_split int) RETURNS VOID 
    AS $$
    \timport src.vsql_core.utils.factored_parser.partition_data as partition_data
    \tpartition_data.partition_data(part_fnc, input_view_name, output_base, num_split)
    $$ LANGUAGE plpythonu;
    """

    return fnc_string

def create_plpythonu_lang_if_not_exists(cur):
    cur.execute("SELECT COUNT(*) FROM pg_language WHERE lanname='plpythonu';")
    (lang_count,) = cur.fetchone()
    if lang_count == 0:
        cur.execute("CREATE LANGUAGE plpythonu;")

def main():
    if len(sys.argv) != 2:
        print "victor path is not specified"
        return
    else:
        victor_path = sys.argv[1]
        conn = global_utils.connect_db()
        cur  = conn.cursor()

        ##create plpythonu lang if it does not exist
        create_plpythonu_lang_if_not_exists(cur)

        library_setup_fnc = create_library_setup_function(victor_path)
        partition_data_fnc = create_partition_data_fnc()

        setup_queries = """
        {create_library_setup}
        {create_partition_data_fnc}

        DROP TABLE IF EXISTS {victor_model_table};
        CREATE TABLE {victor_model_table}({model_name} text, {model_type} text, {model_table_name} text, 
        {model_field_types} text[], {model_field_names} text[], {data_field_types} text[], 
        {data_field_names} text[], {objective} text, {objective_agg} text, {wrapper_fnc} text);

        DROP TABLE IF EXISTS {victor_global_instances};
        CREATE TABLE {victor_global_instances}({mid} serial, {model_instance} text, {model_spec} text);

        DROP TABLE IF EXISTS {victor_model_applications};
        CREATE TABLE {victor_model_applications}({model_app_name} text, {model_inst_name} text, {app_mid} int, 
        {function_content} text);
        """.format(create_library_setup      = library_setup_fnc,
                   create_partition_data_fnc = partition_data_fnc,
                   victor_model_table        = constants.VICTOR_MODEL_TABLE,
                   model_name                = constants.VMT_MODEL_NAME,
                   model_type                = constants.VMT_MODEL_TYPE,
                   model_table_name          = constants.VMT_MODEL_TABLE_NAME,
                   model_field_types         = constants.VMT_MODEL_FIELD_TYPES,
                   model_field_names         = constants.VMT_MODEL_FIELD_NAMES,
                   data_field_types          = constants.VMT_DATA_FIELD_TYPES,
                   data_field_names          = constants.VMT_DATA_FIELD_NAMES,
                   objective                 = constants.VMT_OBJECTIVE,
                   objective_agg             = constants.VMT_OBJECTIVE_ITEM,
                   wrapper_fnc               = constants.VMT_WRAPPER_FNC,
                   victor_global_instances   = constants.VICTOR_GLOBAL_MODEL_ID_TABLE,
                   mid                       = constants.VICTOR_GLOBAL_MID,
                   model_instance            = constants.VICTOR_GLOBAL_INSTANCE,
                   model_spec                = constants.VICTOR_GLOBAL_MODEL_SPEC_NAME,
                   victor_model_applications = constants.VICTOR_MODEL_APP_TABLE,
                   model_app_name            = constants.VMA_MODEL_APP_NAME,
                   model_inst_name           = constants.VMA_MODEL_INST_NAME,
                   app_mid                   = constants.VMA_MID,
                   function_content          = constants.VMA_FUNCTION_CONTENT)
                   
        cur.execute(setup_queries)
        conn.commit()
        print "Victor is set up for the database:{dbname}".format(dbname=sys_appends.dbname)

if __name__ == "__main__":
    main()
