#! /usr/bin/env python
import sys
import traceback
import antlr3
import antlr3.tree
from src.parser.VictorParserLexer import VictorParserLexer
from src.parser.VictorParserParser import VictorParserParser
import src.globals.constants as constants
import src.parser.ParserTypes as ParserTypes
import src.vsql_core.plan_generators.plan_generator as planner
import src.normalization.normalization_front as normalize
import src.normalization.create_model as create_model
import collections, multiprocessing, multiprocessing.sharedctypes
import cPickle, cStringIO
import src.utils.shared_storage as shared_storage


import src.utils.retrieve_and_store_methods as store_model_methods
from multiprocessing import Pool
import src.globals.global_utils as global_utils, time, random, math, array
import main_memory_simple_exec as mm_se
import factor_simple_numpy_hack
import numpy

row_data    = None
col_data    = None
rating_data = None
grad_fn     = None
part_fn     = None
bulk_fn     = None
loss_fn     = None
perm        = None

def execute_grad( mid, nItems, r, c, round_map, chunk_map, item_map, li ):
    (model, model_types) = shared_storage.get_cache_model(mid)
    last_index = li
    while last_index > 0:
        i = last_index
        # assert( r == round_map[i] and c == chunk_map[i] )
        grad_fn(model, (row_data[i], col_data[i], rating_data[i]) )
        last_index = item_map[i]

def _bulk_execute_grad( mid, nItems, r, c, round_map, chunk_map, item_map, li ):
    (model, model_types) = shared_storage.get_cache_model(mid) 
    last_index = li
    data = []
    while last_index > 0:
        i = last_index
        # assert( r == round_map[i] and c == chunk_map[i] )
        data.append( (row_data[i], col_data[i], rating_data[i]) )
        last_index = item_map[i]
    bulk_fn(model, data, [])
            
def compute_loss( (mid, l, nParts, part_id) ):
    (model, model_types) = shared_storage.get_cache_model(mid)
    loss = 0.0
    for i in xrange(0,l,nParts):
        pi = i + part_id
        if(pi < l):
            loss += loss_fn(model, (row_data[pi], col_data[pi], rating_data[pi]) )
    return loss

def slow_loss( mid, nItems ):
    (model, model_types) = shared_storage.get_cache_model(mid)
    loss = 0.0
    for i in xrange(0,nItems):
        loss += loss_fn(model, (row_data[i], col_data[i], rating_data[i]) )
    return loss
               

## Generic Routine
def load_into_shared_memory(sql_data, type_string):
    shared_segments = [ multiprocessing.sharedctypes.Array(typ, map(lambda x: x[i], sql_data, lock=False) ) for (i,typ) in enumerate(type_string) ]
        
## FACTOR EXECUTE HACK
def main_memory_factor_executor(js, ss, nParts=12):
    global row_data, col_data, rating_data, grad_fn, bulk_fn, loss_fn, perm
    mid      = js.mid


    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    t0       = time.time()
    cur.execute(js.data_query)

    sql_data    = cur.fetchall()
    nItems   = len(sql_data)
    
    row_data    = multiprocessing.sharedctypes.Array('i', [row-1 for (row,col,rating,) in sql_data], lock=False)
    col_data    = multiprocessing.sharedctypes.Array('i', [col-1 for (row,col,rating,) in sql_data], lock=False)
    rating_data = multiprocessing.sharedctypes.Array('f', [rating for (row,col,rating,) in sql_data], lock=False)

    
    nRows       = max(row_data) + 1 
    nCols       = max(col_data) + 1
    nSplits     = nParts

    round_map   = multiprocessing.sharedctypes.Array('i', range(nItems))
    chunk_map   = multiprocessing.sharedctypes.Array('i', range(nItems))
    item_map    = multiprocessing.sharedctypes.Array('i', range(nItems))
    print "Data fetched in {}".format(time.time() - t0)

    state    = dict()

    perm     = multiprocessing.Array('i', range(nItems), lock=False)
    
    # Now ground teh functions
    model    = mm_se.ground_function(js.init_function)()
    grad_fn  = mm_se.ground_function(js.grad_function)
    loss_fn  = mm_se.ground_function(js.loss_function)
    stop_fn  = mm_se.ground_function(js.stop_condition)
    part_fn  = mm_se.ground_function(js.partition_function)
    bulk_fn  = None
    if js.bulk_grad is not None:
        bulk_fn = mm_se.ground_function(js.bulk_grad)

    shared_storage.store_cache_model(mid, model, js.model_type)

    #####################################################################
    # To see the shared data properly, we need to setup the pool after  #
    # allocating the shared data                                        #
    #####################################################################
    
    pool     = Pool(processes=nParts)
    (model,model_types) = shared_storage.get_cache_model(mid) 
    del sql_data
    epoch = 1
    # these do not change

    target_fn = execute_grad
    if bulk_fn is not None:
        print ">>> USING BULK"
        target_fn = _bulk_execute_grad
    else:
        print ">>> NO BULK"

    # This factoring code does not work correctly, it is 
    row_perm     = multiprocessing.sharedctypes.Array('i', range(nRows))
    col_perm     = multiprocessing.sharedctypes.Array('i', range(nCols))
    loss_jobs    = [ (mid, nItems, nParts, part_id) for part_id in range(nParts)]
    loss         = multiprocessing.Value('f', 1e100,lock=False)
    rounds       = nSplits
    rounds       = 1 
    chunks       = nSplits

    while not stop_fn(state, loss):
        t0 = time.time()
        random.shuffle(perm)
        random.shuffle(row_perm)
        random.shuffle(col_perm)
        print "Done with Shuffle {}".format(time.time() - t0)

        last_index = [ array.array('i', [-1 for c in range(chunks)]) for r in range(rounds) ]
        
        for i in xrange(nItems):
            pi  = perm[i]
            r0  = row_data[pi]
            r   = row_perm[r0]
            
            c0  = col_data[pi]
            c   = col_perm[c0]
            row_bucket = min(int(r/(nRows/rounds)),rounds-1)
            col_bucket = min(int(c/(nCols/chunks)),chunks-1)
            
            round_map[pi] = row_bucket
            chunk_map[pi] = col_bucket
            item_map[pi]  = last_index[row_bucket][col_bucket]
            last_index[row_bucket][col_bucket] = pi
            
        print "Done with Partition {}".format(time.time() - t0)
        
        t1 = time.time()
                
        for r in xrange(rounds):                
            cs  = [ (r,(r+c) % chunks) for c in xrange(chunks) ]
            ps  = map(lambda (r,c) : multiprocessing.Process(target=target_fn, args=(mid,nItems, r,c, round_map, chunk_map, item_map, last_index[r][c])), cs)
            tmp = [ p.start() for p in ps ]
            ts  = [ p.join()  for p in ps ]
            print "\t Round complete {} of {}".format(r+1, rounds)            
        print "Done with grad step {}".format(time.time() - t1)

        tmp_loss    = 0.0
        losses  = pool.map(compute_loss, loss_jobs)
        tmp_loss    = sum(losses)
        loss.value  = tmp_loss
        print "Done loss={} time={} total_loss={}".format(math.sqrt(tmp_loss)/math.sqrt(nItems), time.time() - t0, tmp_loss)
        
        epoch += 1
        ss.update_segment("epoch", epoch)


    print "Storing back model {}".format(js.mid)
    store_model_methods.store_model(js.mid,model)

        
