#! /usr/bin/env python
import sys
import traceback
import antlr3
import antlr3.tree
from src.parser.VictorParserLexer import VictorParserLexer
from src.parser.VictorParserParser import VictorParserParser
import src.globals.constants as constants
import src.parser.ParserTypes as ParserTypes
import src.vsql_core.plan_generators.plan_generator as planner

import src.normalization.normalization_front as normalize
import src.normalization.create_model as create_model
import src.normalization.create_model_helper as create_model_helper

import collections, multiprocessing, multiprocessing.sharedctypes
import cPickle, cStringIO
import src.utils.shared_storage as shared_storage

from multiprocessing import Pool
import src.globals.global_utils as global_utils, time, random, math, array
import main_memory_simple_exec as mm_se
import main_memory_factor_exec 
import numpy

JOB_SPEC = collections.namedtuple('job_spec', 'mid factor_model init_function stop_condition grad_function bulk_grad loss_function partition_function data_query model_type data_type')



def fetch_model_information(model_name):
    grad_loss_str = """SELECT grad_function_str, loss_function_str, model_field_types, data_field_types, bulk_grad
    FROM victor_model_table where model_name = '{}'""".format(model_name)
    print grad_loss_str
    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    cur.execute(grad_loss_str)
    (grad_str, loss_str, model_str, data_str, bulk_str) = cur.fetchone()
    grad = ParserTypes.PythonFunction(grad_str.split('.'))
    loss = ParserTypes.PythonFunction(loss_str.split('.'))
    bulk = None
    if bulk_str is not None:
        bulk = ParserTypes.PythonFunction(bulk_str.split('.'))
    model_type = tuple(model_str)
    data_type  = tuple(data_str)
    
    print "\t [main_mem_normalize.py] model_type={} data_type={} bulk_type={} ({})".format(model_type, data_type, bulk, bulk_str)
    conn.close()

    return (grad,loss, model_type, data_type, bulk)

def main_memory_create_model_instance(param_dict):
    model_instance_name      = param_dict[constants.PARSER_NAME_FIELD]
    example_table_name       = param_dict[constants.PARSER_EX_FIELD]
    example_fields           = param_dict[constants.PARSER_EX_TYPE_LIST]
    model_name               = param_dict[constants.PARSER_MODEL_SPEC_FIELD]
    init_fnc                 = param_dict[constants.PARSER_INIT_BLOCK]
    stop_condition           = param_dict[constants.PARSER_STOP_CONDITION_NAME]
    partition_function       = create_model.optional_param(param_dict, constants.PARSER_PARTITION_FUNCTION_NAME)

    print example_table_name, example_fields
    q_str = "SELECT {field_names} FROM {example_table_name}".format(field_names=','.join(example_fields),
                                                                  example_table_name=example_table_name)
    print q_str
    factor_model = partition_function is not None
    (grad,loss,model_type, data_type, bulk_grad)=fetch_model_information(model_name)

    mid = create_model_helper.get_model_mid(model_instance_name, model_name)

    job_spec = JOB_SPEC(mid=mid,
                        factor_model=factor_model,
                        init_function=init_fnc, 
                        stop_condition=stop_condition, 
                        grad_function=grad,
                        bulk_grad=bulk_grad,
                        loss_function=loss,
                        partition_function=partition_function,
                        data_query=q_str, 
                        model_type=model_type,
                        data_type=data_type)
    return job_spec

def is_model_spec(t):
    return isinstance(t, ParserTypes.ParserModelInstance)

##########################################
#
# the execture should be in its own module
import src.utils.shared_storage as shared_storage
def main_mem_executor(job_spec, planned_work=[]):
    conn                    = global_utils.connect_db()
    cur                     = conn.cursor()

    ss = shared_storage.start_server()
    ss.create_segment("epoch", 0)
    for pw in planned_work:    
        print "Executing {}".format(pw.setup_query)
        cur.execute(pw.setup_query)

    conn.commit()
    conn.close()
    
    if job_spec.factor_model:
        print "[Main Memory] Main Memory Factor Executor (WARNING: Only Sparse Matrix Supported)"
        main_memory_factor_exec.main_memory_factor_executor(job_spec, ss)
    else:
        print "[Main Memory] Simple Main Memory Executor"
        mm_se.main_memory_line_executor(job_spec, ss)
        

def test_handler(fname):
    print "Parsing {0}...".format(fname)
    tasks = parse_phase(fname)
    good_tasks = filter(is_model_spec,tasks)
    create_tasks = filter(lambda x: not (is_model_spec(x)),tasks)
    for ct in create_tasks:
        ct.normalize()
        
    jobs_specs = [main_memory_create_model_instance(g.params) for g in good_tasks]
    

# if len(sys.argv) == 2:
#     main(sys.argv[1])
# else:
#     print "{0} <specification_file>".format(sys.argv[0])
