from sys_appends import *

import src.globals.constants as constants
import src.globals.global_types as global_types
import src.globals.global_utils as global_utils
import src.globals.victor_exceptions as victor_exceptions
import ova_normalization

import sys, traceback
from src.normalization.create_model_helper import *
import psycopg2



def get_param_type_call_full_str(params):
    def types_to_psql(t):
        if t == constants.PYTHON_FIELD_TYPE:
            return "bytea"
        else:
            return t
    param_types = [ t for (p,t) in params ]
    param_names = [ p for (p,t) in params ]
    param_full  = [ (p + " " + types_to_psql(t)) for (p,t) in params]

    param_type_str = ','.join(param_types)
    param_call_str = ','.join(param_names)
    param_full_str = ','.join(param_full)
    return (param_type_str, param_call_str, param_full_str)

def unpickle_data_types_str(params, indent):
    indent_of_line = ''.join('\t' for i in range(indent))
    unpickle_str = ""
    bin_param_names = []
    for (p, t) in params:
        p_bin = "{0}_unpickled".format(p)
        unpickle_str += "{0}{1} = {2}.unpickle_data({3})\n".format(indent_of_line, p_bin, 
                                                                    constants.VICTOR_PICKLER, p)
        bin_param_names.append(p_bin)

    return (unpickle_str, ','.join(bin_param_names))


def unpickle_data_types(params):
    rets = []
    for (attr_name, typ) in params:
        if typ == constants.PYTHON_FIELD_TYPE:
            rets.append("{victor_pickler}.unpickle_data({attr_name})".format(victor_pickler=constants.VICTOR_PICKLER,
                                                                            attr_name=attr_name))
        else:
            rets.append("{attr_name}".format(attr_name=attr_name))
    return ",".join(rets)


#Assumes path is under PYTHONPATH
#NB: You can check whether modules are available
#But, this might not be a good idea, because 
#functions such as library setup add stuff to the sys.path
#in that case, we screw if we give exception here.
def create_the_wrapper_fnc(curry_fname, fnc, params, grad_fnc, data_type):
    (param_type_str, disused, param_full_str) = get_param_type_call_full_str(params)

    import_str = fnc.get_import_statement()    
    import_str += ", {0}".format(constants.VICTOR_PICKLER) 

    param_call_str   = unpickle_data_types(params)
    
    if grad_fnc: 
        ret_type = "VOID"
        footer = "GD[{0}] = {1}(model, ({2},))".format(constants.MID_FIELD,fnc.get_qualified_name(), param_call_str)
    else:
        ret_type = "FLOAT8"
        footer = "return {0}(model, ({1},))".format(fnc.get_qualified_name(), param_call_str)

    f = """DROP FUNCTION IF EXISTS {function_name} (int, {typed_params}) CASCADE;
    CREATE FUNCTION {function_name} ({mid_field} int, {typed_params})
    \tRETURNS {ret_type}
    AS $$
    \t{import_statement}
    \tmodel = GD[mid]
    \t{footer}
    $$ LANGUAGE plpythonu;""".format(function_name=curry_fname,
                                     typed_params=param_full_str,
                                     mid_field=constants.MID_FIELD, 
                                     ret_type =ret_type,
                                     import_statement=import_str,
                                     footer=footer)
    # print f
    return f

##Creates model
def create_model(param_dict):
    ##TODO: CLEAN THIS PART!!!
    model_name              = param_dict[constants.PARSER_NAME_FIELD]
    model_type              = "DISUSED" # CLEAN ME!
    model_field_types       = param_dict[constants.PARSER_MODEL_TYPE_FIELDS]
    model_field_names       = param_dict[constants.PARSER_MODEL_TYPE_NAMES]
    data_item_type          = "DISUSED"  # CLEAN ME! 
    data_item_field_types   = param_dict[constants.PARSER_DATA_TYPE_FIELDS]
    data_item_field_names   = param_dict[constants.PARSER_DATA_TYPE_NAMES]
    objective_fnc           = param_dict[constants.PARSER_OBJ_FIELD]
    objective_agg           = param_dict[constants.PARSER_OBJ_AGG_FIELD]
    grad_step               = param_dict[constants.PARSER_GRAD_FIELD]
    bulk_grad               = optional_param(param_dict,constants.PARSER_BULKGRAD)
    
    conn                    = global_utils.connect_db()
    cur                     = conn.cursor()


    ###################################################################################################
    ### CHECK INPUTS ARE VALID
    ###################################################################################################
    #Test whether model_name already exists or not
    model_table_name = convert_model_name_to_model_table_name(model_name)
    if check_table_exists_in_db(cur, model_table_name):
        raise victor_exceptions.VictorException("model {0} already exists".format(model_table_name))
    
    #Test whether model types already exist in db
    for t in model_field_types:
        if not check_type_exists_in_db(cur, t):
            raise victor_exceptions.VictorException("model type {0} doesn't exist".format(t))

    #Test whether data types already exist in db
    for t in data_item_field_types:
        if not check_type_exists_in_db(cur, t):
            raise victor_exceptions.VictorException("data type {0} doesn't exist".format(t))

    ###################################################################################################

    #NOW CREATE THE WRAPPER FUNCTIONS
    #create params
    params = []
    for i in range(len(data_item_field_names)):
        params.append((data_item_field_names[i], normalize_type(data_item_field_types[i])))
    #Create grad wrapper
    grad_wrapper_fnc_name = create_fnc_name(model_name, grad_step.function_name)
    cur.execute(create_the_wrapper_fnc(grad_wrapper_fnc_name, grad_step, params, True, data_item_type))

    #Create obj wrapper
    obj_wrapper_fnc_name = create_fnc_name(model_name, objective_fnc.function_name)
    cur.execute(create_the_wrapper_fnc(obj_wrapper_fnc_name, objective_fnc, params, False, data_item_type))

    # Create the raw calls
    grad_function_string      = grad_step.get_serialization_string()
    loss_function_string      = objective_fnc.get_serialization_string()
    
    #NB: Assumes that schema of VMT:
    fields = [model_name, model_type, model_table_name, model_field_types, model_field_names, data_item_field_types, 
              data_item_field_names, obj_wrapper_fnc_name, objective_agg, grad_wrapper_fnc_name,
              grad_function_string, loss_function_string]
    if bulk_grad is not None:
        fields.append(bulk_grad.get_serialization_string())
  
    strs_of_fields = ','.join(convert_text_to_sql(v) for v in fields)
    insert_to_victor_model_table_query = "INSERT INTO {} VALUES ({})".format(constants.VICTOR_MODEL_TABLE, strs_of_fields)

    #inserts new model to victor_model_table
    try:
        cur.execute(insert_to_victor_model_table_query)
    except psycopg2.DatabaseError as a:
        traceback.print_exc(file=sys.stdout)
        print 'error on query: ' + insert_to_victor_model_table_query
        exit(1)

    # create_model_table_query_strs = ["CREATE TABLE ", model_table_name, "(", constants.MID_FIELD, 
    #                                  " int, init_fnc text, train_fnc text, ", 
    #                                  convert_create_statements(model_field_types, model_field_names), ");"]
    create_model_table_query      = "CREATE TABLE {model_table_name} ({mid_field} int, init_fnc text, train_fnc text, {model_field_info});".format(model_table_name=model_table_name, mid_field=constants.MID_FIELD,
                                                                                                                                        model_field_info=convert_create_statements(model_field_types, model_field_names))


    #creates new model table
    try:
        cur.execute(create_model_table_query)
    except psycopg2.DatabaseError as a:
        traceback.print_exc(file=sys.stdout)
        print 'error on query: ' + create_model_table_query
        exit(1)

    conn.commit()
    print model_name + ' is successfully created'
    return 0

def optional_param(pd, c):
    if pd.has_key(c):
        return pd[c]
    else:
        return None

##Creates model instance
def create_model_instance(param_dict):
    model_instance_name      = param_dict[constants.PARSER_NAME_FIELD]
    example_table_name       = param_dict[constants.PARSER_EX_FIELD]
    example_fields           = param_dict[constants.PARSER_EX_TYPE_LIST]
    model_name               = param_dict[constants.PARSER_MODEL_SPEC_FIELD]
    init_fnc                 = param_dict[constants.PARSER_INIT_BLOCK]
    stop_condition           = param_dict[constants.PARSER_STOP_CONDITION_NAME]
    partition_function       = optional_param(param_dict, constants.PARSER_PARTITION_FUNCTION_NAME)
    

    conn = global_utils.connect_db()
    cur = conn.cursor()

    
    model_table_name = convert_model_name_to_model_table_name(model_name)
    #model exists
    if not check_table_exists_in_db(cur, model_table_name):
        raise victor_exceptions.VictorException(model_name + " does not exist")

    #model exists
    if not check_table_exists_in_db(cur, example_table_name):
        raise victor_exceptions.VictorException(example_table_name + " does not exist")

    ova_tuple = ova_normalization.build_ova_structures(cur, model_instance_name, param_dict)

    #get data types in victor_model_table
    query = "SELECT " + constants.VMT_DATA_FIELD_TYPES + " from " + constants.VICTOR_MODEL_TABLE
    query += " WHERE " + constants.VMT_MODEL_NAME + " = '" + model_name + "'"

    field_types_in_victor_model_table = get_single_tuple_result(cur, query)

    #type of example table matches with model declaration
    assert(len(field_types_in_victor_model_table) == len(example_fields))
    for i in range(len(example_fields)):
        field_type = get_type_of_field_in_table(cur, example_fields[i], example_table_name)
        if not compare_two_fields(field_type, field_types_in_victor_model_table[i]):
            warn_str = "[create model instance] the attribute {2} does not match with its supposed type in the victor_model_table: {0} v. {1}".format(field_type, field_types_in_victor_model_table[i], example_fields[i])
            print "WARNING {0}".format(warn_str)
            # raise victor_exceptions.VictorException(warn_str)


    mid = get_model_mid(model_instance_name, model_name)
    select_info_query = """SELECT {0}, {1}, {2}, {3}, {4}, {5}, {6}
    FROM {7} WHERE {8} = '{9}';""".format(constants.VMT_MODEL_TYPE, constants.VMT_MODEL_FIELD_TYPES, 
                                          constants.VMT_MODEL_FIELD_NAMES, constants.VMT_DATA_FIELD_NAMES, 
                                          constants.VMT_OBJECTIVE, constants.VMT_OBJECTIVE_ITEM, 
                                          constants.VMT_WRAPPER_FNC, constants.VICTOR_MODEL_TABLE, 
                                          constants.VMT_MODEL_NAME, model_name)
    cur.execute(select_info_query)
    (model_type, model_field_types, model_field_names,
     useless_example_names, objective, objective_agg, grad_function_name, ) = cur.fetchone()
    # the example names from here are just for the functions...
    # Build the initailization function name
    init_fnc_name = create_fnc_name(model_instance_name, "init")
    init_model_fnc_query = create_init_fnc(init_fnc_name, init_fnc, model_field_types)    
    cur.execute(init_model_fnc_query)
    conn.commit()


    # WARNING: This mid must be late bound!!!
    #init_model_fnc_with_params = "{0}({1})".format(init_fnc_name, mid)    
        



    model_field_names = ','.join(model_field_names)

    # Create the model instance and example table descriptions
    model_table_desc   = global_types.TableDescription(table_name=model_table_name  , field_names=model_field_names)
    example_table_desc = global_types.TableOrViewDescription(declaration=example_table_name, field_names=example_fields)
    # Create a description of the refinement step
    # NB: stop_condition should be converted into a SQL like function
    refinement_desc    = global_types.RefinementDescription(initialization_function=init_fnc_name,
                                                            gradient_step=grad_function_name,
                                                            objective_item=objective,
                                                            objective_aggregate=objective_agg,
                                                            stop_function=stop_condition,
                                                            partition_function=partition_function)

    connect_str = "dbname={0}".format(dbname)

    # TODO: parse these (really logical optimization and normalization) FACTOR THIS OUT!
    optim_desc  = global_types.OptimizationDescription(batch_gradient=False, ova=ova_tuple, factored_execute=False)
    
    model_desc  = global_types.ModelInstanceDescription(mid                = mid,
                                                        connect_str        = connect_str,
                                                        model_table_desc   = model_table_desc,
                                                        example_table_desc = example_table_desc,
                                                        refinement_desc    = refinement_desc,
                                                        opts               = optim_desc)
    return model_desc
