import src.globals.constants as constants
import src.globals.global_utils as global_utils

import psycopg2

def convert_modules_to_import_strs(modules):
    imports = ""
    for i in modules:
        imports = "{0}import {1}\n".format(imports, i)    

    return imports

def convert_statements_to_str(exec_stmts):
    exec_str = ""
    for i in exec_stmts:
        cmd = i[1]
        exec_str = "{0}{1}\n".format(exec_str, cmd)
    return exec_str

#TODO: I couldnt find a way to do this with python classes.
#Search more
# def get_module_name(module_with_fnc):
#     return module_with_fnc.split('.')[0]

# def get_fnc_name(module_with_fnc):
#     return module_with_fnc.split('.')[1]

def exists_in_db(cur, table_name, field_name, field_value):
    
    check_query = """SELECT COUNT(*) as cnt FROM {0} WHERE {1} = '{2}'
    """.format(table_name, field_name, field_value)
    cur.execute(check_query)
    (count,) = cur.fetchone()
    if count:
        return True
    else:
        return False

def check_table_exists_in_db(cur, table_name):
    return exists_in_db(cur, constants.PG_CLASS_TABLE, constants.PG_CLASS_RELNAME, table_name)

def check_type_exists_in_db(cur, type_name):
    #If it is an array type, we should convert x[] to _x, since
    #this is how x[] is stored in pg_type
    #TODO: Do we have more cases like this?
    type_name = normalize_type(type_name)
    return exists_in_db(cur, constants.PG_TYPE_TABLE, constants.PG_TYPE_TYPNAME, type_name)


#fetchs one tuple from the cursor
#and return it
#Assumes that query returns one field
def get_single_tuple_result(cur, query):
    cur.execute(query)
    (ret,) = cur.fetchone()
    return ret


def get_model_mid(model_instance_name, model_name):
    #get unique mid
    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    
    insert_query = "INSERT INTO {model_id_table}({instance_name}, {model_spec_field}) VALUES('{model_instance_name}', '{model_spec_name}');".format(
        model_id_table      = constants.VICTOR_GLOBAL_MODEL_ID_TABLE,
        instance_name       = constants.VICTOR_GLOBAL_INSTANCE,
        model_spec_field    = constants.VICTOR_GLOBAL_MODEL_SPEC_NAME,
        model_instance_name = model_instance_name,
        model_spec_name     = model_name)
    
    cur.execute(insert_query)
    select_query = "SELECT {0} FROM {1} WHERE {2} = '{3}';".format(constants.VICTOR_GLOBAL_MID, 
                                                                   constants.VICTOR_GLOBAL_MODEL_ID_TABLE,
                                                                   constants.VICTOR_GLOBAL_INSTANCE,
                                                                   model_instance_name)
    mid = get_single_tuple_result(cur, select_query)
    conn.close()
    return mid

def normalize_type(type_name):
    type_name = type_name.lower()
    if type_name.find("[]") >= 0:
        stripped_str = type_name.replace("[]", "").rstrip()
        type_name = "_" + stripped_str

    # THIS IS A HACK. you want to detect the int and replace it with
    # any leading number of underscroes.
    if type_name == constants.PYTHON_FIELD_TYPE or type_name == constants.NPYTHON_FIELD_TYPE:
        type_name = "bytea"
    if type_name == constants.NUMPY_FIELD_TYPE:
        type_name = "bytea"
    elif type_name == "int":
        type_name = "int4"
    elif type_name == "_int":
        type_name = "_int4"
    return type_name

#NB: Order is important, because for array types field1 is like _float8 whereas
#field2 is float8[]
def compare_two_fields(field1, field2):
    field1 = normalize_type(field1)
    field2 = normalize_type(field2)
    if field1 == field2 or (field1 == "int4" and field2 == "int"):
        return True
    else:
        return False

##Returns the type of the field from the given field name and 
##the table name
def get_type_of_field_in_table(cur, field_name, table_name):
    query = """SELECT {0} FROM {1} p1, {2} p2 WHERE p1.{3}='{4}' AND p1.{5} = p2.{6} AND 
    {7} IN (SELECT {8} FROM {2} WHERE {9} = 'c' and ({10} = '{11}'));
    """.format(constants.PG_TYPE_TYPNAME, constants.PG_ATTRIBUTE_TABLE, constants.PG_TYPE_TABLE, 
               constants.PG_ATTRIBUTE_ATTNAME, field_name, constants.PG_ATTRIBUTE_ATTTYPID, 
               constants.PG_OID, constants.PG_ATTRIBUTE_ATTRELID, constants.PG_TYPE_TYPRELID, 
               constants.PG_TYPE_TYPTYPE, constants.PG_TYPE_TYPNAME, table_name)
    print query
    return get_single_tuple_result(cur, query)

##Converts non-array text to sql type
def convert_type_to_sql_text(input_type):
    return '\'' + input_type + '\''

##Converts text array to sql type
def convert_types_to_sql_text_array(input_types):
    return '\'{' + ','.join(input_types) + '}\''

##General function that converts either text array or non-array
##text into sql type
def convert_text_to_sql(input_type):
    if type(input_type).__name__ == 'str':
        return convert_type_to_sql_text(input_type)
    else:
        return convert_types_to_sql_text_array(input_type)

##Converts given model name to the model table name
##convention is: model_table_name = model_name + "_instance"
def convert_model_name_to_model_table_name(model_name):
    return model_name + '_instance'


def create_fnc_name(model_name, grad_fnc_name):
    return "{0}_{1}".format(model_name, grad_fnc_name)

##Returns the create statement segment for the fields
## from given field types and field names
## e.g. field_types = [int, float8[]] field_names = [label, vec]
##Returns: label int, vec float8[]
def convert_create_statements(field_types, field_names):
    def types_to_psql(t):
        if t == constants.PYTHON_FIELD_TYPE:
            return "bytea"
        elif t == constants.NUMPY_FIELD_TYPE:
            return "bytea"
        else:
            return t
    assert(len(field_types) == len(field_names))
    return ','.join(field_names[i] + ' ' + types_to_psql(field_types[i]) for i in range(len(field_types)))




## TODO: Move this to a util directory
def strip_braces(s):
    s = s.lstrip(' {')
    return s.rstrip(' }')



def create_init_fnc(init_fnc_name, init_fnc, model_field_types):
    return_type = ""
    return_val = ""    
    import_str = init_fnc.get_import_statement()
    # NB: The name of mid_types should be a function!!
    
    f = """DROP FUNCTION IF EXISTS {init_fn_name} (mid int) CASCADE;
    CREATE FUNCTION {init_fn_name} (mid int)
    \tRETURNS VOID
    AS $$
    \t{import_statement}
    \tmid_types = \"victor_types_\" + str(mid)
    \tGD[mid_types] = tuple({model_field_types})
    \tGD[mid] = {initialization_function}()
    $$ LANGUAGE plpythonu;""".format(init_fn_name=init_fnc_name, import_statement=import_str, 
                                     initialization_function=init_fnc.get_qualified_name(), model_field_types=model_field_types)

    print f
    return f
