grammar VictorParser;

options {
    language=Python;
    output=AST;
    ASTLabelType=CommonTree;
}

@header { 
import src.globals.constants  as constants
import src.parser.ParserTypes as ParserTypes
}

top_level returns [value] :
        a = victor_statement {$value = [a.value]} ';' (a = victor_statement {$value.append(a.value)} ';' )* ;

victor_statement returns [value]: 
        z = create_model_instance      { $value = z.value }
    |   z = create_model_specification { $value = z.value }
    |   z = delete_model_specification { $value = z.value }
    |   z = create_model_application   { $value = z.value } 
    ;

create_model_instance returns [value]: 
        'CREATE MODEL INSTANCE' n=id z = inst_decls 
        {   name   = n.value
            params = [(constants.PARSER_NAME_FIELD, name)] + z.value
            $value = ParserTypes.ParserModelInstance(name=name,params=dict(params))
        }; 

//create model speficiation
create_model_specification returns [value]: 'CREATE MODEL SPECIFICATION' n=id '(' z=decls ')'
            {   name   = n.value
                params = [(constants.PARSER_NAME_FIELD, name)] + z.value
                $value = ParserTypes.ParserModelSpecification(name=name, params=dict(params)) 
            } ;

delete_model_specification returns [value]: 
        'DELETE MODEL SPECIFICATION' n=id 
        {
            name = n.value
            $value  = ParserTypes.ParserDeleteModelSpecification(name=name)
        };

create_model_application returns [value] : 
        'CREATE MODEL APPLICATION FUNCTION' n=id 'function_name' '=' pname=qualified_name '(' tps=typed_param_list ')' 'RETURNS' ret_type=id 'model_instance' '=' mi=id
        { 
            name = n.value
            params = [(constants.PARSER_NAME_FIELD, name), (constants.PARSER_NAME_FIELD, n.value), 
                      (constants.PARSER_MODEL_APP_PYTHON_FNC, ParserTypes.PythonFunction(pname.value)), 
                      (constants.PARSER_MODEL_INST_NAME,mi.value),
                      (constants.PARSER_DATA_TYPE_FIELDS,tps.value),
                      (constants.PARSER_MODEL_APP_RETURN_TYPE, ret_type.value)]
            $value = ParserTypes.ParserModelAppSpecification(name=name, params=dict(params)) 
        };


inst_decls returns [value] : 
    // ',' between inst_decl? not consistent with create_model
    a=inst_decl { $value = a.value } ( b=inst_decl { $value.extend(b.value) } )* ;

inst_decl returns [value]: 
      EXAMPLES   z1=id '(' z=idlist ')' { $value = [(constants.PARSER_EX_FIELD, z1.value), (constants.PARSER_EX_TYPE_LIST, z.value)] } 
    | MODEL SPEC z=id                 { $value = [(constants.PARSER_MODEL_SPEC_FIELD, z.value)] }
    | STOP WHEN  f=qualified_name     { $value = [(constants.PARSER_STOP_CONDITION_NAME, ParserTypes.PythonFunction(f.value))] }
    | PARTITION  f=function_name      { $value = [(constants.PARSER_PARTITION_FUNCTION_NAME, f.value)] }
    | OVA        '(' f=id ',' t=id ',' v=UNINTERP_BLOCK ')'   { $value = [(constants.OVA, True), (constants.PARSER_LABEL_ATTR, f.value), (constants.PARSER_LABEL_TYPE, t.value), (constants.PARSER_LABEL_SET_TABLE, v.getText().encode())] }
    | INIT_FUNCTION f=qualified_name  { $value = [(constants.PARSER_INIT_BLOCK, ParserTypes.PythonFunction(f.value))] }
    ;

decls returns [value] : 
    a=decl { $value = a.value } (',' b=decl { $value.extend(b.value) } )* ;

decl returns [value]: 
      MODEL_T '=' '(' d1=idlist ')' 'as' '(' d2=idlist ')' { $value = [(constants.PARSER_MODEL_TYPE_FIELDS, d1.value), (constants.PARSER_MODEL_TYPE_NAMES, d2.value)] } 
    | OBJECTIVE '=' f=qualified_name { $value = [(constants.PARSER_OBJ_FIELD, ParserTypes.PythonFunction(f.value))] }
    | DATA_T '=' '(' d1=idlist ')' 'as' '(' d2=idlist ')'  { $value = [(constants.PARSER_DATA_TYPE_FIELDS, d1.value), (constants.PARSER_DATA_TYPE_NAMES, d2.value)] }  
    | OBJECTIVE_AGG '=' z=id  { $value= [(constants.PARSER_OBJ_AGG_FIELD, z.value)] }
    | GRAD_STEP '=' f=qualified_name { $value = [(constants.PARSER_GRAD_FIELD, ParserTypes.PythonFunction(f.value))] }
    | BULKGRAD '=' f=qualified_name { $value = [(constants.PARSER_BULKGRAD, ParserTypes.PythonFunction(f.value))] } 
    ;

idlist returns [value] :
    v=id { $value = [v.value] } (',' u=id { $value.append(u.value)  } )*
;


typed_param_list returns [value]:
        p=id t=id { $value = [(p.value,t.value)] } (',' p2=id t2=id { $value.append((p2.value,t2.value))  } )*
    ;

qualified_name returns [value]:
   v=id { $value = [v.value] } ('.' u=id {$value.append(u.value) } )* 
    ;

psql_qualified_name returns [value]:
        'PSQL' ':' libpath=path ':' fname=id  { $value = ParserTypes.PSQLFunction(libpath.value,fname.value) }
        ;

function_name returns [value]:
        | f=psql_qualified_name { $value = f.value } 
        | f=qualified_name      { $value = ParserTypes.PythonFunction(f.value) }
        ;

/* Lexer Rules */
/* These are all the reserved words */
MODEL_T : 'model_type' ;
PYTHON_MODEL_T : 'python_model_type' ;
DATA_T  : 'data_item_type' ;
PYTHON_DATA_T : 'python_data_item_type';
OBJECTIVE : 'objective';
OBJECTIVE_AGG : 'objective_agg'; 
GRAD_STEP    : 'grad_step' ;
BULKGRAD : 'bulk_grad';

EXAMPLES : 'EXAMPLES' ;
MODEL    : 'MODEL';
SPEC     : 'SPEC';
STOP     : 'STOP';
WHEN     : 'WHEN';
PARTITION : 'PARTITION';
MODEL_PARAMS : 'MODEL_PARAMS';
INIT_FUNCTION : 'INIT_FUNCTION';
OVA      : 'OVA';

id     returns [value] : z=ID {$value=z.getText().encode()};
string returns [value] : z=STRING {$value=z.getText().encode()};
path   returns [value] : z=PATH {$value=z.getText().encode()};

STRING     :  ('"' (~'"')* '"')
           | ('\'' (~'\'')* '\'');
ID         : ('a'..'z'|'A'..'Z')('a'..'z'|'A'..'Z' | '0'..'9' | '_' | '[' | ']' )* ;
PATH       : '/' ('a'..'z'|'A'..'Z')('a'..'z'|'A'..'Z' | '0'..'9' | '_' | '[' | ']' | '/' | '.' )* ;
INT        : ('0'..'9')+;
FLOAT      : ('0'..'9')? '.' ('0'..'9')+;
WHITESPACE : ( '\t' | ' ' | '\r' | '\n'| '\u000C' )+ 	{ $channel = HIDDEN; } ;
COMMENT    : '--' (~'\n')* { $channel = HIDDEN; } ;

UNINTERP_BLOCK 
    : '{' ( UNINTERP_BLOCK | ~'}' )* '}'
    ;
