from sys_appends import *
import psycopg2
import src.globals.constants    as constants
import src.globals.global_utils as global_utils
import src.globals.victor_exceptions as victor_exceptions
import src.vsql_core.utils.pickler.victor_pickler as victor_pickler
#Imported for mid_model_string
import src.utils.shared_storage as shared_storage
#Imported for escape_type method. This should be in util?
import src.normalization.ova_normalization as ova

#TODO List: ALTER TABLE victor_global_instances ADD COLUMN model_name text;


import collections
ModelInfo = collections.namedtuple('ModelInfo', 'mid table_name model_type typed_model_fields')
def get_model_info_from_mid(cur, mid):
    select_query = """
    SELECT m.{vmt_model_table_name}, m.{vmt_model_type}, 
    m.{vmt_model_field_types}, m.{vmt_model_field_names} FROM {vmt} m, {vgi} g 
    WHERE m.{vmt_model_name} = g.{vgi_model_name} and g.{mid_field} = {mid};
    """.format(vmt_model_table_name       = constants.VMT_MODEL_TABLE_NAME,
               vmt_model_type             = constants.VMT_MODEL_TYPE,
               vmt_model_field_types      = constants.VMT_MODEL_FIELD_TYPES,
               vmt_model_field_names      = constants.VMT_MODEL_FIELD_NAMES,
               vmt                        = constants.VICTOR_MODEL_TABLE,
               vgi                        = constants.VICTOR_GLOBAL_MODEL_ID_TABLE,
               vmt_model_name             = constants.VMT_MODEL_NAME,
               vgi_model_name             = constants.VICTOR_GLOBAL_MODEL_SPEC_NAME,
               mid_field                  = constants.MID_FIELD,
               mid                        = mid)
    cur.execute(select_query)
    if cur.rowcount == 1:
        (table_name, victor_model_type, field_types, field_names, ) = cur.fetchone()
        assert(len(field_types) == len(field_names))
        model_field_types = [ (field_types[i], field_names[i]) for i in range(len(field_types)) ]
        
        return ModelInfo(table_name         = table_name,
                         mid                = mid,
                         model_type         = victor_model_type,
                         typed_model_fields = model_field_types)
    else:
        return None

def retrieve_model(mid):
    conn = global_utils.connect_db()
    cur  = conn.cursor()

    model_info = get_model_info_from_mid(cur, mid)

    if model_info is not None:
        model_field_names_str  = ','.join( [ name for (type,name) in model_info.typed_model_fields ] )
        select_model_query     = """
        SELECT {model_call_params} 
        FROM {model_inst_table} WHERE {mid_field} = {mid}""".format(model_call_params = model_field_names_str,
                                                                    model_inst_table  = model_info.table_name,
                                                                    mid_field         = constants.MID_FIELD,
                                                                    mid               = model_info.mid)
        cur.execute(select_model_query)
        if (cur.rowcount == 1):
            tup = cur.fetchone()
            ret = []
            for (i, (typ,name)) in enumerate(model_info.typed_model_fields):
                if typ == constants.PYTHON_FIELD_TYPE or typ == constants.NUMPY_FIELD_TYPE:
                    ret.append(victor_pickler.unpickle_data(tup[i]))
                else:
                    ret.append(tup[i])
            if len(ret) == 1:
                return ret[0]
            else:
                return tuple(ret)
        else:
            #TODO: What to do here?
            raise victor_exceptions.VictorException('An error occurred in retrieve_model. Model instance does not exist.')
    else:
        #TODO: What to do here?
        raise victor_exceptions.VictorException('An error occurred in retrieve_model. Either mid or model for that mid does not exist.')
    
## Model must be stored as a list, not a tuple
def store_model(mid, in_model):
    conn = global_utils.connect_db()
    cur  = conn.cursor()

    if isinstance(in_model, tuple):
        model = list(in_model)
    else:
        # assume it is a singleton!
        model = [in_model]
        
    model_info = get_model_info_from_mid(cur, mid)
    
    if model_info is not None:
        # We build the model for return
        update_strs = []
        for (i, (typ,name)) in enumerate(model_info.typed_model_fields):
            if typ == constants.PYTHON_FIELD_TYPE or typ == constants.NUMPY_FIELD_TYPE:
                update_strs.append("{0} = {1}".format(name, psycopg2.Binary(victor_pickler.pickle_data(model[i]))))
            elif typ == 'bytea':
                update_strs.append("{0} = {1}".format(name, psycopg2.Binary(model[i])))
            else:
                update_strs.append("{0} = {1}".format(name, ova.escape_type(model[i], typ)))
                
        update_str = ','.join(update_strs)
        
        update_query = "UPDATE {instance_table} SET {update_str} WHERE {mid_field} = {mid}".format(instance_table = model_info.table_name,
                                                                                                   update_str = update_str,
                                                                                                   mid_field = constants.MID_FIELD,
                                                                                                   mid = mid)
        cur.execute(update_query)
        conn.commit()
    

def build_retrieve_model_sql():
    statements="""
    from sys_appends import *
    import sys
    plpy.info(sys.path)
    plpy.info(sys.version)
    import src.utils.retrieve_and_store_methods as retrieve_store
    GD[mid] = retrieve_store.retrieve_model(mid)
    """
    retrieve_model_sql = shared_storage.mid_model_string.format(function_name="VictorRetrieveModel", statements=statements)
    return retrieve_model_sql

def build_store_model_sql():
    statements="""
    import src.utils.retrieve_and_store_methods as retrieve_store
    assert(GD.has_key(mid))
    retrieve_store.store_model(mid, GD[mid])
    """
    store_model_sql = shared_storage.mid_model_string.format(function_name="VictorStoreModel", statements=statements)
    return store_model_sql
