from sys_appends import *
import src.globals.constants as constants
import src.globals.global_types as global_types
import pickle, collections, multiprocessing
from multiprocessing.managers import SyncManager
import ctypes, numpy, os, mmap


class MyManager(multiprocessing.managers.SyncManager):
    pass

# horrible hack
shared_numpy_segments = dict()
    
### Simple local class around dictionary
class cd:
    def __init__(self):
        self.shared_dict = dict()

    def add(self, x, v):
        self.shared_dict[x] = v
        
    def get(self, x):
        return self.shared_dict[x]
    
    def has_key(self, k):
        return self.shared_dict.has_key(k)

shared_dict = cd()
MyManager.register('get_shared_dict', lambda:shared_dict)

class StorageManager:
    import os
    port = 5001
    def __init__(self, server=False):
        self.segments     = []
        if server:
            self.server      = True                        
            self.manager     = MyManager(address=("", self.port), authkey="translate")
            print "Starting Server".format(os.getpid())
            self.manager.start()
            print "Server Started."            
            self.shared_dict = self.manager.get_shared_dict()
            # aref = self.manager.Array('f', range(10), lock=False)
            # self.shared_dict.add("yes", aref)
            
        else:
            self.server      = False
            self.manager     = MyManager(address=("localhost",self.port), authkey="translate")
            self.manager.connect()
            self.shared_dict = self.manager.get_shared_dict()
            # a = self.shared_dict.get("yes")
            # a[0] += 1.0

    def get_segment_name(self, name):
        return "/tmp/__segment_name_{0}.dat".format(name)
        
    ######################################################
    #
    # These methods create and store general shared objects
    #
    #######################################################
    def create_segment(self, name, value):
        seg_name  = self.get_segment_name(name)
        self.shared_dict.add(seg_name,value)
        
    def read_segment(self,name):
        seg_name = self.get_segment_name(name)
        return self.shared_dict.get(seg_name)
    
    def update_segment(self,name, value):
        seg_name = self.get_segment_name(name)
        self.shared_dict.add(seg_name, value)

    def contains_segment(self, name):
        seg_name  = self.get_segment_name(name)
        return self.shared_dict.has_key(seg_name)
    
    ########################################################
    #
    # This is the new interface for models with types
    #
    ########################################################
    def get_model_name(self, name):
        return "victor_model_{}".format(name)

    def get_model_type_name(self, name):
        return "victor_model_type_{}".format(name)
    
    # def store_model(self, name, model, model_types):
    #     self.create_segment(self.get_model_name(name),model)
    #     self.create_segment(self.get_model_type_name(name), model_types)

    def get_numpy_segment(self, name, i):
        return "victor_numpy_shared_{0}_victor_id_{1}".format(name, i)

    def store_model(self, name, model, model_types):
        # determine if its the first read (no record)
        bFirstRead = not self.contains_segment(self.get_model_type_name(name))

        # Now create the type portion
        self.create_segment(self.get_model_type_name(name), model_types)
        
        not_shared_parts = []
        shared_parts     = []
        for (i,t) in enumerate(model_types):
            if t == constants.NUMPY_FIELD_TYPE:
                shared_parts.append(i)
                not_shared_parts.append(None)
            else:
                not_shared_parts.append(model[i])                

        if bFirstRead:
             for i in shared_parts:
                self.set_numpy_array(self.get_numpy_segment(name, i), model[i])
        # do nothing for the shared parts, they are in shared memory
        # Store the rest back
        self.create_segment(self.get_model_name(name),tuple(not_shared_parts))

    def get_model_with_types(self, name):
        mt           = self.read_segment(self.get_model_type_name(name))
        shared_parts = filter(lambda (i,t): t == constants.NUMPY_FIELD_TYPE, enumerate(mt))

        # The model contains nulls for shared segments
        m            = self.read_segment(self.get_model_name(name))

        # fill in the nulls with their shared memory handlers
        if len(shared_parts) > 0:
            m_list = list(m)
            for (i,t) in shared_parts:
                
                m_list[i] = self.get_numpy_array(self.get_numpy_segment(name, i))
            m = tuple(m_list)
        return (m,mt)
    
    ########################################################
    #
    # These methods specialization to values and arrays
    #
    ########################################################
    def get_value_name(self, name):
        return "vvalue_{0}".format(name)

    def add_shared_value(self, name, typecode, value):
        v = self.manager.Value(typecode, value)
        self.shared_dict.add(self.get_value_name(name), v)

    def get_shared_value(self, name):
        return self.shared_dict.get(self.get_value_name(name))
        
    def get_shared_array(self, name):
        return self.shared_dict.get(self.get_value_name(name))

    ###################
    #
    # Numpy Array handlers
    #
    # Numpy arrays are handled with memory mapped files
    ##################    
    def set_numpy_array(self, name, numpy_array):
        # dt       = self.numpy_to_cmap[numpy_array.dtype]
        # NB: We need to do this with memory mapped files, because the processes need global segments.
        flat = numpy_array.flatten()
        dt   = flat.dtype
        with open("/tmp/" + name, "w+b") as f:
            f.write(flat.tostring())
        print "writing {}".format(dt)
        self.shared_dict.add(name, (dt, numpy_array.shape))
        
    def get_numpy_array(self, name):
        (dt, shape) = self.shared_dict.get(name)
        with open("/tmp/" +name, "r+b") as f:
            if f is None:
                raise ValueError("Tried to open {}".format(name))
            m = mmap.mmap(f.fileno(), 0)
            return numpy.reshape(numpy.frombuffer(m,dtype=dt), shape)
        
this_process_sm = None
def install_sm(sm):
    global this_process_sm
    this_process_sm = sm
    assert(this_process_sm is not None)
    
def get_sm():
    if this_process_sm is None:
        install_sm(StorageManager(server=False))
    return this_process_sm

### Front end for server installation
def start_server():
    sm = StorageManager(server=True)
    install_sm(sm)
    print "Server installed"
    return sm

### TODO: Move this to user facing utils
def get_epoch():
    return get_sm().read_segment("epoch")

###############################################
## Assumes only one process is caching a model


MID = collections.namedtuple('MIDModelDesc', 'function_name statements')
mid_model_string = """DROP FUNCTION IF EXISTS {function_name} (mid int) CASCADE;
    CREATE FUNCTION {function_name} (mid int)
    \tRETURNS VOID
    AS $$
    {statements}
    $$ LANGUAGE plpythonu;"""


def get_cache_model(mid):
    sm  = get_sm()
    return sm.get_model_with_types(mid)

def store_cache_model(mid, model, mid_types):
    sm  = get_sm()
    sm.store_model(mid, model, mid_types)

### To help store this processes storage_manager
def get_moniker(mid):
    return "victor_sm_{0}".format(mid)



def build_get_cache_model_sql():
    statements="""
    import src.utils.shared_storage as shared_storage    
    mid_types = \"victor_types_\" + str(mid)
    (GD[mid],GD[mid_types]) = shared_storage.get_cache_model(mid)
    """
    cache_model_sql = mid_model_string.format(function_name="VictorGetModelCache", statements=statements)
    return cache_model_sql

def build_store_cache_model_sql():
    statements="""
    import src.utils.shared_storage as shared_storage
    assert(GD.has_key(mid))
    mid_types = \"victor_types_\" + str(mid)
    assert(GD.has_key(mid_types))
    shared_storage.store_cache_model(mid, GD[mid], GD[mid_types])
    """
    cache_model_sql = mid_model_string.format(function_name="VictorStoreModelCache", statements=statements)
    return cache_model_sql

def test_handler():
    start_server()                                     
    store_cache_model(10, (range(10), numpy.matrix([[0.0 for i in range(10)] for j in range(3)])), tuple(["python", "numpy"]))
    print get_cache_model(10)

#test_handler()
