import sys_appends
import src.globals.constants as constants
import src.globals.global_types as global_types
import src.globals.global_utils as global_utils

import src.utils.shared_storage as shared_storage

import collections, time

import src.vsql_core.executors.worker_structure as worker_structure


#import src.vsql_core.utils.factored_parser.partition_data as partition_data
def get_split_name(output_name, id):
    return "{output_name}_split_{split_id}".format(output_name = output_name,
                                                   split_id    = id)

from multiprocessing import Pool
# for consistency with main memory shared inferface
class v_wrap:
    def __init__(self, v):
        self.value = v
        
##################################################
### These should really be moved to the planner (or global types)
ShuffleDesc     = collections.namedtuple('ShufflerDesc', 'output_table data_query')
WorkerDesc      = collections.namedtuple('WorkerDesc',   'mid example_table_attrs gradient_function data_query')
ScoreDesc       = collections.namedtuple('ScoreDesc',    'mid objective_agg objective_item example_table_attrs data_query')

FactorChunkStepDesc = collections.namedtuple('FactorChunkerStepDesc', 'seed data_query nSplits epoch mid partition_function')
#FactorChunkGradDesc = collections.namedtuple('FactorChunkGradDesc',   '')
######################
#### Worker Threads
######################
class SQLWorker:
    hard_worker = True
    def ground(self, qs, params, extra_params=None, debug=False):
        self.grounded_queries = map(lambda q: q.format(params, extra_params), qs)
        self.debug = debug
        
    def execute(self):
        conn                    = global_utils.connect_db()
        cur                     = conn.cursor()

        t0 = time.time()
        for q in self.grounded_queries:
            if self.debug:
                print "Executing: {}".format(q)
            cur.execute(q)
        if self.debug:
            print "time={0}".format(time.time() - t0)
        conn.commit()
        conn.close()

# Conventions
# Shufflers put the data into a fresh table
# Workers   clean up (drop) the table when they finish
# Scorers   operate on the original data.
###################
# Shuffler Stuff
setup_query    = "SELECT library_setup();";
class Shuffler(SQLWorker):
    drop_string     = "DROP TABLE IF EXISTS {0.output_table};"    
    shuffler_string = "SELECT * INTO {0.output_table} FROM ({0.data_query}) as x ORDER BY RANDOM ();"
    worker_queries  = [setup_query,drop_string, shuffler_string]
    def __init__(self, params):
        self.ground(self.worker_queries, params)
        
#####################
# Gradient Step Worker
class IncrementalGradientStep(SQLWorker):
    cache_model    = "SELECT VictorGetModelCache({0.mid});"
    grad_step      = "SELECT COUNT(*) FROM (SELECT {0.gradient_function}({0.mid}, {0.example_table_attrs}) FROM {0.data_query}) AS T;"
    store_model    = "SELECT VictorStoreModelCache({0.mid});"
    drop_string    = "DROP TABLE IF EXISTS {0.data_query};" 
    worker_queries = [setup_query,cache_model, grad_step, store_model, drop_string]
    def __init__(self, params):
        self.ground(self.worker_queries, params)

######################################################
## Scoring Worker: Broken Does not return the score.
class ScoreStep(SQLWorker):
    cache_model = "SELECT VictorGetModelCache({0.mid});"
    score_query = "SELECT {0.objective_agg}({0.objective_item}({0.mid},{0.example_table_attrs})) FROM ({0.data_query}) as x;"
    worker_queries = [setup_query,cache_model,score_query]
    def __init__(self, params, debug=False):
        self.params = params
        self.ground(self.worker_queries, params)
        self.debug  = debug
        
    def execute(self):
        conn                    = global_utils.connect_db()
        cur                     = conn.cursor()
        t  = time.time()
        for q in self.grounded_queries:
            if self.debug:
                print "Score: {}".format(q)
            cur.execute(q)
        (loss,) = cur.fetchone()
        t0 = time.time()
        if self.debug:
            print "[ScoreStep time={0} obj={1}]".format(t0-t,loss)
        return loss

## Factor Steps
class FactorChunkerStep(SQLWorker):
    build_view = """drop view if exists victor_view_{0.mid}_{0.seed}; create view victor_view_{0.mid}_{0.seed} as {0.data_query}; COMMIT; """
    cache_model = "SELECT VictorGetModelCache({0.mid});"
    # this guy runs in a separate process, so caching the model will not load it for him.
    build_partition_function = """    
    DROP FUNCTION IF EXISTS victor_partition_{0.mid}_{0.seed}(t victor_view_{0.mid}_{0.seed});
    CREATE FUNCTION victor_partition_{0.mid}_{0.seed}(t victor_view_{0.mid}_{0.seed}) RETURNS INT
    AS $$
    {1.import_string}
    import src.utils.shared_storage as shared_storage
    state_key = 'victor_partition_state_{0.mid}'
    if not GD.has_key(state_key):
       GD[state_key] = dict()
    mid = {0.mid}
    if not GD.has_key(mid):
       (GD[mid],x) = shared_storage.get_cache_model(mid)
    model     = GD[mid]
    return {1.py_partition_function}(GD[state_key], model, {0.seed}, {0.nSplits}, t)
    $$ LANGUAGE plpythonu; COMMIT; 
    """
    perm_and_slice_query = "SELECT partition_data_fnc('victor_partition_{0.mid}_{0.seed}', 'victor_view_{0.mid}_{0.seed}', 'victor_split_table_{0.mid}_{0.epoch}', {0.nSplits});"
    worker_queries       = [setup_query, build_view, build_partition_function, perm_and_slice_query]
    # UNTESTED. Last swtich is the import string
    def __init__(self,params):
        local_extra = collections.namedtuple('local_extra', 'py_partition_function import_string')
        extra_parms  = local_extra(py_partition_function=params.partition_function.get_function_name(),
                                   import_string=params.partition_function.get_import_string())
        self.ground(self.worker_queries, params, extra_params)

## Factor Steps
# THIS IS A SPEC OF CODE -- IT IS NOT OPERATIONAL
class FactorChunkerStepPSQL(SQLWorker):
    build_view = """drop view if exists {1.view_name}; create view {1.view_name} as {0.data_query}; COMMIT; """
    build_function       = """CREATE OR REPLACE FUNCTION {1.partition_function_name}(int, int, {1.view_name}) RETURNS int
    AS '{1.libpath}', '{1.sql_partition_symbol}' LANGUAGE C STRICT;
    """
    perm_and_slice_query = "SELECT partition_data_seed_fnc('{1.partition_function_name}', '{1.view_name}', 'victor_split_table_{0.mid}_{0.epoch}', {0.seed}, {0.nSplits});"
    worker_queries       = [setup_query, build_view, build_function, perm_and_slice_query]

    local_extra = collections.namedtuple('local_extra', 'view_name partition_function_name libpath sql_partition_symbol')
    def __init__(self,params):
        view_name  = "victor_view_{0.mid}_{0.seed}".format(params)
        pfname     = "victor_partition_{0.mid}_{0.seed}".format(params)
        #
        #libpath              = "/Users/chrisre/SCRATCH/DEVEL/VICTOR_SQL/trash/disused/factored_parser/C_Version/factor_partition_wrapper"
        libpath              = params.partition_function.get_libpath()
        
        sql_partition_symbol = params.partition_function.get_fname()
        extra_params         = self.local_extra(view_name=view_name, partition_function_name=pfname, libpath=libpath, sql_partition_symbol=sql_partition_symbol)
        self.ground(self.worker_queries, params, extra_params=extra_params)
        print self.grounded_queries
        
## FactorGradient
## NOT USED
class FactorGrad(SQLWorker):    
    grad_step      = "SELECT COUNT(*) FROM (SELECT {0.gradient_function}({0.example_table_attrs}) FROM {0.data_query}) AS t;"
    worker_queries = [setup_query, grad_step]
    
    def __init__(self,params):
        self.ground(self.worker_queries, params)

    
# This must be an external function for Python reasons
# (Exec cannot be executed in a function that contains a function closure)
def execute_stop_stmt(plan, grounded_loss, env):
    exec plan.stop_condition_imports_str
    loss        = env["loss"]
    stop_state  = env["stop_state"]
    should_stop = eval(grounded_loss)
    env["stop_state"] = stop_state
    return should_stop


def get_output_table_name(mid,epoch):
    return "victor_mid_{0}_epoch_{1}".format(mid,epoch)


##############################################
##### This should really be in the Planner.
##############################################
def convert_task_to_jobs(P, epoch):
    plan = P.plan_object
    def build_shuffler(plan, epoch):
        return ShuffleDesc(output_table = get_output_table_name(plan.mid,epoch),
                           data_query   = plan.data_query)

    def build_worker(plan, epoch):
        return WorkerDesc(mid=plan.mid,
                          example_table_attrs= plan.example_table_call_params,
                          gradient_function  = plan.gradient_step,
                          data_query         = get_output_table_name(plan.mid,epoch)
                          )
    def build_scorer(plan, epoch):
        return ScoreDesc(mid= plan.mid,                         
                         objective_agg       = plan.obj_aggregate_function,
                         objective_item      = plan.obj_row_function,
                         example_table_attrs = plan.example_table_call_params,
                         data_query          = plan.data_query
                         )
    # Shufflers only in epoch 1
    if epoch == 1:
        return worker_structure.WorkSet([Shuffler(build_shuffler(plan,epoch+1))])
    elif epoch == 2:
        return worker_structure.WorkSet([Shuffler(build_shuffler(plan,epoch+1)),
                IncrementalGradientStep(build_worker(plan,epoch))])
    else: # epoch > 1
        return worker_structure.WorkSet([Shuffler(build_shuffler(plan,epoch+1)),
                                         IncrementalGradientStep(build_worker(plan,epoch)),
                                         ScoreStep(build_scorer(plan, epoch-1))])

def get_seed(seed_env, epoch):
    import random
    if not seed_env.has_key(epoch):
        seed_env[epoch] = random.randint(0,1000000)
    return seed_env[epoch]

# This returns slightly more complex structures 
def convert_factor_task_to_jobs(P, epoch, seed_env):
    plan    = P.plan_object
    nSplits = plan.nSplits
    print "[FACTOR]  nSplits = {}".format(nSplits)
    def build_shuffler(plan, epoch):
        assert(plan.partition_function is not None)
        return FactorChunkStepDesc(seed       = get_seed(seed_env, epoch),
                                   data_query = plan.data_query,
                                   nSplits    = nSplits*nSplits, # unfortunate name collision. FIX THIS!
                                   epoch      = epoch,
                                   mid        = plan.mid,
                                   partition_function = plan.partition_function)
    def build_worker(plan, epoch, (roundid, chunkid)):
        # NB: THIS IS USED IN FACTOR CHUNKER!
        output_name= "victor_split_table_{mid}_{epoch}".format(mid=plan.mid, epoch=epoch)
        return IncrementalGradientStep(WorkerDesc(mid=plan.mid,
                                                  example_table_attrs= plan.example_table_call_params,
                                                  gradient_function  = plan.gradient_step,
                                                  data_query         = get_split_name(output_name, roundid*nSplits+chunkid)
                                                  )        )

        # return FactorChunkGradDesc(dbname     = sys_appends.dbname,
        #                            seed       = get_seed(seed_env, epoch),
        #                            data_query = plan.data_query,
        #                            nsplits    = nSplits, 
        #                            epoch_mod  = epoch,
        #                            mid        = plan.mid,
        #                            roundid    = roundid,
        #                            chunkid    = chunkid)

    def build_scorer(plan, epoch):
        return ScoreDesc(mid= plan.mid,                         
                         objective_agg       = plan.obj_aggregate_function,
                         objective_item      = plan.obj_row_function,
                         example_table_attrs = plan.example_table_call_params,
                         data_query          = plan.data_query
                         )
    assert(epoch >= 1)
    
    # if epoch == 1:
    #     return (worker_structure.WorkSet([FactorChunkerStep(build_shuffler(plan,epoch+1))]),worker_structure.WorkSet([]))
    # elif epoch == 2:
    #     workers = worker_structure.WorkList([ worker_structure.WorkSet([ build_worker(plan, epoch, (r,c)) for c in range(nSplits) ]) for r in range(nSplits) ])
    #     return (worker_structure.WorkSet([FactorChunkerStep(build_shuffler(plan,epoch+1))]), workers)
    # else:
    #     workers = worker_structure.WorkList([ worker_structure.WorkSet([ build_worker(plan, epoch, (r,c)) for c in range(nSplits) ]) for r in range(nSplits) ])
    #     return (worker_structure.WorkSet([FactorChunkerStep(build_shuffler(plan,epoch+1)), ScoreStep(build_scorer(plan, epoch-1))]),
    #             workers)
    if epoch == 1:
        return (worker_structure.WorkSet([FactorChunkerStepPSQL(build_shuffler(plan,epoch+1))]),worker_structure.WorkSet([]))
    elif epoch == 2:
        workers = worker_structure.WorkList([ worker_structure.WorkSet([ build_worker(plan, epoch, (r,c)) for c in range(nSplits) ]) for r in range(nSplits) ])
        return (worker_structure.WorkSet([FactorChunkerStepPSQL(build_shuffler(plan,epoch+1))]), workers)
    else:
        workers = worker_structure.WorkList([ worker_structure.WorkSet([ build_worker(plan, epoch, (r,c)) for c in range(nSplits) ]) for r in range(nSplits) ])
        return (worker_structure.WorkSet([FactorChunkerStepPSQL(build_shuffler(plan,epoch+1)), ScoreStep(build_scorer(plan, epoch-1))]),
                workers)
        
########## End Planner

##### Work Structures
##### This is on the other side
def perform_hard_work( (t,id) ):
    t0   = time.time()
    v    = t.execute()
    tend = time.time() - t0
    ret  = [("id",id), ("time".format(id),tend)]
    if v is not None:
        return ret + [("loss", v)]
    else:
        return ret

def check_values(values, plan_states):    
    ###############################
    # Process the return values
    # print values
    for v in values:
        dv = dict(v)
        if dv.has_key("loss"):
            id=dv["id"]
            plan_states[id].env["loss"] = v_wrap(dv["loss"])

### This task in a list of plans and executes them in parallel
def execute_tasks(Plans, process=3, verbose=False):
    #plan = P.plan_object
    # Setup up process pool
    pool         = Pool(processes=process)
    extra_pool   = Pool(processes=2)
    
    # Setup the plans and their enviornemnt
    PlanState   = collections.namedtuple('PlanState', 'id plan grounded_stop env')
    plan_states = [PlanState(plan=P,
                             id=i,
                             grounded_stop="{0.stop_condition_exec_str}(stop_state, loss)".format(P.plan_object),
                             env=dict([("loss",v_wrap(1e100)),("stop_state",dict()),("victor_active",True)])) for (i,P) in enumerate(Plans)]
    
    
    # To understand these conventions (e.g., who touches real data,
    # who does not) see above    
    conn                    = global_utils.connect_db()
    cur                     = conn.cursor()

    # Levent tells me that we need to import the path here before we
    # call the below functions.
    cur.execute("SELECT library_setup()")

    # Setup the storage memory manager
    # and put the models where they belong
    ss = shared_storage.start_server()
    ss.create_segment("epoch", 0)
    for ps in plan_states:
        cur.execute(ps.plan.setup_query)
        print "id={id} mid={mid} query={query}".format(id=ps.id, mid=ps.plan.plan_object.mid, query=ps.plan.setup_query)
        cur.execute("SELECT VictorStoreModelCache({0.mid});".format(ps.plan.plan_object))
    print "Setup Complete for {0} plans".format(len(Plans))

    def get_active(ps):
        return filter(lambda x:x.env["victor_active"], ps)

    def update_active(ps):
        active = False
        for ps in get_active(ps):
            should_stop = execute_stop_stmt(ps.plan.plan_object, ps.grounded_stop, ps.env)
            if should_stop:
                ps.env["active"] = False
            else:
                active = True
        return active
    
    epoch = 1 # epochs are set relative to the worker thread.
    ss.update_segment("epoch", epoch)
    seed_env = dict()
    while update_active(plan_states) > 0:
        # Execute the refinement step (and time it)
        t1          = time.time()
        # Partition out the jobs
        jobs   = []
        extra_jobs = []
        for plan_state in get_active(plan_states):
            if plan_state.plan.plan_object.factorable_exec and constants.use_factor_exec:
                 ((s,w),id) = (convert_factor_task_to_jobs(plan_state.plan, epoch, seed_env),plan_state.id)
                 extra_jobs.append( (s,id) )
                 jobs.append      ( (w,id) )
            else:
                jobs.append( (convert_task_to_jobs(plan_state.plan, epoch),plan_state.id) )

        if verbose:
            print "Jobs={} Active Plans={}".format(jobs, get_active(plan_states))
        active_jobs = jobs
        iteration   = 0

        extra_result = None
        if len(extra_jobs) > 0:
            extra_workers     = [ (w,id) for (j,id) in extra_jobs for w in j.get_round() ]
            extra_result      = pool.map_async(perform_hard_work, extra_workers)
            
        while len(active_jobs) > 0:
            workers     = [ (w,id) for (j,id) in active_jobs for w in j.get_round() ]
            print "\t Iteration {} with {} active hard workers".format(iteration, len(workers))
            values      = pool.map(perform_hard_work, workers)
            check_values(values, plan_states)
            active_jobs = filter(lambda (j,id):j.more_work(), active_jobs)
            iteration   += 1

        # join on extras
        t3 = time.time()
        if extra_result is not None:
            check_values(extra_result.get(), plan_states)
        
        t2 = time.time()
        print "[Epoch {epoch}] time={time} join_time={join_time}".format(epoch=epoch, time=t2-t1, join_time=t2-t3)
                
        for ps in get_active(plan_states):
            print "Plan {id} objective={objective}".format(id=ps.id, objective=ps.env["loss"].value)

        epoch += 1
        ss.update_segment("epoch", epoch)
        
    print "Storing Back"
    conn.commit()
    cur                     = conn.cursor()
    for ps in plan_states:
        q = "SELECT VictorGetModelCache({0.mid});SELECT VictorStoreModel({0.mid});".format(ps.plan.plan_object)
        print q
        cur.execute(q)
    conn.commit()
