
"""
This generates plan for generic linear models
"""
 
import string
import sys
import src.globals.constants as constants
## TODO: Move this to a TableDescription type
def generate_call_params(td):
    return  ','.join(td.field_names)    

class Python_Plan:
    """The Plan Generator"""
    mid=1
    step_size=1e-2
    step_diminish=0.8
    iname='instance_table'
    instance_table='instance_table'
    curry_grad='curry_grad'
    obj_aggregate_function = None
    obj_row_function       = None
    sparse_data = False
    batch_step = None
    
    def __init__(self, mi_desc,sparse_data=False, batch_step=None):
        """
        Init function
        """
        self.mi_desc                = mi_desc
        self.mid                    = mi_desc.mid


        # Package up the refinement portion
        rd = mi_desc.refinement_desc
        self.gradient_step          = rd.gradient_step
        self.obj_aggregate_function = rd.objective_aggregate
        self.obj_row_function       = rd.objective_item        
        self.stop_condition_imports_str = rd.stop_function.get_import_statement()
        self.stop_condition_exec_str    = rd.stop_function.get_qualified_name()
        self.partition_function         = rd.partition_function
        
        self.nSplits                    = 0
        self.factorable_exec            = False
        if self.partition_function is not None:
            self.nSplits                = constants.nSplits
            self.factorable_exec        = True
            print "[DEBUG] nSplits={}".format(self.nSplits)
        else:
            print "[DEBUG] Not factorable Exec {}".format(self.partition_function)
            
        # Fields for data access
        self.example_table_call_params = generate_call_params(mi_desc.example_table_desc)
        self.instance_table            = mi_desc.model_table_desc.table_name
        self.model_field_names         = mi_desc.model_table_desc.field_names        
        self.data_query                = "SELECT {0} FROM {1} as x".format(self.example_table_call_params, mi_desc.example_table_desc.declaration)
    
        # Are these still used?
        self.sparse_data = sparse_data
        self.batch_step = batch_step

    

    def generate_loss_function(self):
        """
        Generates loss function query

        @rtype: string
        @return: query to generate loss function
        """
        loss_eval = None
        if self.obj_aggregate_function and self.obj_row_function:
            loss_eval = "SELECT {0}({1}({2}, {3})) FROM ({4}) as x;\n".format(self.obj_aggregate_function, self.obj_row_function, self.mid, self.example_table_call_params, self.data_query)
        print 'loss_eval query: ' + loss_eval
        return loss_eval
    
    def incremental_plan(self):
        """
        Generates incremental plan
        
        @rtype: tuple
        @return: incremental plan tuple that contains: header, body, loss_eval and footer
        """
        # body   = "SELECT COUNT(*) FROM (SELECT {0}({1}, {2}) FROM ({3}) AS i ORDER BY RANDOM()) AS T;\n".format(self.gradient_step,
        #                                                                                                        self.mid,
        #                                                                                                        self.example_table_call_params,
        #                                                                                                        self.data_query)

        body   = "SELECT COUNT(*) FROM (SELECT {0}({1}, {2}) FROM ({3}) AS i) AS T;\n".format(self.gradient_step,
                                                                                                                self.mid,
                                                                                                                self.example_table_call_params,
                                                                                                                self.data_query)
        footer    = ""
        # self.generate_store_model()
        loss_eval = self.generate_loss_function()


        return [(self.stop_condition_imports_str,self.stop_condition_exec_str), body, loss_eval, footer]
        

    def batchgradient_plan(self):
        """
        Generates batch gradient plan
        
        @rtype: tuple
        @return: batch plan tuple that contains: header, body, loss_eval and foote
        """
        # This is totally broken. Needs to be fixed.
        assert(False)
        #header       = self.generate_cache_model()
        #header      += "SELECT zero_gradient(%d);\n" % (self.mid)

        #prox?
        #assert( self.prox ) # contract says must implement a trivial prox that zeros the gradient
        
        body   = "SELECT COUNT(*) FROM (SELECT {0}({1}, {2}) FROM ({3}) AS i) AS T;\n".format(self.gradient_step,
                                                                                              self.mid,
                                                                                              self.example_table_call_params,
                                                                                              self.data_query)
                
            
        #z   += "SELECT %s(%d, %f);\n" % (self.prox, self.mid,current_step)
        #body.append(z)                

        # generate footer
        footer = self.generate_store_model()
        loss_eval = self.generate_loss_function()
        return [(self.stop_condition_imports_str,self.stop_condition_exec_str), body, loss_eval, footer]

