import sys_appends, time
import copy
#from victor_plan_execution import PlanExecutor
#import generate_plan
import src.vsql_core.plan_generators.generate_plan_factorized as generate_plan_factorized
import src.globals.constants as constants
import src.globals.global_types as global_types


def create_setup_query(mi_desc): 
    mi_table_name    = mi_desc.model_table_desc.table_name
    refine_desc      = mi_desc.refinement_desc
    init_model_query = "SELECT {0.initialization_function}({mid});".format(refine_desc, mid=mi_desc.mid)    

    # This delete should not be here -- we should not allow overwrites unless it is explicit? or not?
    delete_query = "DELETE FROM {0} WHERE {1} = {2};".format(mi_table_name, constants.MID_FIELD, mi_desc.mid)    
    insert_query = "INSERT INTO {model_instance_table}({mid_field}, init_fnc, train_fnc) VALUES ({mid}, '{init_fn_name}', '{gradient_name}');".format(
        mid_field=constants.MID_FIELD,
        model_instance_table=mi_table_name,
        mid=mi_desc.mid,
        init_fn_name=refine_desc.initialization_function,
        gradient_name=refine_desc.gradient_step)


    return '\n'.join([init_model_query, delete_query, insert_query])
    

# General OVA plan
def ova_plan(mi_desc):    
    """
    Generates ova plan
    """
    ova = mi_desc.opts.ova
    
    work_list = []    
    t0 = time.time()
    base_mid  = mi_desc.mid
    for (i,(label,function)) in enumerate(ova.label_functions):
        # Convention is: we start at the base  mid and we make one model per class. This isn't ideal
        # and eventually we need a good (and probably heavier) mechansim
        mid                   = base_mid + i
        # Setup Query
        new_mi_desc                = copy.deepcopy(mi_desc)
        new_mi_desc.mid            = mid        
        setup_query                = create_setup_query(new_mi_desc)

        # Fill in the example view
        ex_table                   = mi_desc.example_table_desc
        attrs                      = filter(lambda x: x <> ova.label_attr, ex_table.field_names)
        updated_query              = "(SELECT {4}, {0}({1}) as {2} FROM {3} as x)".format(function, ova.label_attr, constants.VICTOR_LABEL, ex_table.declaration, ','.join(attrs))
        
        td                         = global_types.TableOrViewDescription(declaration=updated_query, field_names=attrs + [constants.VICTOR_LABEL])

        new_mi_desc.example_table_desc = td
        
        
        P = generate_plan_factorized.Python_Plan(new_mi_desc)
        
        # Generate the plan
        if mi_desc.opts.batch_gradient:        
            plan = P.batchgradient_plan()
        else:
            plan = P.incremental_plan()

        # debugging junk
        print setup_query
        print plan

        work_list.append(global_types.ExecutorTuple(plan_id=mid, connect_str=mi_desc.connect_str, setup_query=setup_query, plan=plan,plan_object=P))
    return work_list



def generate_single_threaded_plan(mi_desc):
    mid              = mi_desc.mid
    mi_table_name    = mi_desc.model_table_desc.table_name

    setup_query      = create_setup_query(mi_desc)
    print 'setup query: ' + setup_query
    opts = mi_desc.opts
    ################################################
    # Now generate a plan for the refinement portion
    #
    P = generate_plan_factorized.Python_Plan(mi_desc)
    if opts.batch_gradient:
        print "**** Using Batch Gradient ****"
        plan = P.batchgradient_plan()
    else:
        plan = P.incremental_plan()

    # debugging junk
    print setup_query
    print plan

    return global_types.ExecutorTuple(plan_id=0, connect_str=mi_desc.connect_str, setup_query=setup_query, plan=plan, plan_object=P) 

def generate_plan(mi_desc):
    if mi_desc.opts.ova is None:
        print "*** SINGLE THREADED PLAN ***"
        return [generate_single_threaded_plan(mi_desc)]
    else:
        print "*** GENERATING OVA PLAN ***"
        return ova_plan(mi_desc)

def simple_planner(mi_desc):
    sh_exec = simple_executor.Shuffler(mi_desc)
    ig_exec = simple_executor.IncrementalGradStep(mi_desc)
    ss_exec = simple_executor.ScoreStep(mi_desc)
    
