"""
This module includes all non-array, 1d and 2d array type information
"""

# OID TABLE from pg_type.h
# oid, monkiker, python type (including endian), byte size

charoid        = 18
booloid        = 16
int2oid        = 21
int4oid        = 23
int8oid        = 20
float4oid      = 700
float8oid      = 701
chararrayoid   = 1002
boolarrayoid   = 1000
int2arrayoid   = 1005
int4arrayoid   = 1007
int8arrayoid   = 1016
float4arrayoid = 1021
float8arrayoid = 1022
float82darrayoid = 0 #TODO: Do we need this?
partition_tupleoid = 0

data_struct_type_field = 1
data_python_type_field = 2
data_size_field = 3

oid_table = [
    #Non-Array Types
    (charoid  , "CHAROID", 'c','c', 1),
    (booloid  , "BOOLOID", 'b', 'b', 1),
    (int2oid  , "Int2OID", '>h', 'h', 2),
    (int4oid  , "Int4OID", '>i', 'i', 4), 
    (int8oid  , "Int8OID", '>q', 'q', 8),
    (float4oid, "Float4OID", '>f', 'f', 4),
    (float8oid, "Float8OID", '>d', 'd', 8),
    #Array Types
    #length?
    (chararrayoid, "_CHAROID", '', '', 0),
    (boolarrayoid, "_BOOLOID", '', '', 0),
    (int2arrayoid, "_Int2OID", '', '', 0),
    (int4arrayoid, "_Int4OID", '', '', 0),
    (int8arrayoid, "_Int8OID", '', '', 0),
    (float4arrayoid, "_Float4OID", '', '', 0),
    (float8arrayoid, "_Float8OID", '', '', 0),
    ##2d arrays
    (float82darrayoid, "", '', '', 0),
    (partition_tupleoid, "", '', '', 0),
    ]

oid_dict = dict()
for  (oid,moniker,struct_type,python_type,size) in oid_table:  
    oid_dict[oid] = (moniker,struct_type,python_type,size)

# Is it more costly to use 
# oid_dict[int4oid][data_size_field] instead of size_int4
size_int = oid_dict[int4oid][data_size_field]
size_int2 = oid_dict[int2oid][data_size_field]
size_float8 = oid_dict[float8oid][data_size_field]
