import src.globals.global_utils as global_utils
import collections
import time, psycopg2
from psql_util import *
DataPartInfo = collections.namedtuple('DataPartitionInfo', 'partition_fnc table_name output_name nSplits')
asize = 20000
nBuf  = 1000000

def get_split_name(output_name, id):
    return "{output_name}_split_{split_id}".format(output_name = output_name,
                                                   split_id    = id)

        
def partition_data(partition_fnc, input_view_name, output_base_name, seed, nSplits):
    conn                    = global_utils.connect_db()
    cur                     = conn.cursor()
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    
    #Added
    cur.execute("select library_setup()")
    ##Create tables that the actual data table
    split_table_names = [ get_split_name(output_base_name, i) for i in range(nSplits) ]
    tw_array = []
    drop_statement    = ';\n'.join([ "DROP TABLE IF EXISTS {split_table_name};;".format(
        split_table_name = split_table_name,
        input_view_name=input_view_name) for split_table_name in split_table_names])

    cur.execute(drop_statement)
    create_statement    = ';\n'.join([ "SELECT * INTO {split_table_name} FROM {input_view_name} as x LIMIT 0".format(
        split_table_name = split_table_name,
        input_view_name=input_view_name) for split_table_name in split_table_names])

    cur.execute(create_statement)

    print "Partition Starts"
    query                   = "SELECT {partition}({seed}, {nSplits}, t.*), * FROM {input_view_name} as t".format(partition  = partition_fnc,
                                                                                        input_view_name=input_view_name, nSplits=nSplits, seed=seed)
    print "{}".format(query)
    conn.commit()
    conn.close()

    
    conn                    = global_utils.connect_db()
    server_side_cur         = conn.cursor(name="ss_cursor")    
    server_side_cur.arraysize = asize

    server_side_cur.execute(query)
    
    schema = []
    first_tuple = server_side_cur.fetchone()
    print "First tuple is back"
    # raise ValueError("YES {}".format(server_side_cur.description))
    
    for (i, d) in enumerate(server_side_cur.description):
        #First field will always be the int (that returns from partition function). So, we need to skip this field
        if i > 0:
            schema.append(d.type_code)

    ts = TupleSchema(schema)
    cur = conn.cursor()
    #Now, we retrieve schema. we can create tuple writers
    tw_array = [ TupleWriter(cur, split_table_names[i], ts, True, nBuf) for i in xrange(nSplits) ]


    
    t1 = time.time()
    tuples   = [first_tuple]
    loop_counter = 1
    while tuples <> []:
        for t in tuples:
            partition_id      = t[0]
            rest_of_the_tuple = t[1:]
            tw_array[partition_id].write_tuple(rest_of_the_tuple)
        if loop_counter % 50 == 0:
            print "\t Loops...{}".format(loop_counter)
        loop_counter += 1
        tuples = server_side_cur.fetchmany(asize)

    print 'write tuple completes in {0}'.format(time.time() - t1)

    t1 = time.time()
    for i in range(nSplits):
        tw_array[i].close()

    print 'copying to postgres is completed in {0}'.format(time.time() - t1)

    t1 = time.time()
    conn.commit()
    print 'committing is completed at {0}'.format(time.time() - t1)

# move this to a separate test handler file
#partition_data('factor_partition_wrapper', 'movielens1m', 'newww_test2_output_base', 10, 9)
