import psycopg2
import src.globals.constants as constants
import src.vsql_core.utils.pickler.victor_pickler as victor_pickler
import src.globals.global_utils as global_utils
import src.normalization.create_model as create_model
import src.globals.victor_exceptions as victor_exceptions
import collections

ModelAppDescription = collections.namedtuple('ModelApplicationDescription', 'model_app_name model_app_fnc model_inst_name data_field_params return_type')

def create_model_application(param_dict):
    model_app_sql_fnc_name     = param_dict[constants.PARSER_NAME_FIELD]
    model_app_python_fnc       = param_dict[constants.PARSER_MODEL_APP_PYTHON_FNC]
    model_instance_name        = param_dict[constants.PARSER_MODEL_INST_NAME]
    data_item_type_fields      = param_dict[constants.PARSER_DATA_TYPE_FIELDS]
    return_type                = param_dict[constants.PARSER_MODEL_APP_RETURN_TYPE]

    ##Check that we have a model variable whose type is model_instance_name
    index = 0
    for (p, t) in data_item_type_fields:
        if t == model_instance_name:
            break
        index += 1
    if index >= len(data_item_type_fields):
        raise victor_exceptions.VictorException('Specified model instance could not be found in the function parameters')
            

    conn = global_utils.connect_db()
    cur = conn.cursor()

    #Retrieve mid from the global mid table
    select_query = "SELECT {mid_field} FROM {global_mid_table} where {model_instance_field} = '{mi_name}';".format(
        mid_field            = constants.MID_FIELD,
        global_mid_table     = constants.VICTOR_GLOBAL_MODEL_ID_TABLE,
        model_instance_field = constants.VICTOR_GLOBAL_INSTANCE,
        mi_name              = model_instance_name)

    cur.execute(select_query)
    (mid, ) = cur.fetchone()

    model_app_desc = ModelAppDescription(model_app_name    = model_app_sql_fnc_name,
                                         model_app_fnc     = model_app_python_fnc,
                                         model_inst_name   = model_instance_name,
                                         data_field_params = data_item_type_fields,
                                         return_type       = return_type)
    function_content = create_wrapper_app_function(mid, model_app_desc, index)
    #create apply_function
    cur.execute(function_content)
    #insert function to the function global table
    insert_query = """INSERT INTO {model_app_table} VALUES('{model_app_name}', '{model_inst_name}', 
    {mid}, '{function_content}')""".format(model_app_table  = constants.VICTOR_MODEL_APP_TABLE,
                                           model_app_name   = model_app_sql_fnc_name,
                                           model_inst_name  = model_instance_name,
                                           mid              = mid,
                                           function_content = function_content)
    cur.execute(insert_query)
    conn.commit()

def create_wrapper_app_function(mid, model_app_desc, model_param_index):
    fname = model_app_desc.model_app_name
    params = model_app_desc.data_field_params
    
    #param call str for python function
    param_call_str   = create_model.unpickle_data_types(params)
    #model name and its type (its type should be model instance name)
    (model_name, model_type) = params[model_param_index]

    #we remove model name and model type because it shouldnt be appear among the parameters of
    #wrapper function
    params.remove((model_name, model_type))
    (param_type_str, disused, param_full_str) = create_model.get_param_type_call_full_str(params)

    import_str = model_app_desc.model_app_fnc.get_import_statement()    
    import_str += ", {0}".format(constants.VICTOR_PICKLER) 

    query = """DROP FUNCTION IF EXISTS {fnc_name}({typed_params}); 
    CREATE FUNCTION {fnc_name}({typed_params}) RETURNS {return_type}
    AS $$
    \t{import_statement}
    \tif SD.has_key({mid}):
    \t\t{model_name} = SD[{mid}]
    \telse:
    \t\timport src.utils.retrieve_and_store_methods as retrieve_store
    \t\tSD[{mid}] = retrieve_store.retrieve_model({mid})
    \t\t{model_name} = SD[{mid}]
    \treturn {python_fnc}({data_call_params})
    $$ LANGUAGE plpythonu;
    """.format(fnc_name         = fname,
               typed_params     = param_full_str,
               import_statement = import_str,
               mid              = mid,
               python_fnc       = model_app_desc.model_app_fnc.get_qualified_name(),
               data_call_params = param_call_str,
               return_type      = model_app_desc.return_type,
               model_name       = model_name)

    print query
    return query
