"""
This is a helper module for the low rank code
"""

import math
import random
# These is a helper module for the low rank code
# Should probably be factored into a common util.
# common utilities should be in their own directory (not in the model descriptions)

def dot(x,y):
    """
    Computes dot products of the given two vectors and returns the dot product value
    
    @type x: vector
    @param x: vector
    @type y: vector
    @param y: vector
    
    @rtype: double
    @return: returns the dot product value of two vectors
    """
    n = min(len(x),len(y))
    ret = 0.0
    for i in range(n):
            ret = ret + x[i]*y[i]
    return ret

def scale_and_add(dst,src,s):
    """
    This method adds scaled vector to the given other vector

    @type dst: vector
    @param dst: dst vector
    @type src: vector
    @param src: src vector which will be scaled
    @type s: double
    @param s: scale factor

    """
    n = min(len(dst),len(src))
    for i in range(n):
        dst[i] = dst[i] + src[i]*s 

def ball_project(x, B, B2):
    """
    This function normalizes given vector if its norm square is greater than the given threshold
    
    @type x: vector
    @param x: given vector to normalized
    @type B: double
    @param B: given norm
    @type B2: double
    @param B2: norm threshold
    """
    norm_square = sum([xi*xi for xi in x])
    if(norm_square > B2):
        norm = math.sqrt(norm_square)
        for i in range(len(x)):
            x[i] = x[i]* B/norm

def make_random_vec(dim):
    """
    Creates random vector with the given dimension
    
    @type dim: number
    @param dim: dimension of the random vector

    @rtype: vector
    @return: random vector with the given dim
    """
    return [ random.gauss(0,1) * 1e-2 for i in range(dim) ]

def make_initial(max_rank, nRows, nCols, B):
    """
    Make initial function
    @type max_rank: double
    @param max_rank: max rank(dim)
    @type nRows: number
    @param nRows: # of rows
    @type nCols: number
    @param nCols: #of cols
    @type B: double
    @param B: B

    @rtype: tuple
    @return: returns L and R
    """
    L = []
    R = []
    for i in range(nRows):
        random_vec = make_random_vec(max_rank)
        ball_project(random_vec, B, B*B)
        L.append(random_vec)
        
    for i in range(nCols):
        random_vec = make_random_vec(max_rank)
        ball_project(random_vec, B, B*B)
        R.append(random_vec)
    return (L,R)


def get_bucket(nsplits, nrows, row, row_map):
    """
    Gets bucket
    @type nsplits: number
    @param nsplits: splits
    @type nrows: number
    @param nrows: # of rows
    @type row: number
    @param row: row index
    @type row_map: vector
    @param row_map: row map

    @rtype: number
    @return: bucket
    """
    row_i = row_map[row]
    row_bucket_size = nrows/nsplits
    return max(row_i/row_bucket_size, nsplits - 1)

def split_a_model(nSplits, nRows, L, row_perm):
    """
    
    """
    splits = [[] for z in nSplits]
    for i in range(len(X)):
        b = get_bucket(nsplits, nRows, i, row_perm)
        splits[b].append(L[i])
    return splits
    
def split_the_model(nSplits, nRows, nCols, L, R, row_perm, col_perm):
    l_split = split_a_model(nSplits, nRows, L, row_perm)
    r_split = split_a_model(nSplits, nCols, R, col_perm)
    return [l_split, r_split]

####################
# Tablename helpers
##################
def tablename_prefix(mid, epoch_sign):
    return "victor_%d_%d" % (mid,epoch_sign)

def tablename_lmodel(mid, epoch_sign, part):
    return "%s_lmodel%d" % (tablename_prefix(mid,epoch_sign), part)


def tablename_rmodel(mid, epoch_sign, part):
    return "%s_rmodel%d" % (tablename_prefix(mid,epoch_sign), part)

def tablename_data(mid,epoch_sign, part):
    return "%s_data_%d" % (tablename_prefix(mid,epoch_sign), part)

def get_l_chunk(nsplits, round, chunk):
    return (round - chunk) % nsplits

def get_r_chunk(nsplits, round, chunk):
    return chunk

def get_part(nsplits, round, chunk):
    l = get_l_chunk(nsplits, round, chunk)
    r = get_r_chunk(nsplits, round, chunk)
    return l * nsplits + r

class ReadList:
    def __init__(self, data):
        self.d = iter(data)


    def readline(self, size=None):
        try:
            (i,vec) = self.iter.next()
        except StopIteration:
            return ''
        else:
            return "%d\t%s\n" % (i,vec)

    def read(self, size=None):
        readline(self, size)

def build_array_string(v):
    return "{%s}" % (','.join([str(x) for x in v]))

#def copy_list(dbname, tablename, l):
#     psql = psycopg2.connect(dbname)
#     psql.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)   
#     c    = psql.cursor()
#     c.execute("DROP TABLE IF EXISTS %s; CREATE TABLE %s (id int, vec float8[]);" % (tablename, tablename))
#     z    = ["%d\t%s" % (id,build_array_string(vec)) for (id,vec) in l]
#     f    = StringIO.StringIO('\n'.join(z))
#     c.copy_from(f, tablename)
    
