"""
This class is the parallel execution for factors
"""

from multiprocessing import Pool
import sys_appends
import time
import sys, getopt
import random

## This is the parallel execution for factors
## This has the convoy problem
## It requires serious effort to fix the convyo problem

import psycopg2

class WorkerThread:
    def __init__(self, mid, max_rank, nRows, nCols, nSplits, connect_str,table_name):
        """
        Initialize method
        """
        self.mid = mid
        self.nRows = nRows
        self.nCols = nCols
        self.nSplits = nSplits
        self.conn = psycopg2.connect(connect_str)
        self.conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        self.psql = self.conn.cursor()
        self.max_rank = max_rank
        self.table_name = table_name
        
    def take_step(self, seed, roundid, chunkid):
        # step 1. Retrieve model and setup permtuations
        #execute(self.retrieve_model_plan, [mid])
        t0 = time.time()
        print "[take_step:%s,%s] Enter" % (roundid,chunkid)
        self.psql.execute("SELECT library_setup();")
        self.psql.execute("SELECT retrieve_model_instance_string(%s)" % (self.mid))
        print "[take_step:%s,%s] Retrieved Model in %s"  % (roundid,chunkid, time.time() - t0)
        t1 = time.time()
        
        self.psql.execute("SELECT create_permutations(%s,%s,%s)" % (self.nRows, self.nCols, seed))
        print "[take_step:%s,%s] Permutations Created. Executing Gradient."  % (roundid,chunkid)

        t2 = time.time()
        exec_str = "SELECT COUNT(*) FROM (SELECT take_step(row - 1, col - 1, rating) FROM " + self.table_name
        exec_str = exec_str + " WHERE round(%s,%s,%s,row-1,col-1) = %s AND chunk(%s,%s,%s,row-1,col-1) = %s) as t;"
        exec_str = exec_str % (self.nSplits, self.nRows, self.nCols, roundid, self.nSplits, self.nRows, self.nCols, chunkid)
        
        # Step 2. execute the step
        self.psql.execute(exec_str)
        (nCount,) = self.psql.fetchone()
        print "[take_step:%s,%s] Finished Step (%d tuples) in %s secs. Storing"  % (roundid,chunkid, nCount, time.time() - t2)

        t3 = time.time()
        # Step 3. transactionally execute the update
        #self.psql.execute("SELECT update_fragment(%s,%s,%s,%s,%s,%s,%s);" % (self.max_rank, self.nSplits, self.nRows, self.nCols, chunkid, roundid, self.mid))
        self.psql.execute("SELECT overwrite_fragment(%s,%s,%s,%s,%s,%s,%s);" % (self.max_rank, self.nSplits, self.nRows, self.nCols, chunkid, roundid, self.mid))
        print "[take_step:%s,%s] Complete in %f sec (update=%s)"  % (roundid,chunkid, time.time() - t0, time.time() - t3)

    def close(self):
        self.psql.close()

class chunk_desc:
    def __init__(self, mid, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name):
        self.mid = mid
        self.max_rank = max_rank
        self.nRows = nRows
        self.nCols = nCols
        self.nsplits = nsplits
        self.connect_str = connect_str
        self.seed = seed
        self.chunkid = chunkid
        self.roundid = roundid
        self.table_name = table_name

def chunk_maker(mid, max_rank, nRows, nCols, nsplits, connect_str, seed, roundid,table_name):
    ret = []
    for chunkid in range(nsplits):
            ret.append(chunk_desc(mid, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name))
    return ret
                       
def perform_work(cd):
    wt = WorkerThread(cd.mid, cd.max_rank, cd.nRows, cd.nCols, cd.nsplits, cd.connect_str, cd.table_name)
    wt.take_step(cd.seed, cd.roundid, cd.chunkid)
    wt.close()
    return (cd.roundid,cd.chunkid)

def install_fresh_database(psql, mid, max_rank, b, stepsize, table_name):
    print "[INIT] Installing fresh model instance to mid=%d with rank=%d b=%f stepsize=%f" % (mid, max_rank, b, stepsize)
    table_query = "SELECT MAX(row), MAX(col), AVG(rating) FROM %s" % table_name
    psql.execute(table_query)
    (nRows, nCols, mean,) = psql.fetchone()

    psql.execute("DELETE FROM lr_model_instance WHERE MID=%d;" % (mid))
    psql.execute("SELECT initialize_model(%d,%d,%d,%f,%f,%f)" % (max_rank, nRows, nCols, b, mean, stepsize))    
    psql.execute("SELECT store_model(%d);" % (mid))
    print "[INIT] Stored model mid=%d" % mid
    return (nRows, nCols)
    
def get_database_stats(psql, mid, table_name):
    print "[INIT] Getting stats about mid=%d and table=%s" % (mid, table_name)
    table_query = "SELECT MAX(row), MAX(col) FROM %s" % table_name
    psql.execute(table_query)
    (nrows, ncols) = psql.fetchone()    
    psql.execute("SELECT array_upper(L,2) as l_rank FROM lr_model_instance WHERE mid=%d LIMIT 1" % (mid))
    (max_rank,) = psql.fetchone()
    return (nrows, ncols, max_rank)
    
def fresh_execute(mid, nsplits, nepochs, table_name, probe_table, max_rank, b, initial_stepsize, diminish):
    connect_str = 'dbname=' + sys_appends.dbname
    conn = psycopg2.connect(connect_str)
    psql = conn.cursor()
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    psql.execute("SELECT library_setup();")
    
    (nRows, nCols) = install_fresh_database(psql, mid, max_rank, b, initial_stepsize, table_name)
            
    pool = Pool(processes=nsplits)
    for epoch in range(nepochs):
        start_epoch = time.time()
        seed = random.randint(1,1000000)
        print "[EPOCH] using seed=%d" % (seed)
        for i in range(nsplits):
            print pool.map(perform_work, chunk_maker(mid, max_rank, nRows, nCols, nsplits, connect_str, seed, i, table_name))
        print "[End Epoch] %s" % (time.time() - start_epoch)
    
        psql.execute("SELECT retrieve_model_instance_string(%d)" % mid)
        if not (probe_table is None):
            psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
            (rmse, ) = psql.fetchone()
            print "[RMSE] %f" % (rmse)
        psql.execute("UPDATE lr_model_instance SET stepsize=stepsize*%f WHERE MID=%d" % (diminish, mid))
    return

def stored_execute(mid, nsplits, nepochs, table_name, probe_table, diminish):
    connect_str = 'dbname=' + sys_appends.dbname
    conn = psycopg2.connect(connect_str)
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    psql = conn.cursor()
    psql.execute("SELECT library_setup();")       

    (nRows, nCols, max_rank) = get_database_stats(psql, mid, table_name)
    
    pool = Pool(processes=nsplits)
    for epoch in range(nepochs):
        start_epoch = time.time()
        seed = random.randint(1,1000000)
        print "[EPOCH] using seed=%d" % (seed)
        for i in range(nsplits):
            print pool.map(perform_work, chunk_maker(mid, max_rank, nRows, nCols, nsplits, connect_str, seed, i, table_name))
        print "[End Epoch] %s" % (time.time() - start_epoch)
        psql.execute("SELECT retrieve_model_instance_string(%d)" % mid)
        if not (probe_table is None):
            psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
            (rmse, ) = psql.fetchone()
            print "[RMSE] %f" % (rmse)
        psql.execute("UPDATE lr_model_instance SET stepsize=stepsize*%f WHERE MID=%d" % (diminish, mid))
        
    psql.close()

def usage(argv):
    print "%s: <mid> <table_name>" % (argv[0])
    
def main():
    try:
        long_opts = ["mid=", "parallel=", "epochs=", "table=", "probe=", "rank=", "stepsize=", "diminish="]
        opts, args = getopt.getopt(sys.argv[1:], "fm:p:e:t:r:b:s:d:", long_opts)
    except getopt.GetoptError:
        print str(err)
        usage(sys.argv)
        sys.exit(2)
    output  = None
    verbose = False

    mid      = None
    nthreads = 1
    epochs   = 10
    rank     = 10
    table_name = None
    probe_name = None
    stepsize   = 1e-2
    diminish   = 0.8
    B          = 1.5
    exec_fresh = True
    
    for o,a in opts:        
        if o in ("-m", "--mid"):
            mid = int(a)
        elif o == "-f":
            exec_fresh = False
        elif o in ("--parallel"):
            nthreads = int(a)
        elif o in ("-e", "--epochs"):
            epochs  = int(a)
        elif o in ("-t", "--table"):
            table_name = a
        elif o in ("-p", "--probe"):
            probe_name = a
        elif o in ("-r", "--rank"):
            rank       = int(a)
        elif o in ("-b"):
            B = float(a)
        elif o in ("-s", "--stepsize"):
            stepsize = float(a)
        elif o in ("-d", "--diminish"):
            diminish = float(a)

    if (mid is None) or (table_name is None):
        print "mid and table name are required"
        usage(sys.argv)
        sys.exit()
    else:
        print "Execute on table %s with mid %d using %d processes" % (table_name,mid, nthreads)
                                                   
    if exec_fresh:
        fresh_execute(mid, nthreads, epochs, table_name, probe_name, rank, B, stepsize, diminish)
    else:
        stored_execute(mid, nthreads, epochs, table_name, probe_name, diminish)
        

main()
