"""
This is the parallel execution for factors
"""

from multiprocessing import Pool
import sys_appends
import time
import sys, getopt
import random

import psycopg2

def get_tablename(tablename, epoch, nSplits, roundid, chunkid):
    epoch_sign = epoch % 2
    linear = roundid*nSplits + chunkid
    s      = "%s%d%d" % ("staging_data", epoch_sign, linear)
    return s

def clean_tables():
    connect_str = 'dbname=' + sys_appends.dbname
    conn = psycopg2.connect(connect_str)
    psql = conn.cursor()
    psql.execute("SELECT tablename from pg_tables where tablename LIKE 'staging_data%'")
    for (tname,) in psql.fetchall():
        print "DROP TABLE %s;" % (tname)

## This is the parallel execution for
## for factors.
class WorkerThread:
    def __init__(self, mid, max_rank, nRows, nCols, nSplits, connect_str,table_name):
        self.mid = mid
        self.nRows = nRows
        self.nCols = nCols
        self.nSplits = nSplits
        self.conn = psycopg2.connect(connect_str)
        self.conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        self.psql = self.conn.cursor()
        self.max_rank = max_rank
        self.table_name = table_name
        
    def take_step(self, seed, epoch, roundid, chunkid):
        # step 1. Retrieve model and setup permtuations
        #execute(self.retrieve_model_plan, [mid])
        t0 = time.time()
        print "[take_step:%s,%s] Enter" % (roundid,chunkid)
        self.psql.execute("SELECT library_setup();")
        self.psql.execute("SELECT retrieve_model_instance_string(%s)" % (self.mid))
        print "[take_step:%s,%s] Retrieved Model %d in %s"  % (roundid,chunkid, self.mid, time.time() - t0)
        t1 = time.time()
        
        #self.psql.execute("SELECT create_permutations(%s,%s,%s)" % (self.nRows, self.nCols, seed))
        #print "[take_step:%s,%s] Permutations Created. Executing Gradient."  % (roundid,chunkid)

        t2 = time.time()
        tname = get_tablename(self.table_name, epoch, self.nSplits, roundid, chunkid)
        exec_str = "SELECT COUNT(*) FROM (SELECT take_step(row - 1, col - 1, rating) FROM %s) AS t;" % (tname)
        print "[take_step:%s,%s] %s" % (roundid, chunkid, exec_str)
        # Step 2. execute the step
        self.psql.execute(exec_str)
        (nCount,) = self.psql.fetchone()
        print "[take_step:%s,%s] Finished Step (%d tuples) in %s secs. Storing"  % (roundid,chunkid, nCount, time.time() - t2)

        t3 = time.time()
        # Step 3. transactionally execute the update
        ow = "SELECT overwrite_fragment(%s,%s,%s,%s,%s,%s,%s);" % (self.max_rank, self.nSplits, self.nRows, self.nCols, chunkid, roundid, self.mid)
        print ow
        self.psql.execute(ow)
        print "[take_step:%s,%s] Complete in %f sec (update=%s)"  % (roundid,chunkid, time.time() - t0, time.time() - t3)

    def close(self):
        self.psql.close()


def chunk_data(psql, query_str, nsplits, nRows,nCols, epoch):
    t0 = time.time()
    epoch_sign = (epoch + 1) % 2
    table_name = "%s%d" % ("staging_data", epoch_sign)
    print "[chunk_data] Chunking the data for epoch %d" % (epoch+1)
    exec_str = "SELECT perm_slice('" + sys_appends.dbname + "', '%s', 'staging_data%d', %d, %d, %d);" % (query_str, epoch_sign, nsplits, nRows, nCols)
    print exec_str
    psql.execute(exec_str)
    print "[chunk_data] Finished Chunking in %s secs."  % (time.time() - t0)
            
class chunk_desc:
    def __init__(self, mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name):
        self.mid = mid
        self.epoch = epoch
        self.max_rank = max_rank
        self.nRows = nRows
        self.nCols = nCols
        self.nsplits = nsplits
        self.connect_str = connect_str
        self.seed = seed
        self.chunkid = chunkid
        self.roundid = roundid
        self.table_name = table_name

def chunk_maker(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, roundid,table_name):
    if roundid == 0:
        ret = [chunk_desc(mid,epoch,max_rank, nRows, nCols, nsplits, connect_str, seed, -1, -1, table_name)]
    else:
        ret = []
    for chunkid in range(nsplits):
            ret.append(chunk_desc(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name))
    return ret
                       
def perform_work(cd):
    if cd.roundid < 0:
        conn = psycopg2.connect(cd.connect_str)
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        psql = conn.cursor()
        qstr = "SELECT row, col, rating FROM %s" % cd.table_name
        chunk_data(psql, qstr, cd.nsplits, cd.nRows, cd.nCols, cd.epoch)
        psql.close()
        conn.close()
    else:
        wt = WorkerThread(cd.mid, cd.max_rank, cd.nRows, cd.nCols, cd.nsplits, cd.connect_str, cd.table_name)
        wt.take_step(cd.seed, cd.epoch, cd.roundid, cd.chunkid)
        wt.close()
    return (cd.roundid,cd.chunkid)

def install_fresh_database(psql, mid, max_rank, b, stepsize, table_name):
    print "[INIT] Installing fresh model instance to mid=%d with rank=%d b=%f stepsize=%f" % (mid, max_rank, b, stepsize)
    table_query = "SELECT MAX(row), MAX(col), AVG(rating) FROM %s" % table_name
    psql.execute(table_query)
    (nRows, nCols, mean,) = psql.fetchone()

    psql.execute("DELETE FROM lr_model_instance WHERE MID=%d;" % (mid))
    psql.execute("SELECT initialize_model(%d,%d,%d,%f,%f,%f)" % (max_rank, nRows, nCols, b, mean, stepsize))    
    psql.execute("SELECT store_model(%d);" % (mid))
    print "[INIT] Stored model mid=%d" % mid
    return (nRows, nCols)
    
def get_database_stats(psql, mid, table_name):
    print "[INIT] Getting stats about mid=%d and table=%s" % (mid, table_name)
    table_query = "SELECT MAX(row), MAX(col) FROM %s" % table_name
    psql.execute(table_query)
    (nrows, ncols) = psql.fetchone()    
    psql.execute("SELECT array_upper(L,2) as l_rank FROM lr_model_instance WHERE mid=%d LIMIT 1" % (mid))
    (max_rank,) = psql.fetchone()
    return (nrows, ncols, max_rank)
    
def fresh_execute(mid, nsplits, nepochs, table_name, probe_table, max_rank, b, initial_stepsize, diminish):
    connect_str = 'dbname=' + sys_appends.dbname
    conn = psycopg2.connect(connect_str)
    psql = conn.cursor()
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    psql.execute("SELECT library_setup();")
    
    (nRows, nCols) = install_fresh_database(psql, mid, max_rank, b, initial_stepsize, table_name)
    if not (probe_table is None):
        psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
        (rmse, ) = psql.fetchone()
        print "[Initial RMSE] %f" % (rmse)
    
    qstr = "SELECT row, col, rating FROM %s" % (table_name)    
    chunk_data(psql, qstr, nsplits, nRows, nCols, -1)
    pool = Pool(processes=nsplits+1)
    for epoch in range(nepochs):
        start_epoch = time.time()
        seed = random.randint(1,1000000)
        print "[EPOCH] %d using seed=%d" % (epoch,seed)
        for i in range(nsplits):            
            print pool.map(perform_work, chunk_maker(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, i, table_name))
        epoch_time = time.time() - start_epoch
        print "[End Epoch] %s" % (epoch_time)
    
        psql.execute("SELECT retrieve_model_instance_string(%d)" % mid)
        if not (probe_table is None):
            psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
            (rmse, ) = psql.fetchone()
            print "[RMSE] epoch=%d time=%f rmse=%f" % (epoch, epoch_time,rmse)
        psql.execute("UPDATE lr_model_instance SET stepsize=stepsize*%f WHERE MID=%d" % (diminish, mid))
    return



def stored_execute(mid, nsplits, nepochs, table_name, probe_table, diminish):
    connect_str = 'dbname=' + sys_appends.dbname
    conn = psycopg2.connect(connect_str)
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    psql = conn.cursor()
    psql.execute("SELECT library_setup();")       

    (nRows, nCols, max_rank) = get_database_stats(psql, mid, table_name)
    qstr = "SELECT row, col, rating FROM %s" % (table_name)
    chunk_data(psql, qstr, nsplits, nRows, nCols, -1)
    
    pool = Pool(processes=nsplits)
    for epoch in range(nepochs):
        start_epoch = time.time()
        seed = random.randint(1,1000000)
        print "[EPOCH] using seed=%d" % (seed)
        for i in range(nsplits):
            print pool.map(perform_work, chunk_maker(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, i, table_name))
        print "[End Epoch] %s" % (time.time() - start_epoch)
        psql.execute("SELECT retrieve_model_instance_string(%d)" % mid)
        if not (probe_table is None):
            psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
            (rmse, ) = psql.fetchone()
            print "[RMSE] %f" % (rmse)
        psql.execute("UPDATE lr_model_instance SET stepsize=stepsize*%f WHERE MID=%d" % (diminish, mid))
        
    psql.close()

def usage(argv):
    print "%s: <mid> <table_name>" % (argv[0])
    
def main():
    try:
        long_opts = ["mid=", "parallel=", "epochs=", "table=", "probe=", "rank=", "stepsize=", "diminish="]
        opts, args = getopt.getopt(sys.argv[1:], "fm:p:e:t:r:b:s:d:", long_opts)
    except getopt.GetoptError:
        print str(err)
        usage(sys.argv)
        sys.exit(2)
    output  = None
    verbose = False

    mid      = None
    nthreads = 1
    epochs   = 10
    rank     = 10
    table_name = None
    probe_name = None
    stepsize   = 5e-2
    diminish   = 0.8
    B          = 1.5
    exec_fresh = True
    
    for o,a in opts:        
        if o in ("-m", "--mid"):
            mid = int(a)
        elif o == "-f":
            exec_fresh = False
        elif o in ("--parallel"):
            nthreads = int(a)
        elif o in ("-e", "--epochs"):
            epochs  = int(a)
        elif o in ("-t", "--table"):
            table_name = a
        elif o in ("-p", "--probe"):
            probe_name = a
        elif o in ("-r", "--rank"):
            rank       = int(a)
        elif o in ("-b"):
            B = float(a)
        elif o in ("-s", "--stepsize"):
            stepsize = float(a)
        elif o in ("-d", "--diminish"):
            diminish = float(a)

    if (mid is None) or (table_name is None):
        print "mid and table name are required"
        usage(sys.argv)
        sys.exit()
    else:
        print "Execute on table %s with mid %d using %d processes" % (table_name,mid, nthreads)
                                                   
    if exec_fresh:
        fresh_execute(mid, nthreads, epochs, table_name, probe_name, rank, B, stepsize, diminish)
    else:
        stored_execute(mid, nthreads, epochs, table_name, probe_name, diminish)
        
clean_tables()
main()
