from multiprocessing import Pool
import time
import sys, getopt
import random
bases = ['/scratch.1/VICTOR/','/scratch.1/chrisre/VICTOR/']
for base in bases:
    sys.path.append(base)
    sys.path.append(base + 'view_language')
    sys.path.append(base + 'factored_parser_udf')
    sys.path.append(base + 'view_language/PYTHON_UTILS/')
    
import psycopg2
import low_rank_helper


## This is the parallel execution for
## for factors.
class WorkerThread:
    def __init__(self, mid, max_rank, nRows, nCols, nSplits, connect_str,table_name):
        self.mid = mid
        self.nRows = nRows
        self.nCols = nCols
        self.nSplits = nSplits
        self.conn = psycopg2.connect(connect_str)
        self.conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        self.psql = self.conn.cursor()
        self.max_rank = max_rank
        self.table_name = table_name
        
    def take_step(self, seed, epoch, nsplits, roundid, chunkid):
        # step 1. Retrieve model and setup permtuations
        #execute(self.retrieve_model_plan, [mid])
        t0 = time.time()
        print "[take_step:%s,%s] Enter" % (roundid, chunkid)
        self.psql.execute("SELECT library_setup();")
        self.psql.execute("SELECT retrieve_memory_map(%d,%d, %d, %d,%d);" % (seed, self.mid, nsplits, roundid,chunkid))
        print "[take_step:%s,%s] Retrieved Memory Mapped Model %d for epoch %d in %s"  % (chunkid,roundid, self.mid, epoch, time.time() - t0)
        t1 = time.time()
        
        t2 = time.time()
        tname = low_rank_helper.tablename_data(self.mid, epoch, low_rank_helper.get_part(nsplits,roundid, chunkid))
        exec_str = "SELECT COUNT(*) FROM (SELECT take_step_sparse(row - 1, col - 1, rating) FROM %s) AS t;" % (tname)
        print "[take_step:%s,%s] %s" % (roundid, chunkid, exec_str)
        # Step 2. execute the step
        self.psql.execute(exec_str)
        (nCount,) = self.psql.fetchone()
        
        print "[take_step:%s,%s] Finished Step (%d tuples) in %s secs. Storing"  % (roundid,chunkid, nCount, time.time() - t2)
        t3 = time.time()
        # Step 3. execute the update
        ow = "select update_mm_segment(%d,%d, %d, %d,%d);" % (seed, self.mid, nsplits, roundid,chunkid)
        print ow
        self.psql.execute(ow)
        print "[take_step:%s,%s] Complete in %f sec (update=%s) for epoch %d"  % (roundid,chunkid, time.time() - t0, time.time() - t3,epoch)

    def close(self):
        self.psql.close()


def chunk_data(psql, seed, query_str, nsplits, epoch, mid):
    t0 = time.time()
    epoch_sign  = epoch
    table_name = "%s%d" % ("staging_data", epoch_sign)
    print "[chunk_data] Chunking the data for epoch %d which has %d sign" % (epoch+1, epoch_sign)
    exec_str = "SELECT permute_and_slice_data('chrisre', %d, '%s', %d, %d, %d);" % (seed, query_str, nsplits, epoch_sign, mid)
    print exec_str
    psql.execute(exec_str)
    print "[chunk_data] Finished Chunking in %s secs."  % (time.time() - t0)
            
class chunk_desc:
    def __init__(self, mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name):
        self.mid = mid
        self.epoch = epoch
        self.max_rank = max_rank
        self.nRows = nRows
        self.nCols = nCols
        self.nsplits = nsplits
        self.connect_str = connect_str
        self.seed = seed
        self.chunkid = chunkid
        self.roundid = roundid
        self.table_name = table_name

def chunk_maker(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, roundid,table_name):
    return [chunk_desc(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seed, chunkid, roundid, table_name) for chunkid in range(nsplits)]

                       
def perform_work(cd):
    if cd.roundid < 0:
        conn = psycopg2.connect(cd.connect_str)
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        psql = conn.cursor()
        psql.execute("SELECT library_setup();")
        qstr = "SELECT row, col, rating FROM %s" % cd.table_name
        chunk_data(psql, cd.seed, qstr, cd.nsplits, cd.epoch, cd.mid)
        psql.close()
        conn.close()
    else:
        wt = WorkerThread(cd.mid, cd.max_rank, cd.nRows, cd.nCols, cd.nsplits, cd.connect_str, cd.table_name)
        wt.take_step(cd.seed, cd.epoch, cd.nsplits, cd.roundid, cd.chunkid)
        wt.close()
    return (cd.roundid,cd.chunkid)

def install_fresh_database(psql, mid, max_rank, b, stepsize, table_name):
    print "[INIT] Installing fresh model instance to mid=%d with rank=%d b=%f stepsize=%f" % (mid, max_rank, b, stepsize)
    t0 = time.time()
    table_query = "SELECT MAX(row), MAX(col), AVG(rating) FROM %s" % table_name
    psql.execute(table_query)
    (nRows, nCols, mean,) = psql.fetchone()
    print "[INIT] model row=%d col=%d and mean computed in %s seconds" % (nRows,nCols, time.time() - t0)

    t0 = time.time()
    psql.execute("DELETE FROM lr_model_instance WHERE MID=%d;" % (mid))
    psql.execute("SELECT initialize_model(%d,%d,%d,%f,%f,%f)" % (max_rank, nRows, nCols, b, mean, stepsize))    
    psql.execute("SELECT store_model(%d);" % (mid))
    print "[INIT] Stored model mid=%d in %f seconds" % (mid, time.time() - t0)
    return (nRows, nCols)
    
def get_database_stats(psql, mid, table_name):
    print "[INIT] Getting stats about mid=%d and table=%s" % (mid, table_name)
    table_query = "SELECT MAX(row), MAX(col) FROM %s" % table_name
    psql.execute(table_query)
    (nrows, ncols) = psql.fetchone()    
    psql.execute("SELECT array_upper(L,2) as l_rank FROM lr_model_instance WHERE mid=%d LIMIT 1" % (mid))
    (max_rank,) = psql.fetchone()
    return (nrows, ncols, max_rank)
    
def fresh_execute(mid, nsplits, nepochs, table_name, probe_table, max_rank, b, initial_stepsize, diminish):
    connect_str = 'dbname=chrisre'
    conn = psycopg2.connect(connect_str)
    psql = conn.cursor()
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    psql.execute("SELECT library_setup();")
    
    (nRows, nCols) = install_fresh_database(psql, mid, max_rank, b, initial_stepsize, table_name)
    conn.commit() # force the commit
    if not (probe_table is None):
        psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
        (rmse, ) = psql.fetchone()
        print "[Initial RMSE] %f" % (rmse)
    
    qstr = "SELECT row, col, rating FROM %s" % (table_name)
    # We save the last seed
    # The way it works is in epoch -1 here, we setup the data for epoch 0
    # then in epoch 0 we need to partition the model according to the same
    # seed that was used in epoch -1. To do that, we save the seed
    seeds     = [random.randint(1,1000000) for i in xrange(nepochs)]

    # kick off the shufflers
    nshufflers       = 2 # TODO: REMOVE THIS HARD CODED
    shufflers        = Pool(processes=nshufflers)
    shuffler_results = [shufflers.apply_async(perform_work, (chunk_desc(mid,epoch,max_rank, nRows, nCols, nsplits, connect_str, seeds[epoch], -1, -1, table_name),)) for epoch in xrange(nepochs)]
    
    # Fire up the process pool
    pool      = Pool(processes=nsplits)    
    model_q = "SELECT load_mm_file('%s',%d)" % (qstr,mid)
    print "[init] %s" % (model_q)
    psql.execute(model_q)
    
    for epoch in range(nepochs):
        start_epoch = time.time()
        # generate a new seed. This is actually used for the next epoch.        
        seed = random.randint(1,1000000)        

        # Acquire the lock on the data
        start_wait = time.time()
        v = shuffler_results[epoch].get()
        end_wait   = time.time() - start_wait
        print "[EPOCH] Aquired data from shuffler. Waited for {0}.".format(end_wait)

        # kick off all the worker threads for this epoch
        for i in range(nsplits):            
            print pool.map(perform_work, chunk_maker(mid, epoch, max_rank, nRows, nCols, nsplits, connect_str, seeds[epoch], i, table_name))        
        print "[End Epoch] %s for epoch %d" % (time.time() - start_epoch, epoch)        
        epoch_time = time.time() - start_epoch
        
        # The model is now scattered in (potentially) many tables
        # reconstruct the sparse model and store it back in the model instance table

        # TODO: CLEAN UP THE MESS!! We are leaving tables all over the database

        # If we are probing then reload the model and compute the rmse
        if not (probe_table is None):
            psql.execute("SELECT read_all_mm_file(%d)" % mid)
            psql.execute("select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from %s" % probe_table)        
            (rmse, ) = psql.fetchone()
            # For convience of experiments we print the time of the epoch -- not the time of computing the RMSE
            print "[RMSE] epoch=%d time=%f rmse=%f wowait=%f" % (epoch, epoch_time,rmse,epoch_time-end_wait)

        # End of the epoch book keeeping
        psql.execute("SELECT lr_diminish(%d,%f)" % (mid,diminish))
        #if epoch == nepochs - 1:
        
    return



def usage(argv):
    print "%s: <mid> <table_name>" % (argv[0])
    
def main():
    try:
        long_opts = ["mid=", "parallel=", "epochs=", "table=", "probe=", "rank=", "stepsize=", "diminish="]
        opts, args = getopt.getopt(sys.argv[1:], "fm:p:e:t:r:b:s:d:", long_opts)
    except getopt.GetoptError:
        print str(err)
        usage(sys.argv)
        sys.exit(2)
    output  = None
    verbose = False

    mid      = None
    nthreads = 1
    epochs   = 10
    rank     = 10
    table_name = None
    probe_name = None
    stepsize   = 5e-2
    diminish   = 0.8
    B          = 1.5
    exec_fresh = True
    
    for o,a in opts:        
        if o in ("-m", "--mid"):
            mid = int(a)
        elif o == "-f":
            exec_fresh = False
        elif o in ("--parallel"):
            nthreads = int(a)
        elif o in ("-e", "--epochs"):
            epochs  = int(a)
        elif o in ("-t", "--table"):
            table_name = a
        elif o in ("-p", "--probe"):
            probe_name = a
        elif o in ("-r", "--rank"):
            rank       = int(a)
        elif o in ("-b"):
            B = float(a)
        elif o in ("-s", "--stepsize"):
            stepsize = float(a)
        elif o in ("-d", "--diminish"):
            diminish = float(a)

    if (mid is None) or (table_name is None):
        print "mid and table name are required"
        usage(sys.argv)
        sys.exit()
    else:
        print "Execute on table %s with mid %d using %d processes" % (table_name,mid, nthreads)
                                                   
    if exec_fresh:
        fresh_execute(mid, nthreads, epochs, table_name, probe_name, rank, B, stepsize, diminish)
    else:
        stored_execute(mid, nthreads, epochs, table_name, probe_name, diminish)
        
main()
