# -------------------------------------
# -- These are the standard installs
# -------------------------------------
# CREATE TABLE Victor_Models (
#     name varchar(1024),
#     objective varchar(1024), 
#     objective_item varchar(1024),
#     gradient varchar(1024),
#     prox varchar(1024),
#     model_type varchar(1024),
#     data_type varchar(1024),
#     PRIMARY KEY (name)
# );
"""
This class contains functions to build PostgreSQL string and query to insert Victor Models table.
"""

victor_models = "VICTOR_MODELS"

# -- CREATE MODEL ell_two_model (
# --  objective_item = ell_two_item
# --  model = float8[], data_type=(float8[],int)
# --  gradient=ell_two_grad, prox=gradient
# -- )
def build_string_constant(v):
    """
    Builds PostgreSQL string from the given string parameter
    @type v: string
    @param v: the string that will be used to build PostgreSQL string

    @rtype: string
    @return: returns 'v' (PostgreSQL string)
    """
    if not v:
        return "NULL"
    else:
        return ''.join(["'", v, "'"]) 

def insert_to_victor_models(model_name, objective_all, objective_item,gradient, prox,model_type, data_type ):
    """
    Constructs PostgreSQL query to insert given model to the victor models table
    @type model_name: string
    @param model_name: name of the model
    @type objective_all: string
    @param objective_all: objectivea ll
    @type objective_item: string
    @param objective_item: objective item
    @type gradient: string
    @param gradient: gradient
    @type prox: string
    @param prox: prox
    @type model_type: string
    @param model_type: type of the model
    @type data_type: string
    @param data_type: type of the data

    @rtype: string
    @return: returns insert query
    """
    v = [model_name, objective_all, objective_item, model_type, data_type, gradient, prox]
    vv = [build_string_constant(vi) for vi in v]
    vals = ','.join(vv)
    st = ["INSERT INTO ", victor_models, " VALUES (", vals, ");\n"]
    return ''.join(st)

# -- we assume this is not another table name... we should add an extra field to be safe.
# CREATE TABLE ell_two_model_instance_table (
#        model_name varchar(1024),
#        instance_name varchar(1024),
#        model_instance_id int REFERENCES Victor_global_instances,
#        data_query varchar(4096),
#        serialized_model float8[],
#        FOREIGN KEY (model_name) REFERENCES Victor_Models(name),
#        PRIMARY KEY (instance_name)
# );
