#include "postgres_util.h"

Victor_Database::Victor_Database(const char *conninfo, std::string table_name, int buf_limit) {
  this->conn = PQconnectdb(conninfo);
  this->buf = (char *) malloc(sizeof(char) * buf_limit);
  this->buf_limit = buf_limit;
  this->bytes_written = 0;
  this->end_of_last_tuple = 0;
  this->start_copy_query = "COPY " + table_name + " FROM STDIN WITH BINARY;";
  init_copy();
}

void
Victor_Database::init_copy() {
  this->copy_to_table("PGCOPY\n\377\r\n\0\0\0\0\0\0\0\0\0", 19);
  end_of_last_tuple = bytes_written;
}

void
Victor_Database::copy_field_to_table(const char *data, int length) {
  uint32_t val = htonl(length);
  this->copy_to_table((char *) &val, 4);
  this->copy_to_table(data, length);
}

void
Victor_Database::check_flush() {
  end_of_last_tuple = bytes_written;
}

void
Victor_Database::copy_to_table(const char *data, int length) {
  if(bytes_written + length > buf_limit) {
    flush();
  }
  memcpy(&(buf[bytes_written]), data, length);
  bytes_written += length;
}

void
Victor_Database::flush() {
  PGresult *res = PQexec(conn, start_copy_query.c_str());
  if (!res || (PQresultStatus(res) != PGRES_COPY_IN))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", start_copy_query.c_str(), PQerrorMessage(conn));
      PQclear(res);
    }

  int ret_val = PQputnbytes(conn, &(buf[0]), end_of_last_tuple);
  if(ret_val != 0) {
    fprintf(stderr, "problem in copy to table method during copying: %s\n", PQerrorMessage(conn));
  }
  
  ret_val = PQendcopy(conn);
  if(ret_val != 0) {
    fprintf(stderr, "problem in copy to table method when ending copying");
  }

  int remained_msg = bytes_written - end_of_last_tuple;

  if (remained_msg > 0) {
    char *temp = (char *) malloc(sizeof(char) * remained_msg);
    memcpy(temp, &(buf[end_of_last_tuple]), remained_msg);
    end_of_last_tuple = 0;
    bytes_written = 0;
    bzero(&(buf[0]), buf_limit);
    init_copy();
    memcpy(&(buf[bytes_written]), temp, remained_msg);
    bytes_written += remained_msg;
    free(temp);
  }
  else {
    end_of_last_tuple = 0;
    bytes_written = 0;
    bzero(&(buf[0]), buf_limit);
    init_copy();
  }
}
