"""
This module contains VictorFile and VictorDatabase classes
"""

from src.vsql_core.utils.factored_parser.tuple_processor import *
import psycopg2
import cStringIO, StringIO
import src.vsql_core.utils.factored_parser.mmap_wrapper as mmap_wrapper

#TODO: Convert all + concatenations to .format

"""
VictorFile implements binary read/write by using
memory-mapped files
"""
class VictorFile:
    def __init__(self, mem_map, schema, is_write):
        self.map = mem_map
        self.ts = TupleSchema(schema)
        if is_write == mmap_wrapper.WRITE_MODE:
            self.tw = TupleWriter(self.map, self.ts, True)

    """
    Reads tuples from the mmap
    @rtype: array
    @return: array of the tuples read from the buffer
    """
    def read_tuples(self):
        tuples = []
        tr = TupleReader(self.map, self.ts, True)

        r = tr.read_tuple()
        while r:
            if r:
                tuples.append(r)
            r = tr.read_tuple()

        return tuples

    """
    Writes given tuples to the mmap
    @type tuples: array
    @param tuples: array of the tuples to be written to the buffer
    """
    def write_tuples(self, tuples):
        for tup in tuples:
            self.tw.write_tuple(tup)

    """
    Writes end of stream to the mmap and flush it
    """
    def flush(self):
        self.tw.write_eos()
        ret_val = self.map.flush()
        self.map.close()

"""
VictorDatabase implements binary read/write by using
postgresql binary copy methods
"""
class VictorDatabase:
    def __init__(self, cur, schema, table_or_query_name, is_table_name):
        self.b = cStringIO.StringIO()
        self.cur = cur
        self.ts = TupleSchema(schema)
        self.tw = TupleWriter(self.b, self.ts, True)

        if is_table_name:
            self.generic_copy_query = "COPY " + table_or_query_name + " {0} WITH BINARY;"
        else:
            self.generic_copy_query = "COPY(" + table_or_query_name + ") {0} WITH BINARY;"

    """
    Reads tuples from the database
    @rtype: array
    @return: array of the tuples read from the buffer
    """
    def read_tuples(self):
        b = cStringIO.StringIO()
        copy_query = self.generic_copy_query.format("TO STDOUT")
        self.cur.copy_expert(copy_query, b)
        tuples = []
        tr = TupleReader(b.getvalue(), self.ts, True)

        r = tr.read_tuple()
        while r:
            if r:
                tuples.append(r)
            r = tr.read_tuple()

        return tuples

    """
    Writes given tuples to the database
    @type tuples: array
    @param tuples: array of the tuples to be written to the buffer
    """
    def write_tuples(self, tuples):
        for tup in tuples:
            self.tw.write_tuple(tup)

    """
    Writes end of stream to the buffer and write it to the database
    """
    def flush(self):
        self.tw.write_eos()

        copy_query = self.generic_copy_query.format("FROM STDIN")
        tmp = cStringIO.StringIO(self.b.getvalue())
        self.cur.copy_expert(copy_query, tmp)
        tmp.close()
        self.b.close()
        self.b = cStringIO.StringIO()
        
        
        

            
