"""
This class is used to generate ova plan
"""

import sys_appends

from victor_plan_execution import PlanExecutor
import generate_plan
from multiprocessing import Pool
import psycopg2
import time

########################
# Main working thread
def execute_work(work_tuple):
    """
    Execute work function
    """
    (plan_id, connect_str, setup_q, plan) = work_tuple
    plan_id = plan_id + 1
    pe = PlanExecutor(plan_id, connect_str, setup_q, plan)
    pe.go()
    pe.close()
    return plan_id



# General OVA plan
def regression_plan(connect_str, dataset,label_set, iname_table, base_mid, epochs, stepsize, diminish, dim, create_function,curried_grad, grad, obj_agg, obj_item,print_only=False, sparse_data=False, batch_plan=False, prox=None):   
    """
    Regression plan function
    """
    work_list = []
    mid       = base_mid        
    setup_query = "DELETE FROM %s where mid=%d;\n" % (iname_table, mid)        
    setup_query = setup_query + "INSERT INTO %s VALUES (%d, %f, %s(%d,%d));\n" % (iname_table, mid, stepsize, create_function, mid, dim)
    if sparse_data:
        data_query   = "SELECT label, k, v FROM %s" % (dataset)    
    else:
        data_query   = "SELECT label, vec FROM %s" % (dataset)    
    P = generate_plan.Python_Plan(mid, epochs, stepsize, diminish, iname_table, curried_grad, grad, data_query, prox, obj_agg, obj_item, sparse_data=sparse_data)
    
    if batch_plan:        
        plan = P.batchgradient_plan()
    else:
        plan = P.incremental_plan()
        
    if print_only:
        print setup_query
        print data_query
        print plan
    else:
        execute_work((0, connect_str, setup_query, plan ))


# General OVA plan
def ova_plan(connect_str, nprocesses, dataset,label_set, iname_table, base_mid, epochs, stepsize, diminish, dim, create_function,curried_grad, grad, obj_agg, obj_item,print_only=False, sparse_data=False, batch_plan=False, prox=None):   
    """
    Generates ova plan
    """
    work_list = []
    t0 = time.time()
    for i in label_set:
        # Convention is: we start at the base  mid and we make one model per class. This isn't ideal
        # and eventually we need a good (and probably heavier) mechansim
        mid       = base_mid + i        
        setup_query = "DELETE FROM %s where mid=%d;\n" % (iname_table, mid)        
        setup_query = setup_query + "INSERT INTO %s VALUES (%d, %f, %s(%d,%d));\n" % (iname_table, mid, stepsize, create_function, mid, dim)
        if sparse_data:
                data_query   = "SELECT 2*CAST((label = %d) AS INT) - 1 as label, k, v FROM %s" % (i, dataset)    
        else:
                data_query   = "SELECT 2*CAST((label = %d) AS INT) - 1 as label, vec FROM %s" % (i, dataset)    
        P    = generate_plan.Python_Plan(mid, epochs, stepsize, diminish, iname_table, curried_grad, grad, data_query, prox, obj_agg, obj_item, sparse_data=sparse_data)

        # Generate the plan
        if batch_plan:        
            plan = P.batchgradient_plan()
        else:
            plan = P.incremental_plan()
    
        if print_only:
            print setup_query
            print data_query
            print plan
        else:
            work_list.append((i, connect_str, setup_query, plan))

    pool = Pool(processes=nprocesses)            
    print pool.map(execute_work, work_list)
    tend = time.time()
    print "Completed execution of %d epochs in %f seconds" % (epochs, tend - t0)
    
