
"""
This generates plan for generic linear models
"""
 
import string
class Python_Plan:
    """The Plan Generator"""
    mid=1
    epochs=20
    nparts= 2
    step_size=1e-2
    step_diminish=0.8
    iname='instance_table'
    iname_table='iname_table'
    curried_grad='curried_grad'
    batch_grad = 'grad'
    prox = 'prox'
    obj_aggregate_function = None
    obj_row_function       = None
    sparse_data = False
    batch_step = None
    
    def __init__(self, mid, epochs, step_size, step_diminish, iname_table, curried_grad, batch_grad, data_query, prox, obj_aggregate_function, obj_row_function, sparse_data=False, batch_step=None):
        """
        Init function
        """
        self.mid = mid
        self.epochs = epochs
        self.step_size = step_size
        self.step_diminsh = step_diminish
        self.iname_table = iname_table
        self.curried_grad = curried_grad
        self.batch_grad = batch_grad
        self.data_query = data_query
        self.prox = prox
        self.nparts = 2 # CHANGE!
        self.obj_aggregate_function = obj_aggregate_function
        self.obj_row_function       = obj_row_function
        self.sparse_data = sparse_data
        self.batch_step = batch_step

    def generate_cache_model(self):
        """
        Generates cache model query

        @rtype: string
        @return: query to generate cache model
        """
        z = "SELECT retrieve_model_instance(%d, serialized_model) FROM %s where mid=%d;\n" % (self.mid, self.iname_table, self.mid)
        return z
    
    # UPDATE ell_two_model_instance_table SET serialized_model=serialize_float8(get_model_instance(mid));
    def generate_store_model(self):
        """
        Generates store model query

        @rtype: string
        @return: query to store the model
        """
        z = "UPDATE %s SET serialized_model=get_model_instance(%d,serialized_model) where mid=%d" % (self.iname_table, self.mid, self.mid)
        return z

    def generate_loss_function(self):
        """
        Generates loss function query

        @rtype: string
        @return: query to generate loss function
        """
        loss_eval = None
        if self.obj_aggregate_function and self.obj_row_function:
            if self.sparse_data:
                loss_eval = "SELECT %s(%s(%d, k,v,label)) FROM (%s) as x;\n" % (self.obj_aggregate_function, self.obj_row_function, self.mid, self.data_query)
            else:
                loss_eval = "SELECT %s(%s(%d, vec,label)) FROM (%s) as x;\n" % (self.obj_aggregate_function, self.obj_row_function, self.mid, self.data_query)
        return loss_eval
    
    def incremental_plan(self):
        """
        Generates incremental plan
        
        @rtype: tuple
        @return: incremental plan tuple that contains: header, body, loss_eval and foote
        """
        header = self.generate_cache_model()
        current_step = self.step_size
        body   = []
        for step in range(self.epochs):
            if self.prox:
                z = ["SELECT COUNT(*) FROM (SELECT ", self.prox, "(", str(self.mid), ", ", str(self.step_size), 
                     ", curried_step(", str(self.mid), ",", str(self.step_size), ", ", 
                     self.curried_grad, "(", str(self.mid), ", vec, label))) FROM (", self.data_query, 
                     ") AS T ORDER BY RANDOM()) AS T;\n"]
                body.append(string.join(z,''))
            else:
                if self.sparse_data:
                    z = "SELECT COUNT(*) FROM (SELECT %s(%d,%f,k,v,label) FROM (%s) AS i ORDER BY RANDOM()) AS T;\n" % ( self.curried_grad, self.mid, current_step, self.data_query)
                else:
                    z = "SELECT COUNT(*) FROM (SELECT %s(%d,%f,vec,label) FROM (%s) AS i ORDER BY RANDOM()) AS T;\n" % ( self.curried_grad, self.mid, current_step, self.data_query)
                body.append(z)                
            current_step = current_step * self.step_diminish

        footer = self.generate_store_model()

        loss_eval = self.generate_loss_function()


        return [header, body, loss_eval, footer]
        

    def batchgradient_plan(self):
        """
        Generates batch gradient plan
        
        @rtype: tuple
        @return: batch plan tuple that contains: header, body, loss_eval and foote
        """
        header       = self.generate_cache_model()
        header      += "SELECT zero_gradient(%d);\n" % (self.mid)
        current_step = self.step_size

        assert( self.prox ) # contract says must implement a trivial prox that zeros the gradient
        
        body = []
        for step in range(self.epochs):
            if self.sparse_data:
                z = "SELECT COUNT(*) FROM (SELECT %s(%d, k,v,label) FROM (%s) as f) as T;\n" % (self.batch_grad, self.mid, self.data_query)
            else:
                z = "SELECT COUNT(*) FROM (SELECT %s(%d, vec,label) FROM (%s) as f) as T;\n" % (self.batch_grad, self.mid, self.data_query)
                
            
            z   += "SELECT %s(%d, %f);\n" % (self.prox, self.mid,current_step)
            body.append(z)                
            current_step = current_step * self.step_diminish

        loss_eval = self.generate_loss_function()
        # generate footer
        footer = self.generate_store_model()
        return [header, body, loss_eval, footer]
    

def test_handler():
    """
    Test this class
    """

    P = Python_Plan(1, 3, 0.9, 0.8, 'ell_two_model_instance', 'c_ell_two_grad', 'full_ell2_grad_step', 'SELECT * FROM FOREST', None, 'SUM', 'full_mse_loss', batch_step='full_gradient_step')
    print "*** Incremental, No Prox"
    print P.incremental_plan()
    print "*** Batch, No Prox"
    print P.batchgradient_plan()
    
    # P = Python_Plan(1, 3, 0.9, 0.8, 'forest_ell_two', 'ell_two_model_instance_table', 'c_ell_two_grad', 'full_ell2_grad_step', 'SELECT * FROM FOREST', 'prox')
    # print "*** Incremental, Prox"
    # print P.incremental_plan()
    # print "*** Batch, Prox"
    # print P.batchgradient_plan()
    

#test_handler()
