import struct
import src.vsql_core.utils.factored_parser.factor_parser as factor_parser
import mmap, random

default_file_name = "model.dat"
sint      = struct.Struct('i')
sdouble   = struct.Struct('d')
size_of_int = 4
size_of_double = 8
class QuickReader:
    def __init__(self, m):
        self.ci = 0
        self.m  = m
        
    def reset(self):
        self.ci = 0
        
    def peek_int32(self):
        (v,) = sint.unpack_from(self.m, self.ci)
        return v
    
    def read_int32(self):
        v = self.peek_int32()
        self.ci += 4
        return v
    
    def write_int32(self, v):
        sint.pack_into(self.m, self.ci, v)
        self.ci += 4
        
    def write_double(self, d):
        sdouble.pack_into(self.m, self.ci, d)
        self.ci += 8

    def peek_double(self):
        (d,)=sdouble.unpack_from(self.m, self.ci)
        return d
    
    def read_double(self):
        d = self.peek_double()
        self.ci += 8
        return d
    
    def advance(self, x):
        self.ci += x


def get_file_name(mid):
    return "/tmp/chickens_{0}_{1}".format(mid, default_file_name)
    
def load_mm_file(dbname, mid):
    import factor_parser
    [L,R,moviemean,stepsize,b,b2] = factor_parser.retrieve_sparse_model_simple(dbname, mid)
    with open(get_file_name(mid), "w+b") as f:
        f.write('\0')
        f.seek(0)
        len_l  = len(L)
        len_r  = len(R)
        rank   = len(L[0]) # since the model is full all rows
        l_size = size_of_double*3 + size_of_int*3 + (len(L)+len(R))*len(L[0])*size_of_double
        m = mmap.mmap(f.fileno(), 0)
        m.resize(l_size)
        qr = QuickReader(m)
        for x in [moviemean,stepsize,b]:
            qr.write_double(x)

        for x in [rank, len_l, len_r]:
            qr.write_int32(x)

        for i in xrange(len_l):
            for j in xrange(rank):
                qr.write_double(L[i][j])
        
        for i in xrange(len_r):
            for j in xrange(rank):
                qr.write_double(R[i][j])
        return (len_l,len_r, rank)
    
def read_mm_file(dbname, seed, mid, nsplits, r, chunk, read_all=False):
    with open(get_file_name(mid), "r+b") as f:
        m = mmap.mmap(f.fileno(), 0)
        qr = QuickReader(m)
       
        moviemean = qr.read_double()
        stepsize  = qr.read_double()
        b         = qr.read_double()
        rank      = qr.read_int32()
        len_l     = qr.read_int32()
        len_r     = qr.read_int32()

        row_perm = range(len_l)        
        col_perm = range(len_r)
        # generate permutations
        random.seed(seed)
        random.shuffle(row_perm)
        random.shuffle(col_perm)

        l_bucket_size = len_l/nsplits
        r_bucket_size = len_r/nsplits

        import low_rank_helper
        l_id = low_rank_helper.get_l_chunk(nsplits, r, chunk)
        r_id = low_rank_helper.get_r_chunk(nsplits, r, chunk)

        
        #L = [ [0.0]*rank for i in xrange(len_l)]
        #R = [ [0.0]*rank for i in xrange(len_r)]
        L = [ [] for i in xrange(len_l)]
        R = [ [] for i in xrange(len_r)]
        for i in xrange(len_l):
            if (factor_parser.get_bucket(row_perm[i],l_bucket_size,nsplits) == l_id) or read_all:
                L[i] = [0.0]*rank
                for j in xrange(rank):
                    L[i][j] = qr.read_double()
            else:
                qr.advance(rank*size_of_double)

        
        for i in xrange(len_r):
            if (factor_parser.get_bucket(col_perm[i],r_bucket_size,nsplits) == r_id) or read_all:
                R[i] = [0.0]*rank
                for j in xrange(rank):        
                    R[i][j] = qr.read_double()
            else:
                qr.advance(rank*size_of_double)
                
        return [L,R,moviemean,stepsize,b,b*b]

def read_all(dbname,mid):
    return read_mm_file(dbname, 0, mid, 1, 0, 0, read_all=True)

def update_mm_segment(dbname, seed, mid, nsplits, r, chunk, L, R):
    with open(get_file_name(mid), "r+b") as f:
        m = mmap.mmap(f.fileno(), 0)
        qr = QuickReader(m)
       
        moviemean = qr.read_double()
        stepsize  = qr.read_double()
        b         = qr.read_double()
        rank      = qr.read_int32()
        len_l     = qr.read_int32()
        len_r     = qr.read_int32()
        assert(len_l == len(L))
        assert(len_r == len(R))

        row_perm = range(len_l)        
        col_perm = range(len_r)
        # generate permutations
        random.seed(seed)
        random.shuffle(row_perm)
        random.shuffle(col_perm)

        l_bucket_size = len_l/nsplits
        r_bucket_size = len_r/nsplits

        import low_rank_helper
        l_id = low_rank_helper.get_l_chunk(nsplits, r, chunk)
        r_id = low_rank_helper.get_r_chunk(nsplits, r, chunk)
        
        rows_updated = 0
        for i in xrange(len_l):
            if factor_parser.get_bucket(row_perm[i],l_bucket_size,nsplits) == l_id:
                rows_updated += 1
                for j in xrange(rank):
                    qr.write_double(L[i][j])
            else:
                qr.advance(rank*size_of_double)
                
        for i in xrange(len_r):
            if factor_parser.get_bucket(col_perm[i],r_bucket_size,nsplits) == r_id:
                rows_updated += 1
                for j in xrange(rank):
                    qr.write_double(R[i][j])
            else:
                qr.advance(rank*size_of_double)
        return rows_updated


def diminish_stepsize(mid, diminish):
    with open(get_file_name(mid), "r+b") as f:
        m = mmap.mmap(f.fileno(), 0)
        qr = QuickReader(m)
        moviemean = qr.read_double()
        stepsize  = qr.peek_double()
        qr.write_double(stepsize * diminish)
        
