grammar CreateModel;
options {
    output=AST;
    ASTLabelType=CommonTree;
    language=Python;

}

@header {
import sys
import traceback

from TestLexer import TestLexer
}

@main {
import SchemaHelper

def main(argv, otherArg=None):
    char_stream = ANTLRFileStream(sys.argv[1])
    lexer = TestLexer(char_stream)
    tokens = CommonTokenStream(lexer)
    parser = TestParser(tokens);

    try:
        z = parser.create_model().value
        print z
        print SchemaHelper.make_create_model_statements(z)
    except RecognitionException:
        traceback.print_stack()
}
// **********************
// CREATE MODEL INSTANCE 
// **********************
create_model_instance returns [value]: 'CREATE MODEL INSTANCE' name=ID 'OF' mname=ID we=with_examples wim=with_initial_model 
{ $value =[name.getText(), mname.getText(), we.value, wim.value] };

with_examples returns [value]: 'WITH EXAMPLES' data_view=ID 'KEY' k=ID 'DATA' attrs=attr_list
        {$value = [data_view.getText(), k.getText(), attrs.value ] } ;

with_initial_model returns [value]: 'WITH INITIAL MODEL' '(' v=(.)* ')' { $value = v.getText() };

attr_list returns [value]: {$value = []} '(' (v=ID { $value.append(v.getText()) })* ')';
// *******************
// CREATE MODEL SYNTAX
// *******************
create_model returns [value]: 'CREATE MODEL' n=ID '(' z=decls ')' ';' { $value = [["name",n.getText()]] + z.value } ;

decls returns [value] : 
    a=decl { $value = [a.value] } (',' b=decl { $value.append(b.value) } )* ;

decl returns [value]: 
      MODEL_T '=' d=typelist { $value = ["model_type", d.value] } 
    | DATA_T '=' d=typelist  { $value = ["data_type", d.value ] }  
    | FACTOR_T '=' ID        { $value = ["factor", $ID.getText() ] }
    | PROX '=' ID            { $value= ["prox", $ID.getText()] }
    | GRAD '=' ID            { $value= ["grad", $ID.getText()] }
    | OBJECTIVE '=' ID       { $value= ["objective", $ID.getText()] }
    | OBJECTIVE_ITEM '=' ID  { $value= ["objective_item", $ID.getText()] }
    ;

typelist returns [value] : 
    ID { $value= [ $ID.getText() ]; }
    | '(' z=idlist ')' { $value = z.value }
    ;

idlist returns [value] :
    v=ID { $value = [v.getText()] } (',' u=ID { $value.append(u.getText())  } )*
;

/* Lexer Rules */
MODEL_T : 'model_type' ;
DATA_T  : 'data_type' ;
FACTOR_T : 'factor' ; 
OBJECTIVE : 'objective';
OBJECTIVE_ITEM : 'objective_item';
PROX    : 'prox' ; 
GRAD    : 'grad' ;

ID         : ('a'..'z'|'A'..'Z')('a'..'z'|'A'..'Z' | '0'..'9' | '_' | '[' | ']' )+ ;
WHITESPACE : ( '\t' | ' ' | '\r' | '\n'| '\u000C' )+ 	{ $channel = HIDDEN; } ;
