# TODO: This is the start of a tool that generates the .sql portion of
#       model descriptions from the raw python model description code
#
# params =[(name, type) ... ]

"""
This module contains a function to build curry function.
"""

def build_curry_function(curry_fname, mid, fname, modulename, path, params, return_type):
    """
    This function generates the .sql portion of model descriptions from the raw python model
    description code

    @type curry_fname: string
    @param curry_fname: curry function name
    @type mid: number
    @param mid: mid
    @type fname: string
    @param fname: function name
    @type modulename: string
    @param modulename: name of the module
    @type path: string
    @param path: path to the module
    @type params: vector
    @param params: parameters of the function
    @type return_type: string
    @param return_type: return type of the SQL function

    @rtype: string
    @return: SQL function
    """
    param_types = [ t for (p,t) in params ]
    param_names = [ p for (p,t) in params ]
    param_full  = [ (p + " " + t) for (p,t) in params]
    
    param_type_str = ','.join(param_types)
    param_call_str = ','.join(param_names)
    param_full_str = ','.join(param_full)
    
    f= ""
    f+= "DROP FUNCTION IF EXISTS " + curry_fname + "(" + param_type_str + ") CASCADE;\n"
    f+= "CREATE FUNCTION " + curry_fname + "(" + param_full_str + ")\n"
    f+= "\tRETURNS " + return_type +"\n"
    f+= "AS $$\n"
    f+= "import sys\n"
    f+= "sys.append.path(path)\n"
    f+= "import " + modulename + "\n"
    f+= "model = GD[" + str(mid) + "]\n"
    f+= "return "+ modulename + "." + fname + "(" + param_call_str + ")\n"
    f+= "$$ LANGUAGE plpythonu;"
    return f
