
-------------------------------------
-- These are the standard installs
-------------------------------------
CREATE TABLE Victor_Models (
    name varchar(1024),
    objective varchar(1024), 
    objective_item varchar(1024),
    gradient varchar(1024),
    prox varchar(1024),
    model_type varchar(1024),
    data_type varchar(1024),
     PRIMARY KEY (name)
);

CREATE TABLE Victor_global_instances (
       mid SERIAL PRIMARY KEY,
       model_name varchar(1024)
       -- keeps a record of all instances
);

-- Here is an example of model creation
-- CREATE MODEL ell_two_model (
--  objective_item = ell_two_item
--  model = (float8[],float8), data_type=(float8[],int)
--  gradient=ell_two_grad, prox=gradient
-- )

insert into 
Victor_Models VALUES('ell_two_model', NULL, 'ell_two_item', 'ell_two_grad', 'gradient', '(float8[],float8)','(float8[],int)');
-- END Model Creation --
------------------------


----------------------------
-- ModelInstance Creation --
----------------------------
-- we assume this is not another table name... 
-- we should add an extra field to be safe.
CREATE TABLE ell_two_model_instance_table (
       model_name varchar(1024),
       instance_name varchar(1024),
       model_instance_id int REFERENCES Victor_global_instances,
       data_query varchar(4096),
       serialized_model_0 float8[],
       serialized_model_1 float8,
       FOREIGN KEY (model_name) REFERENCES Victor_Models(name),
       PRIMARY KEY (instance_name)
);

-- CREATE MODEL INSTANCE forest_ell_two OF ell_two_model
--   	  WITH EXAMPLES Forest KEY id DATA (vec,label)
--        WITH INITIAL MODEL ('{...}','9')
INSERT INTO Victor_global_instances(model_name) VALUES ('ell_two_model');
SELECT mid FROM Victor_global_instances WHERE model_name = 'ell_two_model';

-- generate a new model id, call it mid
-- Incremental gradient code.
-- here mid =1 and we call 
-- SELECT array_dims(vec) FROM forest LIMIT 1
INSERT INTO
ell_two_model_instance_table VALUES 
('ell_two_model', 'forest_ell_two',1, 'SELECT (vec,label) FROM FOREST', zeros(54));


--------------------------------
---- Instantiation::Execution --
--------------------------------
-- one epoch execution plan
-- mid = 1
SELECT cache_model_instance(1, serialized_model) FROM ell_two_model_instance_table where 'forest_ell_two' = instance_name;

-- incremental execution plan
SELECT COUNT(*) FROM (SELECT curried_step(1, 1e-2, ell_two_grad(get_model_instance(1), vec,label)) 
FROM FOREST ORDER BY RANDOM()) as T;

-- batch execution plan
SELECT curried_step(1, 1e-2, vec_sum(ell_two_grad(get_model_instance(1), vec,label)) FROM FOREST as T;

--UPDATE ell_two_model_table SET serialized_model=serialize_float8(get_model_instance(mid));
UPDATE ell_two_model_instance_table SET serialized_model=serialize_float8(get_model_instance(mid));
-------------------
-- END Execution --
-------------------

