-- These are the L2 bias gradients in SQL
-- create table ell_two_model_instance (mid int, serialized_model MODEL_WITH_BIAS, primary key(mid));
CREATE TYPE MODEL_WITH_BIAS AS (model float8[],bias float8);

DROP FUNCTION IF EXISTS create_initial_model_with_bias(int) CASCADE;
CREATE FUNCTION create_initial_model_with_bias(dim int)
       RETURNS MODEL_WITH_BIAS
AS $$
   return ([1e-2 for i in range(dim)], 1e-2)
$$ LANGUAGE plpythonu;

DROP FUNCTION IF EXISTS cache_model_instance(int, MODEL_WITH_BIAS) CASCADE; 
CREATE FUNCTION cache_model_instance(id int, m MODEL_WITH_BIAS)
       RETURNS VOID
AS $$ 
GD[id] = (m["model"], m["bias"])
$$ LANGUAGE plpythonu;

-- signature is this way since there is no return-type covariance
DROP FUNCTION IF EXISTS get_model_instance(int, MODEL_WITH_BIAS) CASCADE; 
CREATE FUNCTION get_model_instance(id int, unused MODEL_WITH_BIAS)
       RETURNS MODEL_WITH_BIAS
AS $$ 
(model,b) = GD[id]
return (model, b)
$$ LANGUAGE plpythonu;


DROP FUNCTION IF EXISTS ell_two_bias_grad(int, x float8[] , label float8 ) CASCADE; 
CREATE FUNCTION ell_two_bias_grad(id int, x float8[], label float8)
       RETURNS MODEL_WITH_BIAS
AS $$
import victor_utils
(w,b) = GD[id]
return victor_utils.grad(w,b,x,label)
$$ LANGUAGE plpythonu;


DROP FUNCTION IF EXISTS curried_bias_grad_g(int, float, float8[], float8) CASCADE; 
CREATE FUNCTION curried_bias_grad_g(id int, step float, x float8[], y float8)
       RETURNS VOID
AS $$ 
import victor_utils
(w,b)   = GD[id]
(dw,db) = victor_utils.grad(w,b,x,y)
w       = victor_utils.scale_and_add(w, dw,-step)
b       = b - (step * db)
GD[id]  = (w,b)
$$ LANGUAGE plpythonu;
