-- These are the L2 bias gradients in SQL
-- create table ell_two_model_instance (mid int, serialized_model MODEL_WITH_BIAS, primary key(mid));
CREATE TYPE LINEAR_MODEL AS (model float8[]);
CREATE TYPE SPARSE_PAIR  AS (k int ,v int);

DROP FUNCTION IF EXISTS create_initial_model(int) CASCADE;
CREATE FUNCTION create_initial_model(dim int)
       RETURNS LINEAR_MODEL
AS $$
   return ([1e-2 for i in range(dim)], 1e-2)
$$ LANGUAGE plpythonu;

DROP FUNCTION IF EXISTS cache_model_instance(int, LINEAR_MODEL, mu double) CASCADE; 
CREATE FUNCTION cache_model_instance(id int, m LINEAR_MODEL, mu double)
       RETURNS VOID
AS $$ 
GD[id] = (m["model"], mu)
$$ LANGUAGE plpythonu;

-- signature is this way since there is no return-type covariance
DROP FUNCTION IF EXISTS get_model_instance(int, LINEAR_MODEL) CASCADE; 
CREATE FUNCTION get_model_instance(id int, unused LINEAR_MODEL)
       RETURNS LINEAR_MODEL
AS $$ 
(model,mu) = GD[id]
return model
$$ LANGUAGE plpythonu;


DROP FUNCTION IF EXISTS logit_l1_sparse(int, float, sparse_pair[], float8) CASCADE; 
CREATE FUNCTION logit_l1_sparse(id int, step float, x sparse_pair[], y float8)
       RETURNS VOID
AS $$ 
import victor_utils
(w,mu)  = GD[id]
wx      = victor_utils.dot_ds(w,x)
sig     = victor_utils.sigma(-wx*y)
# NB: derivative is - y*x sigma
victor_utils.scale_and_add_ds(w,wx,y*sig)
victor_utils.l1_srunk(w, mu*step)
GD[id]  = (w,mu)
$$ LANGUAGE plpythonu;
