-- crappy python dot product
DROP FUNCTION IF EXISTS py_dot(float8[],float8[]) CASCADE; 
CREATE FUNCTION py_dot(x float8[], y float8[])
       RETURNS float8
AS $$
n = min(len(x),len(y))
ret = 0.0
for i in range(n):
    ret = ret + x[i]*y[i]
return ret
$$ LANGUAGE plpythonu;

DROP FUNCTION IF EXISTS vector_add(float8[],float8[]) CASCADE; 
CREATE FUNCTION vector_add(x float8[], y float8[])
RETURNS float8[]
AS $$
if len(x) == 0:
    return y

if len(y) == 0:
    return x
n = min(len(x),len(y))
ret = 0.0
return [x[i] + y[i] for i in range(n)]
$$ LANGUAGE plpythonu;

DROP AGGREGATE vector_sum(float8[]);
CREATE AGGREGATE vector_sum(float8[]) (
    sfunc=vector_add,
    stype=float8[],
    initcond='{}'
);

-- the step function
-- takes a model, a step size, a gradient and produces a model
DROP FUNCTION IF EXISTS step(float8[],float8,float8[]) CASCADE; 
CREATE FUNCTION step(model float8[],step float8,grad float8[])
       RETURNS float8[]
AS $$
n = min(len(model),len(grad)) 
return [ model[i] - step * grad[i] for i in range(n)]
$$ LANGUAGE plpythonu;


-- the curried step function
-- assumes the model is cached in GD[modelid]
-- takes a model, a step size, a gradient and produces a model
DROP FUNCTION IF EXISTS curried_step(int, float8,float8[]) CASCADE; 
CREATE FUNCTION curried_step(modelid int, step float8,grad float8[])
       RETURNS VOID
AS $$
model = GD[modelid]
n = min(len(model),len(grad)) 
GD[modelid] = [model[i] - step * grad[i] for i in range(n)]
$$ LANGUAGE plpythonu;


--- This will eventually be a family of instances cache_
--- caches model instance
DROP FUNCTION IF EXISTS cache_model_instance(int, float8[]) CASCADE; 
CREATE FUNCTION cache_model_instance(modelid int, model float8[])
       RETURNS VOID
AS $$
GD[modelid] = model
$$ LANGUAGE plpythonu;

DROP FUNCTION IF EXISTS cache_model_instance(int, float8[],float8) CASCADE; 
CREATE FUNCTION cache_model_instance(modelid int, model1 float8[], model2 float8)
       RETURNS VOID
AS $$
GD[modelid] = (model1,model2)
$$ LANGUAGE plpythonu;



-- gets a model instance
DROP FUNCTION IF EXISTS get_model_instance(int) CASCADE; 
CREATE FUNCTION get_model_instance(modelid int)
       RETURNS float8[]
AS $$
return GD[modelid] 
$$ LANGUAGE plpythonu;

-- TODO: store and retreive the model
-- This would write it back to its correct location in the database


-- the step function
-- takes a model, a step size, a gradient and produces a model
DROP FUNCTION IF EXISTS zeros(int) CASCADE; 
CREATE FUNCTION zeros(n int)
       RETURNS float8[]
AS $$
return [0 for i in range(n)]
$$ LANGUAGE plpythonu;

-- crappy python dot product
DROP FUNCTION IF EXISTS eval_squares(int, float8[],float) CASCADE; 
CREATE FUNCTION eval_squares(x int, w float8[], b float)
       RETURNS float8
AS $$
model = GD[x]
n = min(len(w), len(model))
ret = sum([model[i]*w[i] for i in range(n)])
return (ret - b)**2
$$ LANGUAGE plpythonu;

--- Some permutation helper functions
DROP FUNCTION IF EXISTS create_perm(int) CASCADE;
CREATE FUNCTION create_perm(n int) RETURNS VOID
AS $$
   import random
   model = range(n)
   random.shuffle(model)
   GD['perm'] = model 
$$ LANGUAGE plpythonu;

--- Some permutation helper functions
DROP FUNCTION IF EXISTS apply_perm(int) CASCADE;
CREATE FUNCTION apply_perm(i int) RETURNS int
AS $$
   return GD['perm'][i]
$$ LANGUAGE plpythonu;


