-----------------------------------------------------------------
-- This is the file that describes a linear model with robust average
-- The model contains a current model 
--  model:float8[]
--  w_hat:float8[] (the averaged model)
--  steps:int      (to compute the moving average)
-- This type of model is used by many different loss functions.
--
-----------------------------------------------------------------
DROP TYPE IF EXISTS LINEAR_MODEL_WITH_AVERAGE CASCADE;
CREATE TYPE LINEAR_MODEL_WITH_AVERAGE AS (model float8[], w_hat float8[], steps int);

-- Create the HOME table for this type.
-- TODO: Make this automatically generated.
DROP TABLE IF EXISTS linear_model_instance;
CREATE TABLE linear_model_instance(mid int, stepsize float8 , serialized_model linear_model_with_average, primary key(mid));

-- Create the linear model and store it in the global dictionary
DROP FUNCTION IF EXISTS create_linear_model(int, int) CASCADE;
CREATE FUNCTION create_linear_model(id int, dim int)
       RETURNS LINEAR_MODEL_WITH_AVERAGE
AS $$
w = [1e-2 for i in range(dim)]
what = w
steps = 0
GD[id] = (w,what,steps,1e-5, 150.0)
return (w,what,steps)
$$ LANGUAGE plpythonu;

-----------------------------------------------------------------------
-- to help with the dispatch (polymorphism) we add in an extra field  -
-- TODO: Make this binary
-----------------------------------------------------------------------
DROP FUNCTION IF EXISTS retrieve_model_instance(int, LINEAR_MODEL_WITH_AVERAGE) CASCADE; 
CREATE FUNCTION retrieve_model_instance(id int, LINEAR_MODEL_WITH_AVERAGE)
       RETURNS VOID
AS $$     
rvs = plpy.execute("SELECT stepsize, serialized_model FROM linear_model_instance WHERE mid=%d" % id)
model = rvs[0]["serialized_model"]
(model_str, what_str, step_str) = model.split('",')
# Parse the model string
model_str = model_str.lstrip('("{')
model_str = model_str.rstrip('}')
w  = [float(x) for x in model_str.split(',')]
# what_str
what_str  = what_str.lstrip('("{')
what_str  = what_str.rstrip('}')
w_hat     = [float(x) for x in model_str.split(',')]
# stepsize
step_str  = step_str.rstrip(')')
steps = int(step_str)

step  = rvs[0]["stepsize"]
plpy.info(model_str)
GD[id] = (w, w_hat, steps, 1e-5, 150.0)
$$ LANGUAGE plpythonu;

-------------------------------------------------------------------------
-- signature is this way since there is no return-type covariance
-------------------------------------------------------------------------
DROP FUNCTION IF EXISTS get_model_instance(int, LINEAR_MODEL_WITH_AVERAGE) CASCADE; 
CREATE FUNCTION get_model_instance(id int, LINEAR_MODEL_WITH_AVERAGE)
       RETURNS LINEAR_MODEL_WITH_AVERAGE
AS $$ 
(w,what,steps,mu,B) = GD[id]
return (w,what,steps)
$$ LANGUAGE plpythonu;


