

-- This fucntion takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS take_step_verbose(int, int, float8) CASCADE; 
CREATE FUNCTION take_step_verbose(i int, j int, rating float8)
RETURNS VOID
AS $$
   import low_rank_helper
   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   plpy.info("model=%s - %s" % (L,R))
   err = low_rank_helper.dot(L[i],R[j]) + moviemean - rating	
   e   = -(stepsize * err)
   tempLi = list(L[i])
   low_rank_helper.scale_and_add(tempLi, R[j],e)   			    
   low_rank_helper.scale_and_add(R[j],   L[i],e)
   L[i] = tempLi
   low_rank_helper.ball_project(L[i],B,B2)
   low_rank_helper.ball_project(R[j],B,B2)
   plpy.info("model=%s - %s" % (L,R))
   GD['lr_model'] =    [L,R,moviemean,stepsize,B,B2] 
$$ LANGUAGE plpythonu;


---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
-- This code is used at the end of a (parallel) update in V-SQL.
-- It should be executed transactionally (else updates may be lost)
-- It reads in the model, updates its fragment in memory and writes it back to disk.
-- The fragment is computed from (chunk,round)
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
DROP FUNCTION IF EXISTS overwrite_fragment_debug(int, int, int, int, int,int,int) CASCADE; 
CREATE FUNCTION overwrite_fragment_debug(max_rank int, nsplits int, nrows int, ncols int, chunk int, round int, mid int)
RETURNS VOID
AS $$   

   row_bucket_size = nrows/nsplits	
   col_bucket_size = ncols/nsplits	
 
   assert(row_bucket_size > 0 and col_bucket_size > 0)   

   col_bucket      = chunk
   col_low         = chunk*col_bucket_size
   col_high        = min((chunk+1)*col_bucket_size, ncols) - 1
   
   row_bucket      = (- (round - col_bucket)) % nsplits   
   row_low         = row_bucket*row_bucket_size
   row_high        = min((row_bucket+1)*row_bucket_size, nrows) - 1
   
   [L_stored,R_stored, mean, stepsize, b, z] = GD['lr_model']	


   # build strings for insert
   L_str = "'{" + ','.join(["{" + ','.join([str(y) for y in l]) + "}" for l in L_stored[row_low:(row_high+1)]]) + "}'"
   R_str = "'{" + ','.join(["{" + ','.join([str(y) for y in r]) + "}" for r in R_stored[col_low:(col_high+1)]]) + "}'"

   # +1 to compensate for postgresql indexing   
   qstr = "UPDATE lr_model_instance SET L[%d:%d][1:%d]=%s, R[%d:%d][1:%d]=%s WHERE mid=%d" % (row_low + 1, row_high + 1, max_rank, L_str, col_low + 1, col_high + 1, max_rank, R_str,mid)
   plpy.info(qstr)
   plpy.execute(qstr)
$$ LANGUAGE plpythonu;


