
-- This fucntion takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS update_segment_sparse(int, int, int, int, int) CASCADE; 
CREATE FUNCTION update_segment_sparse(mid int, epoch_sign int, nsplits int, round int, chunk int)
RETURNS VOID
AS $$
   import low_rank_helper
   import os
   import psycopg2
   import StringIO

   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   dbname = 'dbname=test_victor' # REMOVE THIS!
   
   l_sparse = [(i,L[i]) for i in range(len(L)) if len(L[i]) > 0]
   l_string = '\n'.join(["%d\t%s" % (i, low_rank_helper.build_array_string(li)) for (i, li) in l_sparse])

   r_sparse = [(i,R[i]) for i in range(len(R)) if len(R[i]) > 0]
   r_string = '\n'.join(["%d\t%s" % (i, low_rank_helper.build_array_string(ri)) for (i, ri) in r_sparse])


   l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))   
   r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))

   psql = psycopg2.connect(dbname)		  
   psql.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
   c    = psql.cursor()
   
   c.execute(("DELETE FROM %s;" % l_name))
   c.execute(("DELETE FROM %s;" % r_name))
   c.copy_from(StringIO.StringIO(l_string), l_name)
   c.copy_from(StringIO.StringIO(r_string), r_name)
   psql.close()
$$ LANGUAGE plpythonu;



--------------------------
-- Compute the objective
-- mean-squared error
DROP FUNCTION IF EXISTS retrieve_sparse_model(int, int, int, int, int) CASCADE; 
CREATE FUNCTION retrieve_sparse_model(mid int, epoch_sign int, nsplits int, round int, chunk int)
RETURNS VOID
AS $$
   import low_rank_helper
   v_plan = plpy.prepare("SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank, moviemean as moviemean, stepsize as stepsize, b as b FROM lr_model_instance WHERE mid=$1", ["int"]) 
   GD['sparse_model_info'] = (mid,epoch_sign, nsplits, round, chunk)

   rv = plpy.execute(v_plan, [mid], 1)
   nRows     = rv[0]["nrows"]
   l_rank    = rv[0]["l_rank"]
   nCols     = rv[0]["ncols"]
   r_rank    = rv[0]["r_rank"]
   stepsize  = rv[0]["stepsize"]
   moviemean = rv[0]["moviemean"]
   b         = rv[0]["b"]


   l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))
   plpy.info("Looking for table %s " % (l_name))	  
   L = [[] for i in range(nRows)]
   rvs    = plpy.execute("SELECT id, vec from %s" % (l_name));
   for rv in rvs:
       id = rv["id"]
       vec = rv["vec"]
       L[id] = vec
   
   r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))
   plpy.info("Looking for table %s " % (r_name))	  
   R = [[] for i in range(nCols)]
   rvs    = plpy.execute("SELECT id, vec from %s" % (r_name));
   for rv in rvs:
       id = rv["id"]
       vec = rv["vec"]
       R[id] = vec
   
   GD['lr_model'] = [L,R,moviemean,stepsize, b, b*b]

$$ LANGUAGE plpythonu;


-- This fucntion takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS take_step_sparse(int, int, float8) CASCADE; 
CREATE FUNCTION take_step_sparse(i int, j int, rating float8)
RETURNS VOID
AS $$
   import low_rank_helper
   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   if (len(L[i]) == 0):
      (mid,epoch_sign, nsplits, round, chunk) = GD['sparse_model_info']
      plpy.error("Zero length: L index %d (%d,%d) from %s!" % (i, round, chunk, low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))))

   if (len(R[j]) == 0):
      (mid,epoch_sign, nsplits, round, chunk) = GD['sparse_model_info']
      plpy.error("Zero length: R index %d (%d,%d) from %s!" % (j, round, chunk, low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))))

   err = low_rank_helper.dot(L[i],R[j]) + moviemean - rating	
   e   = -(stepsize * err)
   tempLi = list(L[i])
   low_rank_helper.scale_and_add(tempLi, R[j],e)   			    
   low_rank_helper.scale_and_add(R[j],   L[i],e)
   L[i] = tempLi
   low_rank_helper.ball_project(L[i],B,B2)
   low_rank_helper.ball_project(R[j],B,B2)
   GD['lr_model'] =    [L,R,moviemean,stepsize,B,B2] 
$$ LANGUAGE plpythonu;


-- This function takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS sparse_rebuild_model(int, int, int) CASCADE; 
CREATE FUNCTION sparse_rebuild_model(mid int, epoch_sign int, nsplits int)
RETURNS VOID
AS $$
   import low_rank_helper
   l_body = [ "(SELECT id,vec FROM %s)" % (low_rank_helper.tablename_lmodel(mid, epoch_sign, i)) for i in range(nsplits)]
   l_query = "SELECT id,vec FROM (%s) as t ORDER BY id;" % (' UNION '.join(l_body))
   #plpy.info(l_query)
   lrvs = plpy.execute(l_query)
   L_str = low_rank_helper.build_array_string([ low_rank_helper.build_array_string(lrv["vec"]) for lrv in lrvs ])
   plpy.info("Rebuild one")

   r_body = [ "(SELECT id,vec FROM %s)" % (low_rank_helper.tablename_rmodel(mid, epoch_sign, i)) for i in range(nsplits)]
   r_query = "SELECT id,vec FROM (%s) as t ORDER BY id;" % (' UNION '.join(r_body))
   rrvs = plpy.execute(r_query)
   plpy.info(r_query)
   R_str = low_rank_helper.build_array_string([ low_rank_helper.build_array_string(rrv["vec"]) for rrv in rrvs ])

   #plpy.info("update")
   plpy.execute("UPDATE lr_model_instance SET L='%s', R='%s' WHERE mid=%d" % (L_str, R_str, mid))
$$ LANGUAGE plpythonu;


