#! /usr/bin/env python
import sys
import traceback
import antlr3
import antlr3.tree
from src.parser.VictorParserLexer import VictorParserLexer
from src.parser.VictorParserParser import VictorParserParser
import src.globals.constants as constants
import src.parser.ParserTypes as ParserTypes
import src.vsql_core.plan_generators.plan_generator as planner
import src.normalization.normalization_front as normalize
import src.normalization.create_model as create_model
import collections, multiprocessing, multiprocessing.sharedctypes
import cPickle, cStringIO
import src.utils.shared_storage as shared_storage

from multiprocessing import Pool
import src.globals.global_utils as global_utils, time, random, math, array
import main_memory_simple_exec as mm_se
import factor_simple_numpy_hack
import numpy
JOB_SPEC = collections.namedtuple('job_spec', 'init_function stop_condition grad_function loss_function partition_function data_query model_type data_type')
### THESE ARE EXPLORATORY HACKS
crazy_hack = False
pure_hack  = True


row_data    = None
col_data    = None
rating_data = None
grad_fn     = None
part_fn     = None
loss_fn     = None
perm        = None

def execute_grad( mid, nItems, r, c, round_map, chunk_map, item_map, li ):
    (model, model_types) = shared_storage.get_cache_model(mid) 
    last_index = li
    while last_index > 0:
        i = last_index
        # assert( r == round_map[i] and c == chunk_map[i] )
        grad_fn(model, row_data[i], col_data[i], rating_data[i])
        last_index = item_map[i]


            
def compute_loss( (mid, l, nParts, part_id) ):
    (model, model_types) = shared_storage.get_cache_model(mid)
    loss = 0.0
    for i in xrange(0,l,nParts):
        pi = i + part_id
        if(pi < l):
            loss += loss_fn(model, row_data[pi], col_data[pi], rating_data[pi])
    return loss

## Generic Routine
def load_into_shared_memory(sql_data, type_string):
    shared_segments = [ multiprocessing.sharedctypes.Array(typ, map(lambda x: x[i], sql_data, lock=False) ) for (i,typ) in enumerate(type_string) ]
        
## FACTOR EXECUTE HACK
def main_memory_factor_executor(mid, js, nParts=4):
    global row_data, col_data, rating_data, grad_fn, loss_fn, perm

    ss       = shared_storage.start_server()
    ss.create_segment("epoch", 0)


    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    t0       = time.time()
    cur.execute(js.data_query)

    sql_data    = cur.fetchall()
    nItems   = len(sql_data)
    
    row_data    = multiprocessing.sharedctypes.Array('i', [row-1 for (row,col,rating,) in sql_data], lock=False)
    col_data    = multiprocessing.sharedctypes.Array('i', [col-1 for (row,col,rating,) in sql_data], lock=False)
    rating_data = multiprocessing.sharedctypes.Array('f', [rating for (row,col,rating,) in sql_data], lock=False)

    
    nRows       = max(row_data) + 1 
    nCols       = max(col_data) + 1
    nSplits     = nParts

    round_map   = multiprocessing.sharedctypes.Array('i', range(nItems))
    chunk_map   = multiprocessing.sharedctypes.Array('i', range(nItems))
    item_map    = multiprocessing.sharedctypes.Array('i', range(nItems))
    print "Data fetched in {}".format(time.time() - t0)

    state    = dict()

    loss     = 1e100
    perm     = multiprocessing.Array('i', range(nItems), lock=False)
    
    # Now ground teh functions
    model    = mm_se.ground_function(js.init_function)()
    grad_fn  = mm_se.ground_function(js.grad_function)
    loss_fn  = mm_se.ground_function(js.loss_function)
    stop_fn  = mm_se.ground_function(js.stop_condition)
    part_fn  = mm_se.ground_function(js.partition_function)

    shared_storage.store_cache_model(mid, model, js.model_type)

    #####################################################################
    # To see the shared data properly, we need to setup the pool after  #
    # allocating the shared data                                        #
    #####################################################################
    
    pool     = Pool(processes=nParts)
    (model,model_types) = shared_storage.get_cache_model(mid) 
    del sql_data
    epoch = 1
    # these do not change
    # This factoring code does not work correctly, it is 
    row_perm     = multiprocessing.sharedctypes.Array('i', range(nRows))
    col_perm     = multiprocessing.sharedctypes.Array('i', range(nCols))
    loss_jobs    = [ (mid, nItems, nParts, part_id) for part_id in range(nParts)]
    while not stop_fn(state, loss):
        t0 = time.time()
        random.shuffle(perm)
        random.shuffle(row_perm)
        random.shuffle(col_perm)
        print "Done with Shuffle {}".format(time.time() - t0)
        
        last_index = [ array.array('i', [-1 for r in range(nSplits)]) for c in range(nSplits) ]
        if crazy_hack:
            last_index = factor_simple_numpy_hack.quick_partition_two(perm, row_perm, row_data, col_perm, col_data, round_map, chunk_map, item_map, nSplits, nRows, nCols, nItems)
        else:
            for i in xrange(nItems):
                pi  = perm[i]
                r0  = row_data[pi]
                r   = row_perm[r0]
                
                c0  = col_data[pi]
                c   = col_perm[c0]
                row_bucket = min(int(r/(nRows/nSplits)),nSplits-1)
                col_bucket = min(int(c/(nCols/nSplits)),nSplits-1)
                
                round_map[pi] = row_bucket
                chunk_map[pi] = col_bucket
                item_map[pi]  = last_index[row_bucket][col_bucket]
                last_index[row_bucket][col_bucket] = pi

        
        # j  = part_fn(state, model, 10, nParts*nParts, { "row" : row_data[pi], "col" : col_data[pi], "rating" : rating_data[pi] }  )
        # partitions[j].append(pi)
        print "Done with Partition {}".format(time.time() - t0)

        t1 = time.time()
        for r in xrange(nParts):
            if crazy_hack:
                ps  = [ multiprocessing.Process(target=factor_simple_numpy_hack.execute_grad,
                                                args=(mid,row_data, col_data, rating_data, nItems, r,c, round_map, chunk_map, item_map, last_index[r][c])) for c in xrange(nParts)]
            else:
                ps  = [ multiprocessing.Process(target=execute_grad, args=(mid,nItems, r,c, round_map, chunk_map, item_map, last_index[r][c])) for c in xrange(nParts)]
            tmp = [ p.start() for p in ps ]
            ts  = [ p.join() for p in ps ]
            print "\t Round complete {} of {}".format(r, nParts)
            
        #pool.map(execute_grad, grad_jobs)
        print "Done with grad step {}".format(time.time() - t1)

        loss    = 0.0
        losses  = pool.map(compute_loss, loss_jobs)
        loss    = sum(losses)
        print "Done loss={} time={}".format(math.sqrt(loss)/math.sqrt(nItems), time.time() - t0)
        
        epoch += 1
        ss.update_segment("epoch", epoch)


def shuffler(perm_ref):
    random.shuffle(perm_ref)

def fetch_model_information(model_name):
    grad_loss_str = """SELECT grad_function_str, loss_function_str, model_field_types, data_field_types
    FROM victor_model_table where model_name = '{}'""".format(model_name)
    print grad_loss_str
    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    cur.execute(grad_loss_str)
    (grad_str, loss_str, model_str, data_str) = cur.fetchone()
    grad = ParserTypes.PythonFunction(grad_str.split('.'))
    loss = ParserTypes.PythonFunction(loss_str.split('.'))
    model_type = tuple(model_str)
    data_type  = tuple(data_str)
    print " {} {} ".format(model_type, data_type)
    conn.close()

    return (grad,loss, model_type, data_type)

def main_memory_create_model_instance(param_dict):
    model_instance_name      = param_dict[constants.PARSER_NAME_FIELD]
    example_table_name       = param_dict[constants.PARSER_EX_FIELD]
    example_fields           = param_dict[constants.PARSER_EX_TYPE_LIST]
    model_name               = param_dict[constants.PARSER_MODEL_SPEC_FIELD]
    init_fnc                 = param_dict[constants.PARSER_INIT_BLOCK]
    stop_condition           = param_dict[constants.PARSER_STOP_CONDITION_NAME]
    partition_function       = create_model.optional_param(param_dict, constants.PARSER_PARTITION_FUNCTION_NAME)

    print example_table_name, example_fields
    q_str = "SELECT {field_names} FROM {example_table_name}".format(field_names=','.join(example_fields),
                                                                  example_table_name=example_table_name)
    print q_str
    factor_model = partition_function is not None
    (grad,loss,model_type, data_type)=fetch_model_information(model_name)

    job_spec = JOB_SPEC(init_function=init_fnc, 
                        stop_condition=stop_condition, 
                        grad_function=grad,
                        loss_function=loss,
                        partition_function=partition_function,
                        data_query=q_str, 
                        model_type=model_type,
                        data_type=data_type)
    mid = 10
    if factor_model:
        print "Factor model"
        main_memory_factor_executor(mid, job_spec)
    else:
        mm_se.main_memory_line_executor(mid, job_spec)
        
def parse_phase(fname):
    char_stream = antlr3.ANTLRFileStream(fname)
    lexer  = VictorParserLexer(char_stream)
    tokens = antlr3.CommonTokenStream(lexer)
    parser = VictorParserParser(tokens)
    
    return parser.top_level().value

def is_model_spec(t):
    return isinstance(t, ParserTypes.ParserModelInstance)

def main(fname):
    print "Parsing {0}...".format(fname)
    tasks = parse_phase(fname)
    good_tasks = filter(is_model_spec,tasks)
    create_tasks = filter(lambda x: not (is_model_spec(x)),tasks)
    for ct in create_tasks:
        ct.normalize()
        
    for g in good_tasks:
        main_memory_create_model_instance(g.params)

if len(sys.argv) == 2:
    main(sys.argv[1])
else:
    print "{0} <specification_file>".format(sys.argv[0])
