import collections, multiprocessing
import src.globals.global_utils as global_utils, time, random, math, array
import src.utils.shared_storage as shared_storage
import src.vsql_core.utils.pickler.victor_pickler as vp
import numpy

perms      = None
nShufflers = 2
def unpack_data( typ , v ):
    if (typ == "python") or (typ == "numpy"):
        return vp.unpickle_data(v)
    elif typ == "npython":
        return numpy.array(vp.unpickle_data(v))
    else:
        return v
    
def shuffler_job(perms, epoch):
    t0    = time.time()
    random.shuffle(perms[(epoch + 1) % (nShufflers + 1)] )
    print "Done with Shuffle {}".format(time.time() - t0)

def ground_function(py_fn):
    exec py_fn.get_import_statement()
    # assert(perm is not None)
    return eval(py_fn.get_qualified_name(), globals(), locals())

def main_memory_line_executor(mid, js, nParts   = 12):
    ss       = shared_storage.start_server()
    ss.create_segment("epoch", 0)


    conn     = global_utils.connect_db()
    cur      = conn.cursor()
    t0       = time.time()

    cur.execute(js.data_query)
    sql_data    = cur.fetchall()

    print "Data fetched in {}".format(time.time() - t0)
    t1        = time.time()

    # Should we convert these to a column layout? Just to make the indexing and deconstruction time lower
    data      = [ tuple([unpack_data(js.data_type[i],v) for (i,v) in enumerate(row)])  for row in sql_data  ]
    print "Unpickled in {}".format(time.time() - t1)
        
    state    = dict()

    loss     = 1e100
    nItems   = len(sql_data)
    perms    = [ multiprocessing.Array('i', range(nItems), lock=False) for i in xrange(nShufflers + 1) ]

    ######################################################
    # Ground out the functions so that we can use them.
    #####################################################
    model    = ground_function(js.init_function)()
    grad_fn  = ground_function(js.grad_function)
    loss_fn  = ground_function(js.loss_function)
    stop_fn  = ground_function(js.stop_condition)

    def worker_job(epoch):
        perm     =  perms[epoch % (nShufflers + 1)]
        t0 = time.time()
        for i in xrange(nItems):
            v = data[perm[i]]
            grad_fn(model, v)        
        print "Done with grad step {}".format(time.time() - t0)

    def loss_job(epoch):
        if(epoch < 1):
            return 1e100
        
        loss    = 0.0
        for i in xrange(nItems):
            v     = data[i]
            loss += loss_fn(model, v)
        print "Done loss={} time={}".format(math.sqrt(loss)/math.sqrt(nItems), time.time() - t0)
        return loss
    
    #####################################################################
    # To see the shared data properly, we need to setup the pool after  #
    # allocating the shared data                                        #
    #####################################################################
    pool     = multiprocessing.Pool(processes=3)
    epoch    = 0
    shuffler_job(perms, epoch)
    while not stop_fn(state, loss):
        epoch += 1
        ss.update_segment("epoch", epoch)
        
        ps = [ multiprocessing.Process(target=f, args=a) for (f,a) in [(shuffler_job, (perms,epoch)), (loss_job, (epoch,))]]

        for p in ps:
            p.start()
            
        worker_job(epoch)  

        for p in ps:
            p.join()
