#include "simple_grad.h"
#include <cassert>
double
dot(double *x, double *y, int nSize) {
  double ret = 0.0;
  for(int i = nSize - 1; i >= 0; i--) {ret += x[i]*y[i];}
  return ret;
}

void 
add_and_scale(double *x, int size, double *y, double c) {
  for(int i = size - 1; i >= 0; i--) {
    x[i] += y[i]*c;
  }
}

LinearModel::
LinearModel(int _nSize) : nSize(_nSize) 
{ data = new double[nSize];
  for(int i = nSize - 1; i >= 0; i--) {data[i] = 1e-2;}
  stepsize = 0.1;
  nSteps = 1;
}

void
LinearModel::
grad_step( double *x, int size, int y) {
  assert(size == nSize);
  double wx  = dot(data,x,nSize);
  double err = (wx - y);
  double step = stepsize / (1 + nSteps);
  double etd = - step * err;
  nSteps++;
  add_and_scale(data, size, x, etd);  
}

double
LinearModel::
se_loss(double *x, int size, int y) {
  assert(size == nSize); 
  double wx  = dot(data,x,nSize);
  double err = (wx - y);
  return err*err;
}
