import numpy

cdef extern from "numpy/arrayobject.h":
    ctypedef int intp
    ctypedef extern class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef int nd
        cdef intp *dimensions
        cdef intp *strides
        cdef int flags
        
cdef extern from "simple_grad.h":
    cdef cppclass LinearModel:
        LinearModel(int)
        void grad_step(double*, int, int)
        double se_loss(double*, int, int)

cdef class LinearModelHolder:
    cdef LinearModel *m
    def __cinit__(self, int n):
        self.m = new LinearModel(n)

    def __dealloc__(self):
        del self.m

    def grad_step(self, ndarray a, int label):
        self.m.grad_step(<double*> a.data, a.dimensions[0], label)

    def se_loss(self, ndarray a, int label):
        return self.m.se_loss(<double*> a.data, a.dimensions[0], label)


cdef class PyrexModelHolder:
    cdef int nSteps
    cdef ndarray m
    cdef double stepsize
    
    def __init__(self, int n):
        self.m = numpy.array([1e-2 for i in range(n)])
        self.nSteps = 1
        
    def grad_step(self, ndarray x, int y):
        cdef double wx = numpy.dot(self.m, x)
        cdef double step =self.stepsize / (1 + self.nSteps)
        self.m += -step*(wx - y)*x
        self.nSteps += 1

    def se_loss(self, ndarray x, int y):
        cdef double wx = numpy.dot(self.m, x)
        return (wx - y)*(wx - y)
    
def init():
    return (LinearModelHolder(54),)

def init_pyrex():
    return (PyrexModelHolder(54),)

def grad_helper(t_lmh, v):
    a = numpy.array(v[0])
    l = v[1]
    (lmh,) = t_lmh
    lmh.grad_step(a, l)

def loss_fn(t_lmh, v):
    a = numpy.array(v[0])
    l = v[1]
    (lmh,) = t_lmh
    return lmh.se_loss(a, l)


def stopping_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
    s['state'] += 1
    return s['state'] > 10
