"""
This module implements the unit tests for
read and write together of the atomic data types.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
import cStringIO

class AtomicDataProcessorRoundTripRegressionTests(unittest.TestCase):
    
    def generic_test_round_trip_data(self, is_array_type, random_fnc, write_fnc, read_fnc, method_name):
        """
        Tests whether random data types are successfully read and written by atomic_data_processor
        @type is_array_type: number
        @param is_array_type: whether data type is non-array, 1d array or 2d array
        @type random_fnc: function
        @param random_fnc: random function that creates a random atomic data
        @type write_fnc: function
        @param write_fnc: function in the atomic_data_processor to write the atomic data
        @type read_fnc: function
        @param read_fnc: function in the atomic_data_processor to read the atomic data
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        for i in range(test_num):
            b = cStringIO.StringIO()
            #create rand data
            rand_data = create_rand_data(is_array_type, random_fnc)
            write_fnc(b, rand_data)
            ret_data = read_fnc(VictorBuffer(b.getvalue()))
            #compare the data
            compare_data(is_array_type, self, rand_data, ret_data, method_name, exact_equal)

    
    def test_round_trip_char(self):
        """
        Tests whether random chars are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_char, 
                                          Atomic_Data_Processor().write_char, 
                                          Atomic_Data_Processor().read_char, 'round_trip_char')


    
    def test_round_trip_bool(self):
        """
        Tests whether random bools are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_bool, 
                                          Atomic_Data_Processor().write_bool, 
                                          Atomic_Data_Processor().read_bool, 'round_trip_bool')

    
    def test_round_trip_int2(self):
        """
        Tests whether random int2s are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_int2, 
                                          Atomic_Data_Processor().write_int2, 
                                          Atomic_Data_Processor().read_int2, 'round_trip_int2')

    
    def test_round_trip_int4(self):
        """
        Tests whether random int4s are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_int4, 
                                          Atomic_Data_Processor().write_int4, 
                                          Atomic_Data_Processor().read_int4, 'round_trip_int4')

    
    def test_round_trip_int8(self):
        """
        Tests whether random int8s are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_int8, 
                                          Atomic_Data_Processor().write_int8, 
                                          Atomic_Data_Processor().read_int8, 'round_trip_int8')

    
    def test_round_trip_float4(self):
        """
        Tests whether random float4s are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_float4, 
                                          Atomic_Data_Processor().write_float4, 
                                          Atomic_Data_Processor().read_float4, 'round_trip_float4')

    
    def test_round_trip_float8(self):
        """
        Tests whether random float8s are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_non_array, create_random_float8, 
                                          Atomic_Data_Processor().write_float8, 
                                          Atomic_Data_Processor().read_float8, 'round_trip_float8')

    
    def test_round_trip_char_array(self):
        """
        Tests whether random char arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_char, 
                                          Atomic_Data_Processor().write_char_array, 
                                          Atomic_Data_Processor().read_char_array, 'round_trip_char_array')

    
    def test_round_trip_bool_array(self):
        """
        Tests whether random bool arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_bool, 
                                          Atomic_Data_Processor().write_bool_array, 
                                          Atomic_Data_Processor().read_bool_array, 'round_trip_bool_array')

    
    def test_round_trip_int2_array(self):
        """
        Tests whether random int2 arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_int2, 
                                          Atomic_Data_Processor().write_int2_array, 
                                          Atomic_Data_Processor().read_int2_array, 'round_trip_int2_array')

    
    def test_round_trip_int4_array(self):
        """
        Tests whether random int4 arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_int4, 
                                          Atomic_Data_Processor().write_int4_array, 
                                          Atomic_Data_Processor().read_int4_array, 'round_trip_int4_array')

    
    def test_round_trip_int8_array(self):
        """
        Tests whether random int8 arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_int8, 
                                          Atomic_Data_Processor().write_int8_array, 
                                          Atomic_Data_Processor().read_int8_array, 'round_trip_int8_array')

    
    def test_round_trip_float4_array(self):
        """
        Tests whether random float4 arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_float4, 
                                          Atomic_Data_Processor().write_float4_array, 
                                          Atomic_Data_Processor().read_float4_array, 'round_trip_float4_array')

    
    def test_round_trip_float8_array(self):
        """
        Tests whether random float8 arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_1d_array, create_random_float8, 
                                          Atomic_Data_Processor().write_float8_array, 
                                          Atomic_Data_Processor().read_float8_array, 'round_trip_float8_array')

    
    def test_round_trip_float8_2d_array(self):
        """
        Tests whether random float8 2d arrays are successfully read and written by atomic_data_processor
        """
        self.generic_test_round_trip_data(type_2d_array, create_random_float8, 
                                          Atomic_Data_Processor().write_float8_2d_array, 
                                          Atomic_Data_Processor().read_float8_2d_array, 'round_trip_float8_2d_array')



