"""
This module implements the unit tests for
write of the atomic data types.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
import cStringIO

class AtomicDataProcessorWriteRegressionTests(unittest.TestCase):
    def generic_test_write_data(self, is_array_type, random_fnc, elm_size, elm_type, pack_fnc,
                                write_fnc, method_name):
        """
        Tests whether random data types are successfully written by atomic_data_processor
        @type is_array_type: number
        @param is_array_type: whether data type is non-array, 1d array or 2d array
        @type random_fnc: function
        @param random_fnc: random function that creates a random atomic data
        @type elm_size: number
        @param elm_size: Size of the atomic type (e.g. char is 1)
        @type elm_type: number
        @param elm_type: oid of the data type
        @type pack_fnc: function
        @param pack_fnc: pack function that converts atomic data to the binary representation
        @type write_fnc: function
        @param write_fnc: function in the atomic_data_processor to write the atomic data
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        for i in range(test_num):
            b = cStringIO.StringIO()
            comp_b = cStringIO.StringIO()
            #create rand data
            rand_data = create_rand_data(is_array_type, random_fnc)
            #write rand data to the buffer
            write_rand_data(is_array_type, comp_b, rand_data, elm_size, elm_type, pack_fnc)

            write_fnc(b, rand_data)
            self.assertEqual(b.getvalue(), comp_b.getvalue(), method_name + ' fails')

    
    def test_write_char(self):
        """
        Tests whether random chars are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_char, oid_dict[charoid][data_size_field], 
                                     charoid, struct_char.pack, Atomic_Data_Processor().write_char, 'write_char')

    
    def test_write_bool(self):
        """
        Tests whether random bools are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_bool, oid_dict[booloid][data_size_field], 
                                     booloid, struct_bool.pack, Atomic_Data_Processor().write_bool, 'write_bool')

    
    def test_write_int2(self):
        """
        Tests whether random int2s are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_int2, oid_dict[int2oid][data_size_field], 
                                     int2oid, struct_int2.pack, Atomic_Data_Processor().write_int2, 'write_int2')

    
    def test_write_int4(self):
        """
        Tests whether random int4s are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_int4, oid_dict[int4oid][data_size_field], 
                                     int4oid, struct_int4.pack, Atomic_Data_Processor().write_int4, 'write_int4')

    
    def test_write_int8(self):
        """
        Tests whether random int8s are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_int8, oid_dict[int8oid][data_size_field], 
                                     int8oid, struct_int8.pack, Atomic_Data_Processor().write_int8, 'write_int8')
        

    
    def test_write_float4(self):
        """
        Tests whether random float4s are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_float4, oid_dict[float4oid][data_size_field], 
                                     float4oid, struct_float4.pack, Atomic_Data_Processor().write_float4, 'write_float4')
        
    
    def test_write_float8(self):
        """
        Tests whether random float8s are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_non_array, create_random_float8, oid_dict[float8oid][data_size_field], 
                                     float8oid, struct_float8.pack, Atomic_Data_Processor().write_float8, 'write_float8')


    
    def test_write_char_array(self):
        """
        Tests whether random char arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_char, oid_dict[charoid][data_size_field], 
                                     charoid, struct_char.pack, Atomic_Data_Processor().write_char_array, 'write_char_array')

    
    def test_write_bool_array(self):
        """
        Tests whether random bool arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_bool, oid_dict[booloid][data_size_field], 
                                     booloid, struct_bool.pack, Atomic_Data_Processor().write_bool_array, 'write_bool_array')

    
    def test_write_int2_array(self):
        """
        Tests whether random int2 arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_int2, oid_dict[int2oid][data_size_field], 
                                     int2oid, struct_int2.pack, Atomic_Data_Processor().write_int2_array, 'write_int2_array')

    
    def test_write_int4_array(self):
        """
        Tests whether random int4 arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_int4, oid_dict[int4oid][data_size_field], 
                                     int4oid, struct_int4.pack, Atomic_Data_Processor().write_int4_array, 'write_int4_array')

    
    def test_write_int8_array(self):
        """
        Tests whether random int8 arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_int8, oid_dict[int8oid][data_size_field], 
                                     int8oid, struct_int8.pack, Atomic_Data_Processor().write_int8_array, 'write_int8_array')

    
    def test_write_float4_array(self):
        """
        Tests whether random float4 arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_float4, oid_dict[float4oid][data_size_field], 
                                     float4oid, struct_float4.pack, Atomic_Data_Processor().write_float4_array, 
                                     'write_float4_array')

    
    def test_write_float8_array(self):
        """
        Tests whether random float8 arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_1d_array, create_random_float8, oid_dict[float8oid][data_size_field], 
                                     float8oid, struct_float8.pack, Atomic_Data_Processor().write_float8_array, 
                                     'write_float8_array')

    
    def test_write_float8_2d_array(self):
        """
        Tests whether random float8 2d arrays are successfully written by atomic_data_processor
        """
        self.generic_test_write_data(type_2d_array, create_random_float8, oid_dict[float8oid][data_size_field], 
                                     float8oid, struct_float8.pack, Atomic_Data_Processor().write_float8_2d_array, 
                                     'write_float8_2d_array')


