import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from common_unittest_methods import *

from parse_fast import *
from operator import itemgetter, attrgetter
import psycopg2
import cStringIO

general_row_size = 500

test_read_all_array_table_name = 'test_read_all_array_tuples'
test_triples_table_name = 'test_triples'
test_2d_arrays = 'test_2d_arrays'
test_read_all_ia_pairs_table_name = 'test_ia_pairs'

class VictorBinReadRegressionTestCases(unittest.TestCase):        
    def test_read_all_array_tuples(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Read all array table that contains vec float8[]
        create_table_query = 'DROP TABLE IF EXISTS ' + test_read_all_array_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_read_all_array_table_name + '(vec float8[]);'

        cur.execute(create_table_query)

        # TODO: Convert this part into binary write
        for i in range(num_of_tuples):
            vec_str = '\'{'
            for j in range(vec_size):
                vec_str += str(random.random())
                if (j < vec_size - 1):
                    vec_str += ', '
            vec_str += '}\''
            cur.execute('INSERT INTO ' + test_read_all_array_table_name + ' VALUES(' + vec_str + ')')


        b = cStringIO.StringIO()
        cur.copy_expert('COPY ' + test_read_all_array_table_name + '(vec) TO STDOUT WITH BINARY', b)

        pb   = PSQLByteParser(b.getvalue())
        v = pb.read_all_array_tuples()

        cur2 = conn.cursor()
        cur2.execute('SELECT vec FROM ' + test_read_all_array_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)
            
        sorted(all_records, key=itemgetter(0))
        sorted(v, key=itemgetter(0))

        self.assertEqual(len(all_records), len(v), 'wrong # of tuples')

        for i in range(len(all_records)):
            error_msg = 'vector length differs: ' + str(len(all_records[i][0])) + ' ' +  str(len(v[i]))
            self.assertEqual(len(all_records[i][0]), len(v[i]), error_msg)
            self.assertEqual(len(all_records[i][0]), vec_size)
            self.assertEqual(len(v[i]), vec_size)
            for j in range(len(all_records[i][0])):
                self.assertEqual(all_records[i][0][j], v[i][j], 'array content differs')

        conn.commit()
        cur.close()
        cur2.close()
        conn.close()

    def test_read_triples(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Read triples table that contains row int, col int, rating float8
        create_table_query = 'DROP TABLE IF EXISTS ' + test_triples_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_triples_table_name + '(row int, col int, rating float8);'
        cur.execute(create_table_query)

        # TODO: Convert this part into binary write
        for i in range(num_of_tuples):
            rating = random.uniform(1, 5)
            row = random.randint(1, 10000)
            col = random.randint(1, 10000)
            cur.execute('INSERT INTO ' + test_triples_table_name + ' VALUES(' + str(row) 
                        + ', ' + str(col) + ', ' + str(rating) + ')')


        b = cStringIO.StringIO()
        cur.copy_expert('COPY ' + test_triples_table_name + '(row,col,rating) TO STDOUT WITH BINARY', b)

        pb = PSQLByteParser(b.getvalue())
        v = pb.read_all_triples()

        cur2 = conn.cursor()
        cur2.execute('SELECT * FROM ' + test_triples_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)
            
        sorted(all_records, key=itemgetter(0, 1, 2))
        sorted(v, key=itemgetter(0, 1, 2))

        self.assertEqual(len(all_records), len(v), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], v[i][0], 'row values differ')
            self.assertEqual(all_records[i][1], v[i][1], 'column values differ')
            self.assertEqual(all_records[i][2], v[i][2], 'rating values differ')

        conn.commit()
        cur.close()
        cur2.close()
        conn.close()


    def create_2d_double_array_str(self, row_num, col_num):
        vec_str = '\'{'
        for i in range(row_num):
            vec_str += '{'
            for j in range(col_num):
                vec_str += str(random.random())
                if (j < col_num - 1):
                    vec_str += ','
            vec_str += '}'
            if (i < row_num - 1):
                vec_str += ','
        vec_str += '}\''

        return vec_str

    def compare_2d_arrays(self, row_num, col_num, array1, array2):
        self.assertEqual(len(array1), len(array2), 'row numbers differ')
        self.assertEqual(len(array1), row_num, 'row numbers differ')
        self.assertEqual(row_num, len(array2), 'row numbers differ')
        for i in range(len(array1)):
            self.compare_1d_arrays(col_num, array1[i], array2[i])

    def compare_1d_arrays(self, dim, array1, array2):
        self.assertEqual(len(array1), len(array2), 'row numbers differ')
        self.assertEqual(len(array1), dim, 'row numbers differ')
        self.assertEqual(dim, len(array2), 'row numbers differ')
        for i in range(len(array1)):
            self.assertEqual(array1[i], array2[i])
    

    def test_read_single_pair_2d(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # 2d double array table that contains l float8[], r float8[]
        create_table_query = 'DROP TABLE IF EXISTS ' + test_2d_arrays + ';'
        create_table_query += 'CREATE TABLE ' + test_2d_arrays + '(id int, l float8[], r float8[]);'
        cur.execute(create_table_query)

        # TODO: Convert this part into binary write
        for i in range(num_of_tuples):
            l_vector_str = self.create_2d_double_array_str(l_row_num, col_num)
            r_vector_str = self.create_2d_double_array_str(r_row_num, col_num)
            cur.execute('INSERT INTO ' + test_2d_arrays + ' VALUES(' + str(i) + ', ' + l_vector_str + ', ' + r_vector_str + ')')

        conn.commit()

        for i in range(num_of_tuples):
            b = cStringIO.StringIO()
            cur.copy_expert('COPY (SELECT l, r FROM ' + test_2d_arrays + ' WHERE id = ' + str(i) + ') TO STDOUT WITH BINARY', b)

            pb = PSQLByteParser(b.getvalue())
            v = pb.read_single_pair_2d()

            cur2 = conn.cursor()
            cur2.execute('SELECT l, r FROM ' + test_2d_arrays + ' WHERE id = ' + str(i) + ';')
            all_records = []
            for record in cur2:
                all_records.append(record)

            # select statement has to return only 1 tuple
            self.assertEqual(len(all_records), 1, 'wrong # of tuples')

            self.compare_2d_arrays(l_row_num, col_num, all_records[0][0], v[0]) # compare l arrays
            self.compare_2d_arrays(r_row_num, col_num, all_records[0][1], v[1]) # compare r arrays

        conn.commit()
        cur.close()
        conn.close()


    def test_read_all_ia_pairs(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Read all ia paris that contains val int, vec float8[]
        create_table_query = 'DROP TABLE IF EXISTS ' + test_read_all_ia_pairs_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_read_all_ia_pairs_table_name + '(val int, vec float8[]);'

        cur.execute(create_table_query)

        # TODO: Convert this part into binary write
        for i in range(num_of_tuples):
            int_val = random.randint(1, 100000)
            vec_str = '\'{'
            for j in range(vec_size):
                vec_str += str(random.random())
                if (j < vec_size - 1):
                    vec_str += ', '
            vec_str += '}\''
            cur.execute('INSERT INTO ' + test_read_all_ia_pairs_table_name + ' VALUES(' + str(int_val) + ', ' + vec_str + ');')

        b = cStringIO.StringIO()
        cur.copy_expert('COPY ' + test_read_all_ia_pairs_table_name + ' TO STDOUT WITH BINARY', b)

        pb = PSQLByteParser(b.getvalue())
        v = pb.read_all_ia_pairs()

        cur2 = conn.cursor()
        cur2.execute('SELECT * FROM ' + test_read_all_ia_pairs_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)
            
        sorted(all_records, key=itemgetter(0, 1))
        sorted(v, key=itemgetter(0, 1))

        self.assertEqual(len(all_records), len(v), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], v[i][0], 'int values differ')
            error_msg = 'vector length differs: ' + str(len(all_records[i][1])) + ' ' +  str(len(v[i][1]))
            self.assertEqual(len(all_records[i][1]), len(v[i][1]), error_msg)
            self.assertEqual(len(all_records[i][1]), vec_size)
            self.assertEqual(len(v[i][1]), vec_size)
            for j in range(len(all_records[i][1])):
                self.assertEqual(all_records[i][1][j], v[i][1][j], 'array content differs')

        conn.commit()
        cur.close()
        cur2.close()
        conn.close()

    def test_read_int32(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            val = random.randint(1, 1000000)
            b.write(struct_int32.pack(val))
            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_int32()
            self.assertEqual(val, ret_val, 'read_int32 failed')

    def test_read_int64(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            val = random.randint(1, 1000000)
            b.write(struct_int64.pack(val))
            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_int64()
            self.assertEqual(val, ret_val, 'read_int64 failed')

    def test_read_int16(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            val = random.randint(1, 32600) # short int
            b.write(struct_int16.pack(val))
            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_int16()
            self.assertEqual(val, ret_val, 'read_int16 failed')

    def test_read_double(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            val = random.uniform(1, 1000000)
            b.write(struct_double.pack(val))
            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_double()
            self.assertEqual(val, ret_val, 'read_double failed')

    # generic method to test psqltype
    def read_generic_psqltype(self, elm_type_str):
        elm_type = int32oid
        elm_size = size_int32
        struct_x = struct_int32
        array = []

        if (elm_type_str == 'int16'):
            elm_type = int16oid
            elm_size = size_int16
            struct_x = struct_int16
            for j in range(test_num):
                val = random.randint(1, 32000)
                array.append(val)
        elif (elm_type_str == 'int32'):
            elm_type = int32oid
            elm_size = size_int32
            struct_x = struct_int32
            for j in range(test_num):
                val = random.randint(1, 100000)
                array.append(val)
        elif (elm_type_str == 'int64'):
            elm_type = int64oid
            elm_size = size_int64
            struct_x = struct_int64
            for j in range(test_num):
                val = random.randint(1, 100000)
                array.append(val)
        elif (elm_type_str == 'double'):
            elm_type = doubleoid
            elm_size = size_double
            struct_x = struct_double
            for j in range(test_num):
                val = random.uniform(1, 100000)
                array.append(val)

        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_x.pack(array[i]))
            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_psqltype(elm_type)
            self.assertEqual(array[i], ret_val, 'read psqltype ' + elm_type_str + ' failed')

    def test_read_psqltype_int32(self):
        self.read_generic_psqltype('int32')

    def test_read_psqltype_int64(self):
        self.read_generic_psqltype('int64')

    def test_read_psqltype_int16(self):
        self.read_generic_psqltype('int16')

    def test_read_psqltype_double(self):
        self.read_generic_psqltype('double')

    def generic_create_and_write_1d_array_with_length(self, b, elm_type_str):
        elm_size = size_int16
        if (elm_type_str == 'int16'):
            elm_size = size_int16
        elif (elm_type_str == 'int32'):
            elm_size = size_int32
        elif (elm_type_str == 'int64'):
            elm_size = size_int64
        elif (elm_type_str == 'double'):
            elm_size = size_double
        b.write(struct_int32.pack(size_int32 * 5 + vec_size * (elm_size))) #array size
        return self.generic_create_and_write_1d_array(b, elm_type_str)

    def generic_create_and_write_1d_array(self, b, elm_type_str):
        elm_type = int32oid
        elm_size = size_int32
        struct_x = struct_int32
        array = []

        if (elm_type_str == 'int16'):
            elm_type = int16oid
            elm_size = size_int16
            struct_x = struct_int16
            for j in range(vec_size):
                val = random.randint(1, 32000)
                array.append(val)
        elif (elm_type_str == 'int32'):
            elm_type = int32oid
            elm_size = size_int32
            struct_x = struct_int32
            for j in range(vec_size):
                val = random.randint(1, 100000)
                array.append(val)
        elif (elm_type_str == 'int64'):
            elm_type = int64oid
            elm_size = size_int64
            struct_x = struct_int64
            for j in range(vec_size):
                val = random.randint(1, 100000)
                array.append(val)
        elif (elm_type_str == 'double'):
            elm_type = doubleoid
            elm_size = size_double
            struct_x = struct_double
            for j in range(vec_size):
                val = random.uniform(1, 100000)
                array.append(val)


        b.write(struct_int32.pack(1)) #ndims
        b.write(struct_int32.pack(0)) #has null
        b.write(struct_int32.pack(elm_type)) #element type
        b.write(struct_int32.pack(vec_size)) #length
        b.write(struct_int32.pack(1)) #bound

        for j in range(vec_size):
            b.write(struct_int32.pack(elm_size))
            b.write(struct_x.pack(array[j]))

        return array

    #generic array to test read_array method
    def read_generic_array(self, elm_type_str):
        for i in range(test_num):
            b = cStringIO.StringIO()
            array = self.generic_create_and_write_1d_array(b, elm_type_str)

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_array(0) #TODO: Max length is not used?
            self.compare_1d_arrays(vec_size, array, ret_val)

    def test_read_array_int32(self):
        self.read_generic_array('int32')

    # TODO: python type (q) is not supported for array
    # http://docs.python.org/library/array.html
    def test_read_array_int64(self):
        #TODO: Type?
        self.read_generic_array('int32')

    def test_read_array_int16(self):
        self.read_generic_array('int16')

    def test_read_array_double(self):
        self.read_generic_array('double')

    def create_2d_double_array(self, row, col):
        val_array = []

        for i in range(row):
            val_array.append(create_1d_double_array(col))

        return val_array

    def test_read_binary_2d_double_array(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int32.pack(2)) #ndims
            b.write(struct_int32.pack(0)) #has null
            b.write(struct_int32.pack(doubleoid)) #element type
            b.write(struct_int32.pack(general_row_size)) #length
            b.write(struct_int32.pack(1)) #bound
            b.write(struct_int32.pack(col_num)) #length
            b.write(struct_int32.pack(1)) #bound

            val_array = self.create_2d_double_array(general_row_size, col_num)

            for j in range(general_row_size):
                for k in range(col_num):
                    b.write(struct_int32.pack(size_double))
                    b.write(struct_double.pack(val_array[j][k]))

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_binary_2d_double_array()
            self.assertEqual(len(ret_val), len(val_array), 'vec size differs')
            self.assertEqual(len(ret_val), general_row_size, 'vec size differs')
            for j in range(len(ret_val)):
                self.assertEqual(len(ret_val[j]), len(val_array[j]), 'vec size differs')
                self.assertEqual(len(ret_val[j]), col_num, 'vec size differs')
                for k in range(len(ret_val[j])):
                    self.assertEqual(ret_val[j][k], val_array[j][k], 'vector content differs')
                    

    def test_read_all_2darrays(self):
        b = cStringIO.StringIO()
        general_val_array = []

        for i in range(test_num):
            b.write(struct_int32.pack(2)) #ndims
            b.write(struct_int32.pack(0)) #has null
            b.write(struct_int32.pack(doubleoid)) #element type
            b.write(struct_int32.pack(general_row_size)) #length
            b.write(struct_int32.pack(1)) #bound
            b.write(struct_int32.pack(col_num)) #length
            b.write(struct_int32.pack(1)) #bound

            val_array = [[0.0] * col_num for i in xrange(general_row_size)]

            for j in range(general_row_size):
                for k in range(col_num):
                    val = random.random()
                    val_array[j][k] = val
                    b.write(struct_int32.pack(size_double))
                    b.write(struct_double.pack(val))

            general_val_array.append(val_array)

        pb = PSQLByteParser(b.getvalue(), False, False)
        ret_val = pb.read_all_2darrays()
        self.assertEqual(len(ret_val), len(general_val_array), 'vec size differs')
        self.assertEqual(len(ret_val), test_num, 'vec size differs')
        for i in range(len(ret_val)):
            self.assertEqual(len(ret_val[i]), len(general_val_array[i]), 'vec size differs')
            self.assertEqual(len(ret_val[i]), general_row_size, 'vec size differs')
            for j in range(len(ret_val[i])):
                self.assertEqual(len(ret_val[i][j]), len(general_val_array[i][j]), 'vec size differs')
                self.assertEqual(len(ret_val[i][j]), col_num, 'vec size differs')
                for k in range(len(ret_val[i][j])):
                    self.assertEqual(ret_val[i][j][k], general_val_array[i][j][k], 'vec content differs')

    def write_2d_double_array_with_array_size(self, b, row, col, array):
        b.write(struct_int32.pack(size_int32 * 5 + row * col * (size_int32 + size_double))) #array size
        b.write(struct_int32.pack(2)) #ndims
        b.write(struct_int32.pack(0)) #has null
        b.write(struct_int32.pack(doubleoid)) #element type
        b.write(struct_int32.pack(row)) #length
        b.write(struct_int32.pack(1)) #bound
        b.write(struct_int32.pack(col)) #length
        b.write(struct_int32.pack(1)) #bound
        self.write_2d_double_array(b, row, col, array)

    def write_2d_double_array(self, b, row, col, array):
        for i in range(row):
            write_1d_double_array(b, col, array[i])


    def test_read_single_pair_2d(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int16.pack(2)) #nFields

            #l array
            l_array = self.create_2d_double_array(l_row_num, col_num)
            self.write_2d_double_array_with_array_size(b, l_row_num, col_num, l_array)

            #r array
            r_array = self.create_2d_double_array(r_row_num, col_num)
            self.write_2d_double_array_with_array_size(b, r_row_num, col_num, r_array)

            pb = PSQLByteParser(b.getvalue(), False, False)
            (ret_l, ret_r) = pb.read_single_pair_2d()

            # compare l_array
            self.compare_2d_arrays(l_row_num, col_num, l_array, ret_l)
            # compare r_array
            self.compare_2d_arrays(r_row_num, col_num, r_array, ret_r)


    def test_read_double_array(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int32.pack(1)) #ndims
            b.write(struct_int32.pack(0)) #has null
            b.write(struct_int32.pack(doubleoid)) #element type
            b.write(struct_int32.pack(vec_size)) #length
            b.write(struct_int32.pack(1)) #bound
            array = []

            for j in range(vec_size):
                val = random.random()
                array.append(val)
                b.write(struct_int32.pack(size_double))
                b.write(struct_double.pack(val))

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_double_array(0) #TODO: Max length is not used?
            self.assertEqual(len(ret_val), vec_size, 'vec size differs')
            for j in range(vec_size):
                self.assertEqual(array[j], ret_val[j], 'vector content differs')

    def test_read_int_array_pair(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int16.pack(2)) #nFields
            b.write(struct_int32.pack(size_int32)) #int size
            int_val = random.randint(1, 100000)
            b.write(struct_int32.pack(int_val)) #int value

            #1d array
            array = create_1d_double_array(vec_size)
            write_1d_double_array_with_array_size(b, vec_size, array)

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_int_array_pair()
            self.assertEqual(ret_val[0], int_val, 'int values differ')

            #compare arrays
            self.compare_1d_arrays(vec_size, array, ret_val[1])


    def test_read_bin_data(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            val = random.random()
            b.write(struct_int32.pack(size_double))
            b.write(struct_double.pack(val))
            b.write(struct_int32.pack(size_int32))
            val = random.randint(1, 100000)
            b.write(struct_int32.pack(val))
            pb = PSQLByteParser(b.getvalue(), False, False)
            self.assertEqual(b.getvalue(), pb.read_bin_data(len(b.getvalue())), 'read bin data fails')

    def test_read_triple(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            (row, col, rating) = create_triple()
            b.write(struct_int16.pack(3))
            b.write(struct_int32.pack(size_int32))
            b.write(struct_int32.pack(row))
            b.write(struct_int32.pack(size_int32))
            b.write(struct_int32.pack(col))
            b.write(struct_int32.pack(size_double))
            b.write(struct_double.pack(rating))
            pb = PSQLByteParser(b.getvalue(), False, False)
            (ret_row, ret_col, ret_rating) = pb.read_triple()
            self.assertEqual(row, ret_row, 'row values differ')
            self.assertEqual(col, ret_col, 'col values differ')
            self.assertEqual(rating, ret_rating, 'rating values differ')
            
    def test_read_array_tuple(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int16.pack(1)) #nFields

            #1d array
            array = create_1d_double_array(vec_size)
            write_1d_double_array_with_array_size(b, vec_size, array)

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_array = pb.read_array_tuple()

            # compare l_array
            self.compare_1d_arrays(vec_size, array, ret_array)

    def test_read_tuple(self):
        for i in range(test_num):
            b = cStringIO.StringIO()
            b.write(struct_int16.pack(5)) #nFields
            arrayint16 = self.generic_create_and_write_1d_array_with_length(b, 'int16')
            arrayint32 = self.generic_create_and_write_1d_array_with_length(b, 'int32')
            arrayint16_2 = self.generic_create_and_write_1d_array_with_length(b, 'int16')
            arraydouble = self.generic_create_and_write_1d_array_with_length(b, 'double')
            arrayint32_2 = self.generic_create_and_write_1d_array_with_length(b, 'int32')

            pb = PSQLByteParser(b.getvalue(), False, False)
            ret_val = pb.read_tuple()

            self.compare_1d_arrays(vec_size, arrayint16, ret_val[0][0])
            self.compare_1d_arrays(vec_size, arrayint32, ret_val[1][0])
            self.compare_1d_arrays(vec_size, arrayint16_2, ret_val[2][0])
            self.compare_1d_arrays(vec_size, arraydouble, ret_val[3][0])
            self.compare_1d_arrays(vec_size, arrayint32_2, ret_val[4][0])

