import random
import unittest
import StringIO
import HTMLTestRunner
import sys

from common_unittest_methods import *

from parse_fast import *
from factor_parser import *
from operator import itemgetter, attrgetter
import psycopg2
import cStringIO

class VictorBinReadWriteRegressionTestCases(unittest.TestCase):
    def test_read_write_triples(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Write triples table that contains row int, col int, rating float8
        create_table_query = 'DROP TABLE IF EXISTS ' + test_triples_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_triples_table_name + '(row int, col int, rating float8);'
        cur.execute(create_table_query)

        b = cStringIO.StringIO()
        output_buf = OutputBuffer(cur, test_triples_table_name, b)

        # create triples
        triples = []

        for i in range(num_of_tuples):
            (row, col, rating) = create_triple()
            triples.append( (row, col, rating) )
            output_buf.output_triple( (row, col, rating) )

        # flush
        output_buf.close()

        read_b = cStringIO.StringIO()
        cur.copy_expert('COPY ' + test_triples_table_name + '(row,col,rating) TO STDOUT WITH BINARY', read_b)

        pb = PSQLByteParser(read_b.getvalue())
        all_records = pb.read_all_triples()
            
        sorted(all_records, key=itemgetter(0, 1, 2))
        sorted(triples, key=itemgetter(0, 1, 2))

        self.assertEqual(len(all_records), len(triples), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], triples[i][0], 'row values differ')
            self.assertEqual(all_records[i][1], triples[i][1], 'column values differ')
            self.assertEqual(all_records[i][2], triples[i][2], 'rating values differ')

        conn.commit()
        cur.close()
        conn.close()

    def test_read_write_ia_pairs(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Write all ia paris that contains val int, vec float8[]
        create_table_query = 'DROP TABLE IF EXISTS ' + test_read_all_ia_pairs_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_read_all_ia_pairs_table_name + '(val int, vec float8[]);'

        cur.execute(create_table_query)

        b = cStringIO.StringIO()
        output_buf = OutputBuffer(cur, test_read_all_ia_pairs_table_name, b)

        all_ia_arrays = []
        for i in range(num_of_tuples):
            int_val = random.randint(1, 100000)
            vec_array = []
            for j in range(vec_size):
                vec_array.append(random.random())
            all_ia_arrays.append( (int_val, vec_array) )
            output_buf.output_double_array(int_val, vec_array)

        # flush
        output_buf.close()

        read_b = cStringIO.StringIO()
        cur.copy_expert('COPY ' + test_read_all_ia_pairs_table_name + ' TO STDOUT WITH BINARY', read_b)

        pb = PSQLByteParser(read_b.getvalue())
        all_records = pb.read_all_ia_pairs()
            
        sorted(all_records, key=itemgetter(0, 1))
        sorted(all_ia_arrays, key=itemgetter(0, 1))

        self.assertEqual(len(all_records), len(all_ia_arrays), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], all_ia_arrays[i][0], 'int values differ')
            error_msg = 'vector length differs: ' + str(len(all_records[i][1])) + ' ' +  str(len(all_ia_arrays[i][1]))
            self.assertEqual(len(all_records[i][1]), len(all_ia_arrays[i][1]), error_msg)
            self.assertEqual(len(all_records[i][1]), vec_size)
            self.assertEqual(len(all_ia_arrays[i][1]), vec_size)
            for j in range(len(all_records[i][1])):
                self.assertEqual(all_records[i][1][j], all_ia_arrays[i][1][j], 'array content differs')

        conn.commit()
        cur.close()
        conn.close()
