import random
import unittest
import StringIO
import HTMLTestRunner
import sys

from common_unittest_methods import *

from parse_fast import *
from factor_parser import *
from operator import itemgetter, attrgetter
import psycopg2
import cStringIO

almost_equal_places = 12

class VictorBinWriteRegressionTestCases(unittest.TestCase):
    def test_write_triples(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Write triples table that contains row int, col int, rating float8
        create_table_query = 'DROP TABLE IF EXISTS ' + test_triples_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_triples_table_name + '(row int, col int, rating float8);'
        cur.execute(create_table_query)

        b = cStringIO.StringIO()
        output_buf = OutputBuffer(cur, test_triples_table_name, b)

        # create triples
        triples = []

        for i in range(num_of_tuples):
            rating = random.uniform(1, 5)
            row = random.randint(1, 10000)
            col = random.randint(1, 10000)
            triples.append( (row, col, rating) )
            output_buf.output_triple( (row, col, rating) )

        # flush
        output_buf.close()

        cur2 = conn.cursor()
        cur2.execute('SELECT * FROM ' + test_triples_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)
            
        sorted(all_records, key=itemgetter(0, 1, 2))
        sorted(triples, key=itemgetter(0, 1, 2))

        self.assertEqual(len(all_records), len(triples), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], triples[i][0], 'row values differ')
            self.assertEqual(all_records[i][1], triples[i][1], 'column values differ')
            # NB: This should be assertEqual when writing and reading are both binary
            self.assertAlmostEqual(all_records[i][2], triples[i][2], almost_equal_places, 'rating values differ')

        conn.commit()
        cur.close()
        cur2.close()
        conn.close()

    def test_write_ia_pairs(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Write all ia paris that contains val int, vec float8[]
        create_table_query = 'DROP TABLE IF EXISTS ' + test_read_all_ia_pairs_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_read_all_ia_pairs_table_name + '(val int, vec float8[]);'

        cur.execute(create_table_query)

        b = cStringIO.StringIO()
        output_buf = OutputBuffer(cur, test_read_all_ia_pairs_table_name, b)

        all_ia_arrays = []
        for i in range(num_of_tuples):
            int_val = random.randint(1, 100000)
            vec_array = []
            for j in range(vec_size):
                vec_array.append(random.random())
            all_ia_arrays.append( (int_val, vec_array) )
            output_buf.output_double_array(int_val, vec_array)

        # flush
        output_buf.close()

        cur2 = conn.cursor()
        cur2.execute('SELECT * FROM ' + test_read_all_ia_pairs_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)
            
        sorted(all_records, key=itemgetter(0, 1))
        sorted(all_ia_arrays, key=itemgetter(0, 1))

        self.assertEqual(len(all_records), len(all_ia_arrays), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], all_ia_arrays[i][0], 'int values differ')
            error_msg = 'vector length differs: ' + str(len(all_records[i][1])) + ' ' +  str(len(all_ia_arrays[i][1]))
            self.assertEqual(len(all_records[i][1]), len(all_ia_arrays[i][1]), error_msg)
            self.assertEqual(len(all_records[i][1]), vec_size)
            self.assertEqual(len(all_ia_arrays[i][1]), vec_size)
            for j in range(len(all_records[i][1])):
                self.assertAlmostEqual(all_records[i][1][j], all_ia_arrays[i][1][j], almost_equal_places, 'array content differs')

        conn.commit()
        cur.close()
        cur2.close()
        conn.close()


    def create_2d_double_array(self, row_num, col_num):
        array = []
        for i in range(row_num):
            part_array = []
            for j in range(col_num):
                part_array.append(random.random())
            array.append(part_array)

        return array

    def test_write_lr_model(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        # prepare the dataset on the db to test
        # Write lr_model table that contains mid int, l float8[], r float8[], moviemean float8, stepsize float8, b float8
        create_table_query = 'DROP TABLE IF EXISTS ' + test_lr_model_table_name + ';'
        create_table_query += 'CREATE TABLE ' + test_lr_model_table_name + '(mid int, l float8[], r float8[], moviemean float8, '
        create_table_query += 'stepsize float8, b float8);'

        cur.execute(create_table_query)

        b = cStringIO.StringIO()
        output_buf = OutputBuffer(cur, test_lr_model_table_name, b)

        all_lr_table_rows = []
        # mid int, l float8[], r float8[], moviemean float8, stepsize float8, b float8
        for i in range(num_of_tuples):
            mid = random.randint(1, 100)
            l_array = self.create_2d_double_array(l_row_num, col_num)
            r_array = self.create_2d_double_array(r_row_num, col_num)
            moviemean = random.uniform(1, 5)
            stepsize = random.random()
            b = random.uniform(1, 2)
            all_lr_table_rows.append((mid, l_array, r_array, moviemean, stepsize, b))
            output_buf.output_lr_model(mid, l_array, r_array, moviemean, stepsize, b)

        # flush
        output_buf.close()

        cur2 = conn.cursor()
        cur2.execute('SELECT * FROM ' + test_lr_model_table_name + ';')
        all_records = []
        for record in cur2:
            all_records.append(record)

        sorted(all_records, key=itemgetter(0, 1, 2, 3, 4, 5))
        sorted(all_lr_table_rows, key=itemgetter(0, 1, 2, 3, 4, 5))

        self.assertEqual(len(all_records), len(all_lr_table_rows), 'wrong # of tuples')

        for i in range(len(all_records)):
            self.assertEqual(all_records[i][0], all_lr_table_rows[i][0], 'mid values differ')
            ## Check l vector
            self.assertEqual(len(all_records[i][1]), len(all_lr_table_rows[i][1]), 'l vector size differs')
            self.assertEqual(len(all_records[i][1]), l_row_num, 'l vector size differs')
            self.assertEqual(len(all_lr_table_rows[i][1]), l_row_num, 'l vector size differs')
            for j in range(len(all_records[i][1])):
                self.assertEqual(len(all_records[i][1][j]), len(all_lr_table_rows[i][1][j]), 'l vector size differs')
                self.assertEqual(len(all_records[i][1][j]), col_num, 'l vector size differs')
                self.assertEqual(len(all_lr_table_rows[i][1][j]), col_num, 'l vector size differs')
                for k in range(len(all_records[i][1][j])):
                    self.assertAlmostEqual(all_records[i][1][j][k], all_lr_table_rows[i][1][j][k], almost_equal_places,
                                           'vector content differs')

            # Check r vector
            self.assertEqual(len(all_records[i][2]), len(all_lr_table_rows[i][2]), 'l vector size differs')
            self.assertEqual(len(all_records[i][2]), r_row_num, 'l vector size differs')
            self.assertEqual(len(all_lr_table_rows[i][2]), r_row_num, 'l vector size differs')
            for j in range(len(all_records[i][2])):
                self.assertEqual(len(all_records[i][2][j]), len(all_lr_table_rows[i][2][j]), 'l vector size differs')
                self.assertEqual(len(all_records[i][2][j]), col_num, 'l vector size differs')
                self.assertEqual(len(all_lr_table_rows[i][2][j]), col_num, 'l vector size differs')
                for k in range(len(all_records[i][2][j])):
                    self.assertAlmostEqual(all_records[i][2][j][k], all_lr_table_rows[i][2][j][k], almost_equal_places, 
                                           'vector content differs')

            # moviemean
            self.assertAlmostEqual(all_records[i][3], all_lr_table_rows[i][3], almost_equal_places, 'moviemean values differ')
            # stepsize
            self.assertAlmostEqual(all_records[i][4], all_lr_table_rows[i][4], almost_equal_places, 'stepsize values differ')
            # b
            self.assertAlmostEqual(all_records[i][5], all_lr_table_rows[i][5], almost_equal_places, 'b values differ')

    def test_write_int32(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        for i in range(test_num):
            b = cStringIO.StringIO()
            output_buf = OutputBuffer(cur, 'test', b, False, False)
            val = random.randint(1, 1000000)
            output_buf.write_int32(val)
            ret_b = cStringIO.StringIO()
            ret_b.write(struct_int32.pack(val))
            self.assertEqual(ret_b.getvalue(), output_buf.buffer.getvalue(), 'write_int32 failed')

    def test_write_double(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        for i in range(test_num):
            b = cStringIO.StringIO()
            output_buf = OutputBuffer(cur, 'test', b, False, False)
            val = random.uniform(1, 1000000)
            output_buf.write_double(val)
            ret_b = cStringIO.StringIO()
            ret_b.write(struct_double.pack(val))
            self.assertEqual(ret_b.getvalue(), output_buf.buffer.getvalue(), 'write_double failed')

    def test_internal_write_triple(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        for i in range(test_num):
            b = cStringIO.StringIO()
            output_buf = OutputBuffer(cur, 'test', b, False, False)
            t = create_triple()
            output_buf.internal_write_triple(t)
            ret_b = cStringIO.StringIO()
            ret_b.write(struct_int16.pack(3))
            ret_b.write(struct_int32.pack(size_int32))
            ret_b.write(struct_int32.pack(t[0]))
            ret_b.write(struct_int32.pack(size_int32))
            ret_b.write(struct_int32.pack(t[1]))
            ret_b.write(struct_int32.pack(size_double))
            ret_b.write(struct_double.pack(t[2]))
            self.assertEqual(ret_b.getvalue(), output_buf.buffer.getvalue(), 'write_triple failed')
                
    def test_internal_write_double_array(self):
        conn = psycopg2.connect('dbname=' + dbname)
        cur = conn.cursor()

        for i in range(test_num):
            b = cStringIO.StringIO()
            output_buf = OutputBuffer(cur, 'test', b, False, False)

            # int val
            int_val = random.randint(1, 100000)
            # 1d array
            array = create_1d_double_array(vec_size)
            
            #write pair
            output_buf.internal_write_double_array(int_val, array)

            ret_b = cStringIO.StringIO()

            ret_b.write(struct_int16.pack(2)) #nFields
            ret_b.write(struct_int32.pack(size_int32)) #int size
            ret_b.write(struct_int32.pack(int_val)) #int value
            write_1d_double_array_with_array_size(ret_b, vec_size, array)

            self.assertEqual(ret_b.getvalue(), output_buf.buffer.getvalue(), 'internal_write_double_array failed')
