"""
This module implements the unit tests for
write of the atomic data types by TupleWriter.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
import cStringIO

class TupleWriterWriteRegressionTests(unittest.TestCase):
    
    def generic_test_write_data(self, schema, type_params, method_name):
        """
        Tests whether random data types are successfully written by TupleWriter
        @type schema: tuple
        @param schema: represents the data types in the tuple
        @type type_params: tuple
        @param type_params: contains information(is_array_type, elm_size, random_fnc, elm_type, pack_fnc)
        about each data type in the tuple.
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        for i in range(test_num):
            b = cStringIO.StringIO()
            comp_b = cStringIO.StringIO()
            comp_b.write(struct_int2.pack(len(type_params))) #nFields
            rand_tuple = []
            for (is_array_type, elm_size, random_fnc, elm_type, pack_fnc) in type_params:
                # create data
                data = create_rand_data(is_array_type, random_fnc)

                #append to the tuple
                rand_tuple.append(data)

                #write data
                write_rand_data(is_array_type, comp_b, data, elm_size, elm_type, pack_fnc)

            ts = TupleSchema(schema)
            tw = TupleWriter(b, ts)
            tw.write_tuple(rand_tuple)

            self.assertEqual(b.getvalue(), comp_b.getvalue(), method_name + ' fails')

    
    def test_write_single_char_tuple(self):
        """
        Tests whether tuples that only contain random char are successfully written by TupleWriter
        """
        self.generic_test_write_data([charoid], [char_type_param], 'write_single_char_tuple')

    
    def test_write_single_bool_tuple(self):
        """
        Tests whether tuples that only contain random bool are successfully written by TupleWriter
        """
        self.generic_test_write_data([booloid], [bool_type_param], 'write_single_bool_tuple')

    
    def test_write_single_int2_tuple(self):
        """
        Tests whether tuples that only contain random int2 are successfully written by TupleWriter
        """
        self.generic_test_write_data([int2oid], [int2_type_param], 'write_single_int2_tuple')

    
    def test_write_single_int4_tuple(self):
        """
        Tests whether tuples that only contain random int4 are successfully written by TupleWriter
        """
        self.generic_test_write_data([int4oid], [int4_type_param], 'write_single_int4_tuple')

    
    def test_write_single_int8_tuple(self):
        """
        Tests whether tuples that only contain random int8 are successfully written by TupleWriter
        """
        self.generic_test_write_data([int8oid], [int8_type_param], 'write_single_int8_tuple')

    
    def test_write_single_float4_tuple(self):
        """
        Tests whether tuples that only contain random float4 are successfully written by TupleWriter
        """
        self.generic_test_write_data([float4oid], [float4_type_param], 'write_single_float4_tuple')

    
    def test_write_single_float8_tuple(self):
        """
        Tests whether tuples that only contain random float8 are successfully written by TupleWriter
        """
        self.generic_test_write_data([float8oid], [float8_type_param], 'write_single_float8_tuple')

    
    def test_write_single_char_array_tuple(self):
        """
        Tests whether tuples that only contain random char array are successfully written by TupleWriter
        """
        self.generic_test_write_data([chararrayoid], [char_array_type_param], 'write_single_char_array_tuple')

    
    def test_write_single_bool_array_tuple(self):
        """
        Tests whether tuples that only contain random bool array are successfully written by TupleWriter
        """
        self.generic_test_write_data([boolarrayoid], [bool_array_type_param], 'write_single_bool_array_tuple')

    
    def test_write_single_int2_array_tuple(self):
        """
        Tests whether tuples that only contain random int2 array are successfully written by TupleWriter
        """
        self.generic_test_write_data([int2arrayoid], [int2_array_type_param], 'write_single_int2_array_tuple')

    
    def test_write_single_int4_array_tuple(self):
        """
        Tests whether tuples that only contain random int4 array are successfully written by TupleWriter
        """
        self.generic_test_write_data([int4arrayoid], [int4_array_type_param], 'write_single_int4_array_tuple')

    
    def test_write_single_int8_array_tuple(self):
        """
        Tests whether tuples that only contain random int8 array are successfully written by TupleWriter
        """
        self.generic_test_write_data([int8arrayoid], [int8_array_type_param], 'write_single_int8_array_tuple')

    
    def test_write_single_float4_array_tuple(self):
        """
        Tests whether tuples that only contain random float4 array are successfully written by TupleWriter
        """
        self.generic_test_write_data([float4arrayoid], [float4_array_type_param], 'write_single_float4_array_tuple')

    
    def test_write_single_float8_array_tuple(self):
        """
        Tests whether tuples that only contain random float8 array are successfully written by TupleWriter
        """
        self.generic_test_write_data([float8arrayoid], [float8_array_type_param], 'write_single_float8_array_tuple')

    
    def test_write_single_float8_2d_array_tuple(self):
        """
        Tests whether tuples that only contain random float8 2d array are successfully written by TupleWriter
        """
        self.generic_test_write_data([float82darrayoid], [float8_2d_array_type_param], 'write_single_float8_2d_array_tuple')

    
    def test_tuples_with_non_array_data(self):
        """
        This method tests write of tuples with possible non-array data type combinations.
        Since there are 7 non-array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_non_array_data(self.generic_test_write_data, limit_test, without_sql)

    
    def test_tuples_with_array_data(self):
        """
        This method tests write of tuples with possible array(1d or 2d) data type combinations.
        Since there are 8 array data types, all combinations produce 255 different schema.
        """
        test_tuples_with_array_data(self.generic_test_write_data, limit_test, without_sql)

    
    def test_tuples_with_data(self):
        """
        This method tests write of tuples with possible non-array and array(1d or 2d) data type combinations.
        Since there are 15 data types, all combinations should produce 2 ** 15 - 1 different schema.
        """
        test_tuples_with_data(self.generic_test_write_data, limit_test, without_sql)

