"""
This module includes helper variables and methods for the unittests.
"""

import random
from sys_appends import *

import sys
import struct

sys.path.append(b + 'src/vsql_core/utils/factored_parser')

from parse_fast import *
from tuple_processor import *

#limit of the enumeration tests
limit_test = 200

#num of tuples in the database
tuple_size = 1
#we run the tests for test_num times
test_num = tuple_size
#size of the vectors that are used in the experiments
vec_size = 100

#row num of 2d arrays
row_num   = 30
#col num of 2d arrays
col_num   = 50

#non-array data type
type_non_array = 0
#1d array data type
type_1d_array = 1
#2d array data type
type_2d_array = 2

#test has sql interaction
with_sql = 1
#test doesnt have sql interaction
without_sql = 0

size_int = 4
struct_int4   = struct.Struct('>i')

temp_table_name = 'temp_tbl_for_test'
temp_file_name = "temp_file.dat"

#during the comparison, we use assertEqual
exact_equal = 0
#during the comparison, we use assertAlmostEqual
almost_equal = 1
almost_eq_places = 6

#enumerate function
enumerate_fnc = lambda x: [[y for j, y in enumerate(set(x)) if (i >> j) & 1] for i in range(2**len(set(x)))]

def compare_non_array_data(cls, data1, data2, method_name, is_almost):
    """
    Compares non-array data type
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type data1: data
    @param data1: data1 to compare
    @type data2: data
    @param data2: data2 to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    if is_almost:
        cls.assertAlmostEqual(data1, data2, almost_eq_places, method_name + ': tuple content is different')
    else:
        cls.assertEqual(data1, data2, method_name + ': tuple content is different')

def compare_array_data(cls, array1, array2, method_name, is_almost):
    """
    Compares 1d array data type
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type array1: array
    @param array1: array1 to compare
    @type array2: array
    @param array2: array2 to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    cls.assertEqual(len(array1), len(array2), method_name + ': array sizes are different')
    for i in range(len(array1)):
        compare_non_array_data(cls, array1[i], array2[i], method_name, is_almost)


def compare_2d_array_data(cls, array1, array2, method_name, is_almost):
    """
    Compares 2d array data type
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type array1: array
    @param array1: array1 to compare
    @type array2: array
    @param array2: array2 to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    cls.assertEqual(len(array1), len(array2), method_name + ': row sizes are different')
    for i in range(len(array1)):
        compare_array_data(cls, array1[i], array2[i], method_name, is_almost)

def compare_data(is_array_type, cls, rand_data, ret_data, method_name, is_almost):
    """
    Compares data type(non-array, 1d or 2d array)
    @type is_array_type: number
    @param is_array_type: whether data is non-array, 1d or 2d array
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type rand_data: data
    @param rand_data: data to compare
    @type ret_data: data
    @param ret_data: data to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    # compare the data value
    if is_array_type == type_2d_array:
        compare_2d_array_data(cls, rand_data, ret_data, method_name, is_almost)
    elif is_array_type == type_1d_array:
        compare_array_data(cls, rand_data, ret_data, method_name, is_almost)
    elif is_array_type == type_non_array:
        compare_non_array_data(cls, rand_data, ret_data, method_name, is_almost)

def compare_tuples(cls, type_params, rand_tuple, ret_tuple, method_name, is_almost):
    """
    Compares tuples that can contain one or more data types
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type type_params: tuple
    @param type_params: contains information about the data types in the tuple
    @type rand_tuple: array
    @param rand_tuple: array to compare
    @type ret_tuple: array
    @param ret_tuple: array to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    cls.assertEqual(len(rand_tuple), len(ret_tuple), 'tuple sizes are different')
    for i in range(len(rand_tuple)):
        compare_data(type_params[i][0], cls, rand_tuple[i], ret_tuple[i], method_name, 
                     is_almost and (type_params[i][3] == float8oid))

def compare_set_of_tuples(cls, type_params, tuple_set1, tuple_set2, method_name, is_almost):
    """
    Compares set of tuples that can contain one or more tuples that can contain one or more data types
    @type cls: class
    @param cls: unittest class for assertEqual functions
    @type type_params: tuple
    @param type_params: contains information about the data types in the tuple
    @type tuple_set1: array
    @param tuple_set1: array to compare
    @type tuple_set2: array
    @param tuple_set2: array to compare
    @type method_name: string
    @param method_name: name of the method
    @type is_almost: bool
    @param is_almost: whether we want exact comparison or almost comparison
    """
    cls.assertEqual(len(tuple_set1), len(tuple_set2), method_name + '# of tuples differs')
    for i in range(len(tuple_set1)):
        compare_tuples(cls, type_params, tuple_set1[i], tuple_set2[i], method_name, is_almost)

def convert_char(v):
    """
    Converts char to sql representation
    @type v: char
    @param v: character
    @rtype: string
    @return sql representation of the given type
    """
    return '\'' + v + '\''

def convert_bool_and_numeric(v):
    """
    Converts bool and numeric types to sql representation
    @type v: bool or a numeric type
    @param v: bool or a numeric type
    @rtype: string
    @return sql representation of the given type
    """
    return repr(v)

def convert_char_array(v):
    """
    Converts char array to sql representation
    @type v: char array
    @param v: character array
    @rtype: string
    @return sql representation of the given type
    """
    str = '\'{'
    str += ','.join(v)
    str += '}\''
    return str

def convert_bool_and_numeric_array(v):
    """
    Converts bool and numeric array types to sql representation
    @type v: bool or a numeric array type
    @param v: bool or a numeric array type
    @rtype: string
    @return sql representation of the given type
    """
    return '\'' + convert_bool_and_numeric_array_without_quotes(v) + '\''

def convert_bool_and_numeric_array_without_quotes(v):
    """
    Converts bool and numeric array types to sql representation without quotes
    @type v: bool or a numeric array type
    @param v: bool or a numeric array type
    @rtype: string
    @return sql representation of the given type
    """
    return '{' + ','.join([repr(el) for el in v]) + '}'

def convert_bool_and_numeric_2d_array(v):
    """
    Converts bool and numeric 2d array types to sql representation
    @type v: bool or a numeric 2d array type
    @param v: bool or a numeric 2d array type
    @rtype: string
    @return sql representation of the given type
    """
    str = '\'{'
    str += ','.join([convert_bool_and_numeric_array_without_quotes(el) for el in v])
    str += '}\''
    return str

def write_non_array_data(b, data, data_size, pack_fnc):
    """
    Writes non-array data to the buffer
    @type b: buffer
    @param b: buffer that the data will be written
    @type data: data
    @param data: non-array type data
    @type data_size: number
    @param data_size: size of the data
    @type pack_fnc: function
    @param pack_fnc: pack function for the data
    """
    # write data size
    b.write(struct_int4.pack(data_size))
    # write rand data to the buffer
    b.write(pack_fnc(data))

def write_array_data(b, data_array, elm_size, elm_type, pack_fnc):
    """
    Writes 1d array data to the buffer
    @type b: buffer
    @param b: buffer that the data will be written
    @type data_array: data
    @param data_array: 1d array type data
    @type elm_size: number
    @param elm_size: size of the data
    @type elm_type: number
    @param elm_type: oid of the element
    @type pack_fnc: function
    @param pack_fnc: pack function for the data
    """
    # write data size
    b.write(struct_int4.pack(size_int * 5 + len(data_array) * (size_int + elm_size))) #Field size
    b.write(struct_int4.pack(1)) #ndims
    b.write(struct_int4.pack(0)) #has_null
    b.write(struct_int4.pack(elm_type)) #elem_type
        
    b.write(struct_int4.pack(len(data_array))) #length
    b.write(struct_int4.pack(1)) #lbound
        
    for val in data_array:
        b.write(struct_int4.pack(elm_size))
        b.write(pack_fnc(val))

def write_2d_array_data(b, data_array, elm_size, elm_type, pack_fnc):
    """
    Writes 2d array data to the buffer
    @type b: buffer
    @param b: buffer that the data will be written
    @type data_array: data
    @param data_array: 2d array type data
    @type elm_size: number
    @param elm_size: size of the data
    @type elm_type: number
    @param elm_type: oid of the element
    @type pack_fnc: function
    @param pack_fnc: pack function for the data
    """
    # write data size
    b.write(struct_int4.pack(size_int * (3 + 4) + len(data_array) * len(data_array[0]) * (size_int + elm_size))) #Field size
    b.write(struct_int4.pack(2)) #ndims
    b.write(struct_int4.pack(0)) #has_null
    b.write(struct_int4.pack(elm_type)) #elem_type
        
    b.write(struct_int4.pack(len(data_array))) #length
    b.write(struct_int4.pack(1)) #lbound
    b.write(struct_int4.pack(len(data_array[0]))) #length
    b.write(struct_int4.pack(1)) #lbound
        
    for x in data_array:
        for val in x:
            b.write(struct_int4.pack(elm_size))
            b.write(pack_fnc(val))

def write_rand_data(is_array_type, b, rand_data, elm_size, elm_type, pack_fnc):
    """
    Writes data(non-array, 1d or 2d array) to the buffer
    @type is_array_type: number
    @param is_array_type: whether data is non-array, 1d or 2d array
    @type b: buffer
    @param b: buffer that the data will be written
    @type rand_data: data
    @param rand_data: non-array, 1d or 2d array type data
    @type elm_size: number
    @param elm_size: size of the data
    @type elm_type: number
    @param elm_type: oid of the element
    @type pack_fnc: function
    @param pack_fnc: pack function for the data
    """
    if is_array_type == type_2d_array:
        write_2d_array_data(b, rand_data, elm_size, elm_type, pack_fnc)
    elif is_array_type == type_1d_array:
        write_array_data(b, rand_data, elm_size, elm_type, pack_fnc)
    elif is_array_type == type_non_array:
        write_non_array_data(b, rand_data, elm_size, pack_fnc)

def create_rand_data(is_array_type, random_fnc):
    """
    Creates random data(non-array, 1d or 2d array)
    @type is_array_type: number
    @param is_array_type: whether data is non-array, 1d or 2d array
    @type random_fnc: function
    @param random_fnc: random function to create a random element
    @rtype: data
    @return: returns random data(non-array, 1d or 2d array)
    """        
    # create data
    if is_array_type == type_2d_array:
        rand_data = create_2d_generic_array(random_fnc)
    elif is_array_type == type_1d_array:
        rand_data = create_1d_generic_array(random_fnc)
    elif is_array_type == type_non_array:
        rand_data = random_fnc()

    return rand_data

def create_random_char():
    """
    Creates random char
    @rtype: char
    @return: random char
    """        
    rand = random.randint(0,35)
    ch = '0'
    if(rand < 10):
        ch = str(rand)
    else:
        ch = chr(rand + 55)
    return ch

def create_random_bool():
    """
    Creates random bool
    @rtype: bool
    @return: random bool
    """        
    rand = random.randint(0,1)
    if rand == 1:
        return True
    else:
        return False

def create_random_int2():
    """
    Creates random int2
    @rtype: int2
    @return: random int2
    """
    return random.randint(-2 ** 15, 2 ** 15 - 1)

def create_random_int4():
    """
    Creates random int4
    @rtype: int4
    @return: random int4
    """        
    return random.randint(-2 ** 31, 2 ** 31 - 1)

def create_random_int8():
    """
    Creates random int8
    @rtype: int8
    @return: random int8
    """        
    return random.randint(-2 ** 63, 2 ** 63 - 1)

#TODO: FIX THIS!!!
def create_random_float4():
    """
    Creates random float4
    @rtype: float4
    @return: random float4
    """        
    return 3.5

#TODO: FIX THIS!!!
def create_random_float8():
    """
    Creates random float8
    @rtype: float8
    @return: random float8
    """        
    return random.uniform(0, 2000000)

def create_1d_generic_array(create_random_elm_fnc):
    """
    Creates 1d array
    @type create_random_elm_fnc: function
    @param create_random_elm_fnc: function to create random elements of the array
    @rtype: array
    @return: random 1d array with the given type
    """        
    ret = []
    for i in range(vec_size):
        ret.append(create_random_elm_fnc())
        
    return ret

def create_2d_generic_array(create_random_elm_fnc):
    """
    Creates 2d array
    @type create_random_elm_fnc: function
    @param create_random_elm_fnc: function to create random elements of the array
    @rtype: array
    @return: random 2d array with the given type
    """        
    ret = []
    for i in range(row_num):
        row_array = []
        for j in range(col_num):
            row_array.append(create_random_elm_fnc())
        ret.append(row_array)

    return ret

def get_vdb_tuple(t, convert_fnc, sql_type):
    """
    Includes a function(to convert data to sql type) and sql type of data
    to the given type param
    @type t: tuple
    @param t: type param
    @type convert_fnc: function
    @param convert_fnc: function to convert data to the sql representation
    @type sql_type: string
    @param sql_type: data type in sql (e.g. 'char' for char, 'float8[]' for float8 array)
    @rtype: tuple
    @return: new type param that includes convert function and sql type
    """
    new_t = []
    for elm in t:
        new_t.append(elm)
    new_t.append(convert_fnc)
    new_t.append(sql_type)
    return new_t

#type params
char_type_param = (type_non_array, oid_dict[charoid][data_size_field], create_random_char, charoid, struct_char.pack)
bool_type_param = (type_non_array, oid_dict[booloid][data_size_field], create_random_bool, booloid, struct_bool.pack)
int2_type_param = (type_non_array, oid_dict[int2oid][data_size_field], create_random_int2, int2oid, struct_int2.pack)
int4_type_param = (type_non_array, oid_dict[int4oid][data_size_field], create_random_int4, int4oid, struct_int4.pack)
int8_type_param = (type_non_array, oid_dict[int8oid][data_size_field], create_random_int8, int8oid, struct_int8.pack)
float4_type_param = (type_non_array, oid_dict[float4oid][data_size_field], create_random_float4, float4oid, struct_float4.pack)
float8_type_param = (type_non_array, oid_dict[float8oid][data_size_field], create_random_float8, float8oid, struct_float8.pack)
char_array_type_param = (type_1d_array, oid_dict[charoid][data_size_field], create_random_char, charoid, struct_char.pack)
bool_array_type_param = (type_1d_array, oid_dict[booloid][data_size_field], create_random_bool, booloid, struct_bool.pack)
int2_array_type_param = (type_1d_array, oid_dict[int2oid][data_size_field], create_random_int2, int2oid, struct_int2.pack)
int4_array_type_param = (type_1d_array, oid_dict[int4oid][data_size_field], create_random_int4, int4oid, struct_int4.pack)
int8_array_type_param = (type_1d_array, oid_dict[int8oid][data_size_field], create_random_int8, int8oid, struct_int8.pack)
float4_array_type_param = (type_1d_array, oid_dict[float4oid][data_size_field], create_random_float4, float4oid, struct_float4.pack)
float8_array_type_param = (type_1d_array, oid_dict[float8oid][data_size_field], create_random_float8, float8oid, struct_float8.pack)
float8_2d_array_type_param = (type_2d_array, oid_dict[float8oid][data_size_field], create_random_float8, float8oid, struct_float8.pack)

#type params for vdb that inclues convert to sql type and sql create type
char_vdb_type_param = get_vdb_tuple(char_type_param, convert_char, 'char')
bool_vdb_type_param = get_vdb_tuple(bool_type_param, convert_bool_and_numeric, 'bool')
int2_vdb_type_param = get_vdb_tuple(int2_type_param, convert_bool_and_numeric, 'int2')
int4_vdb_type_param = get_vdb_tuple(int4_type_param, convert_bool_and_numeric, 'int4')
int8_vdb_type_param = get_vdb_tuple(int8_type_param, convert_bool_and_numeric, 'int8')
float4_vdb_type_param = get_vdb_tuple(float4_type_param, convert_bool_and_numeric, 'float4')
float8_vdb_type_param = get_vdb_tuple(float8_type_param, convert_bool_and_numeric, 'float8')
bool_array_vdb_type_param = get_vdb_tuple(bool_array_type_param, convert_bool_and_numeric_array, 'bool[]')
int2_array_vdb_type_param = get_vdb_tuple(int2_array_type_param, convert_bool_and_numeric_array, 'int2[]')
int4_array_vdb_type_param = get_vdb_tuple(int4_array_type_param, convert_bool_and_numeric_array, 'int4[]')
int8_array_vdb_type_param = get_vdb_tuple(int8_array_type_param, convert_bool_and_numeric_array, 'int8[]')
float4_array_vdb_type_param = get_vdb_tuple(float4_array_type_param, convert_bool_and_numeric_array, 'float4[]')
float8_array_vdb_type_param = get_vdb_tuple(float8_array_type_param, convert_bool_and_numeric_array, 'float8[]')
float8_2d_array_vdb_type_param = get_vdb_tuple(float8_2d_array_type_param, convert_bool_and_numeric_2d_array, 'float8[][]')

def enumerate_and_test_tuples(test_fnc, type_params, oids, limit_test):
    """
    This method enumerates all subsets of the given types and creates all combination
    of different schemas (e.g. [char, int2] or [char, int4, float8array]).
    Then, for each combination, test method(test_fnc) is called.
    # of tests is: 2 ** len(type_params) - 1
    @type test_fnc: function
    @param test_fnc: name of the generic test function
    @type type_params: tuple
    @param type_params: contains information(is_array_type, elm_size, random_fnc, elm_type, pack_fnc)
    about each data type in the tuple.
    @type oids: tuple
    @param oids: oids of the data types in the tuple
    @type limit_test: number
    @param limit_test: when enumerations are too many to execute, limit_test is used to limit the number of enumerations
    """
    # We assert that type params have same length with oids
    assert(len(type_params) == len(oids))
    enumerated_list = enumerate_fnc(xrange(len(type_params)))
    
    for i in range(len(enumerated_list)):
        if(len(enumerated_list[i]) != 0 and i < limit_test):
            param_array_for_fnc = []
            oid_array_for_fnc   = []
            method_name = ''
            for j in range(len(enumerated_list[i])):
                param_array_for_fnc.append(type_params[enumerated_list[i][j]])
                oid_of_elm = oids[enumerated_list[i][j]]
                oid_array_for_fnc.append(oid_of_elm)
                method_name += '_' + oid_dict[oid_of_elm][0]

            test_fnc(oid_array_for_fnc, param_array_for_fnc, method_name)

def test_tuples_with_non_array_data(test_fnc, limit_test, sql_interaction):
    """
    This method tests tuples with possible non-array data type combinations by calling test_fnc.
    Since there are 7 non-array data types, all combinations produce 127 different schema.
    Therefore, this method includes min(127, limit_test) tests
    @type test_fnc: function
    @param test_fnc: name of the generic test function
    @type limit_test: number
    @param limit_test: when enumerations are too many to execute, limit_test is used to limit the number of enumerations
    @type sql_interaction: bool
    @param sql_interaction: whether test involves postgres or not
    """
    if sql_interaction:
        type_params = [char_vdb_type_param, bool_vdb_type_param, int2_vdb_type_param, 
                       int4_vdb_type_param, int8_vdb_type_param, float4_vdb_type_param, float8_vdb_type_param]
    else:
        type_params = [char_type_param, bool_type_param, int2_type_param, int4_type_param, int8_type_param,
                       float4_type_param, float8_type_param]

    oids         = [charoid, booloid, int2oid, int4oid, int8oid, float4oid, float8oid]
    enumerate_and_test_tuples(test_fnc, type_params, oids, limit_test)

def test_tuples_with_array_data(test_fnc, limit_test, sql_interaction):
    """
    This method tests tuples with possible array data type combinations by calling test_fnc.
    Since there are 8 array data types(for the ones with sql interaction, we dont use char arrays, since they are 
    interpreted as strings, so it is 7), all combinations produce 255(127) different schema.
    Therefore, this method includes min(255(127), limit_test) tests
    @type test_fnc: function
    @param test_fnc: name of the generic test function
    @type limit_test: number
    @param limit_test: when enumerations are too many to execute, limit_test is used to limit the number of enumerations
    @type sql_interaction: bool
    @param sql_interaction: whether test involves postgres or not
    """
    if sql_interaction:
        type_params = [bool_array_vdb_type_param, int2_array_vdb_type_param, 
                       int4_array_vdb_type_param, int8_array_vdb_type_param, 
                       float4_array_vdb_type_param, float8_array_vdb_type_param, 
                       float8_2d_array_vdb_type_param]
        oids         = [boolarrayoid, int2arrayoid, int4arrayoid, 
                        int8arrayoid, float4arrayoid, float8arrayoid, float82darrayoid]
    else:
        type_params = [char_array_type_param, bool_array_type_param, int2_array_type_param, 
                       int4_array_type_param, int8_array_type_param, float4_array_type_param, 
                       float8_array_type_param, float8_2d_array_type_param]
        oids         = [chararrayoid, boolarrayoid, int2arrayoid, int4arrayoid, 
                        int8arrayoid, float4arrayoid, float8arrayoid, float82darrayoid]

    enumerate_and_test_tuples(test_fnc, type_params, oids, limit_test)

def test_tuples_with_data(test_fnc, limit_test, sql_interaction):
    """
    This method tests tuples with possible non-array and array data type combinations by calling test_fnc.
    Since there are 7 non-array and 8 array data types(for the ones with sql interaction, we dont use char arrays, 
    since they are interpreted as strings, so it is 7), all combinations produce 2 ** 15 - 1(2 ** 14 - 1) different schema.
    Therefore, this method includes min(2 ** 15 - 1(2 ** 14 - 1), limit_test) tests
    @type test_fnc: function
    @param test_fnc: name of the generic test function
    @type limit_test: number
    @param limit_test: when enumerations are too many to execute, limit_test is used to limit the number of enumerations
    @type sql_interaction: bool
    @param sql_interaction: whether test involves postgres or not
    """
    if sql_interaction:
        type_params = [char_vdb_type_param, bool_vdb_type_param, int2_vdb_type_param, 
                       int4_vdb_type_param, int8_vdb_type_param, float4_vdb_type_param, 
                       float8_vdb_type_param, bool_array_vdb_type_param, 
                       int2_array_vdb_type_param, int4_array_vdb_type_param, 
                       int8_array_vdb_type_param, float4_array_vdb_type_param, 
                       float8_array_vdb_type_param, float8_2d_array_vdb_type_param]
        oids         = [charoid, booloid, int2oid, int4oid, int8oid, float4oid, float8oid,
                        boolarrayoid, int2arrayoid, int4arrayoid, int8arrayoid, float4arrayoid, 
                        float8arrayoid, float82darrayoid]
    else:
        type_params = [char_type_param, bool_type_param, int2_type_param, int4_type_param, 
                       int8_type_param, float4_type_param, float8_type_param,
                       char_array_type_param, bool_array_type_param, int2_array_type_param, 
                       int4_array_type_param, int8_array_type_param, float4_array_type_param, 
                       float8_array_type_param, float8_2d_array_type_param]
        oids         = [charoid, booloid, int2oid, int4oid, int8oid, float4oid, float8oid,
                        chararrayoid, boolarrayoid, int2arrayoid, int4arrayoid, 
                        int8arrayoid, float4arrayoid, float8arrayoid, float82darrayoid]

    enumerate_and_test_tuples(test_fnc, type_params, oids, limit_test)

def create_table_from_fields(type_params):
    """
    From given type params, this method creates "create table query statement"
    @type type_params: array
    @param type_params: type param for each data type in the query
    @rtype: string
    @return: create table statement
    """    
    create_table_query = 'DROP TABLE IF EXISTS ' + temp_table_name + ';'
    create_table_query += 'CREATE TABLE ' + temp_table_name + '('
    create_table_query += ','.join('f' + str(i) + ' ' + type_params[i][6] for i in range(len(type_params)))
            
    create_table_query += ');'
            
    return create_table_query
